/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * $Id: dcf77.c,v 1.81 2005/09/27 19:20:01 phk Exp $
 *
 * This receiver use the reverse principle of all other DCF77 receivers I have
 * seen implemented, instead of relying on the S/N to be good enough for the
 * correct minute-synchronization to trivially stand out (ie: by the absense
 * of the 59th second pulse), I instead start out with 60 possible hypothesis
 * for where the minute starts, and disprove them as I collect bits.
 *
 * There are many features of the DCF77 timegram which can be checked even
 * though we do not know what time/date it actually is.  For instance a minute
 * reading of 71 would obviously be wrong, as would illegal BCD digits, wrong
 * parity or the two timezone bits being identical.
 *
 * It is possible to also check that the first 14 bits follow a particular
 * pattern, they have been zero every single time I have looked but the
 * PTB reserves the right to play with them.
 *
 * The principle behind this is based on the observation that noise almost
 * always consist of missing pulses, and that '0' and '1' bit pulses very
 * rarely gets confused.
 *
 * This method allows us to tighten the frontend filters much more than
 * other receivers because we can tolerate a high ratio of missing pulses.
 *
 * It is interesting to note that this method does not even require a full
 * minute worth of pulses before the top of the minute is uniquely determined:
 * even though we only have one or two bits actually supporting it, all other
 * possibilities have two or more bits rejecting them as possibilities.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <math.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <stdlib.h>

#include <ntpns.h>
#include "dcf77_cfg.h"
#include "dcf77.h"

static struct logh *l_filter;
struct logh *l_msg;

/*
 * These are our default tweak parameters
 */
static struct dcf77 defaults = {
	.max_gap = 10,
	.max_jitter = .020,
	.low_zero = .070,
	.high_zero = .140,
	.low_one = .165,
	.high_one = .250,
	.ts_zero = .145,
	.ts_one = .255,
	.disproof = 2,
	.squelch = .33,
	.timeout = 600,
	.lockbits = 20,
	.untrack = 2,
	.delay = 0
};

/**********************************************************************/

/*
 * This is slightly complicated, but there is a good reason:  In order to 
 * minimize the effects of noise which mostly take the form of spurious extra
 * pulses, we want to poll the PPS-API as soon as possible after the likely
 * end of the pulse.
 *
 * Once we have a pulse of valid length, we poll again 1.135 second after
 * the start time of that pulse, to look for a following short pulse.
 *
 * If we find a valid pulse at that time, we start over from above, if no 
 * valid pulse is found, we look for a long pulse .1 second later.
 *
 * If we find a valid pulse at that time, we start over from above, if none
 * is found we look for a short pulse .9 second later.
 *
 * A slight complication here is that our pulses are timestamped on
 * CLOCK_REALTIME, but eventlib may run on CLOCK_MONOTONIC and the fact
 * that we can only fiddle the interval setting on recurring timers.
 *
 * All the magic numbers are tweakable.
 *
 *
 *     A     B  C  D  E                                 a     b  c  d  e
 *     v     v  v  v  v                                 v     v  v  v  v
 *
 * ----       ----- ------------------------------------       ----- -------
 *     |     |     |                                    |     |     |
 *     |     |     |                                    |     |     |
 *      ----- -----                                      ----- -----
 *
 *
 */

static void
dcf77_timer(evContext ctx, void *uap, struct timespec due, struct timespec inter __unused)
{
	struct dcf77 *ip;
	double dt, dw, delta;
	struct timespec ts1, ts2;
	int i, j, class;
	time_t t;

	ip = uap;

	if (ip->clock_step != clock_step) {
		dcf77_change_state(ip, DCF_INIT);
		ip->clock_step = clock_step;
		return;
	}
	time(&t);
	t -= ip->t0;
	if (t < 0 || t > ip->timeout) {
		Log(l_msg, "%d No luck, try again t %d", ip->t0, t);
		dcf77_change_state(ip, DCF_INIT);
		return;
	}

	if (ip->lasti == -1) {
		ts2 = evConsTime(0,  90000000);
	} else if (ip->timer_state == SHORT) {
		/*
		 * We are at 'C' and set up things for the case where we find
		 * no pulse, so we aim the timer at 'E' to look for a long
		 * pulse.  In theory we could exploit the PPS-API to tell us
		 * if the first flank of the long pulse happened, but there
		 * seems to be little actual benefit.
		 */
		ip->timer_state = LONG;
		ts2 = evSubTime(ip->sample_one, ip->sample_zero);
	} else {
		/*
		 * Here we are at 'E' and set up things for the case were we
		 * find no pulse, so we aim for 'c'.
		 */
		ip->timer_state = SHORT;
		ts2 = ip->sample_zero;
		ts2.tv_sec++;
		ts2 = evSubTime(ts2, ip->sample_one);
	}
	j = 0;
	i = PpsPickup(&ip->pps, ip->ts, 2);
	if (i != 2) {
		/* Too bad, no pulse, aim timer, done. */
		i = evResetTimer(ctx, ip->tid, dcf77_timer, ip, due, ts2);
		if (i < 0)
			err(1, "evResetTimer() failed %s:%d",
			     __FILE__, __LINE__);
		return;
	}
	record("DCF77PPS", ip->unit, ip->ts, sizeof ip->ts);

	/* So we had a pulse, when and how big ? */

	dw = TsDiff(&ip->ts[1], &ip->ts[0]);

	/* Update pulse-width histogram */
	i = dw / NDWBUCKET;
	if (i >= NDWHIST)
		i = NDWHIST - 1;
	ip->dwhist[i]++;

	if (dw < ip->low_zero)
		class = -2;		/* too short */
	else if (dw < ip->high_zero)
		class = -1;		/* Short pulse, ie: A-B */
	else if (dw < ip->low_one)
		class = 0;		/* Not short, not long, throw away */
	else if (dw < ip->high_one)
		class = 1;		/* Long pulse, ie: A-D */
	else
		class = 2;		/* Too long */

	if (ip->lasti == -1) {
		dt = 1.0;
	} else {
		dt = TsDiff(&ip->ts[0], &ip->ts[2]);
	}

	/* Update pulse-spacing histogram */
	i = (int)(dt + .5);
	if (i < NDTHIST)
		ip->dthist[i]++;
	else
		ip->dthist[NDTHIST - 1]++;

	if (dt > 59 || dt < 0) {
		/*
		 * Loss of signal or timetravel: eject, EJECT!
		 *
		 * '59' is not entirely random.  The blameallocation code
		 * depends on knowing the previos minute and it would be
		 * confused by a missing minute.  It doesn't pay off to
		 * fix this confusion because an empty minute does not
		 * assign any blame anyway.
		 */
		Log(l_msg, "%d loss of signal dt=%f", ip->t0, dt);
		dcf77_change_state(ip, DCF_INIT);
		return;
	}

	if ((fabs(dt / i - 1.0) > ip->max_jitter && dt < ip->max_gap) ||
	    (class != -1 && class != 1)) {
		/*
		 * dt fell out side the wedge-shaped window or dw fell
		 * outside the two permissible ranges: reject pulse.
		 */
		char buf0[30], buf1[30], buf2[30];
		TsFormat(&ip->ts[0], buf0);
		TsFormat(&ip->ts[1], buf1);
		TsFormat(&ip->ts[2], buf2);
		Log(l_filter,
		   "Reject (%d) i %d dw %f dt %f last %s this %s ... %s\n",
		   class, i, dw, dt, buf2, buf0, buf1);
		ip->noise++;
	} else {
		/* Valid pulse, aim timer at 'c' relative to 'A' */

		/* ts1 = 'A' (start of pulse) [UTC sec] */
		TsTimespec(&ip->ts[0], &ts1);

		/* ts2 = 'C' or 'E' (time now) [UTC sec] */
		clock_gettime(CLOCK_REALTIME, &ts2);

		/* ts2 = how long since pulse started [sec] */
		ts2 = evSubTime(ts2, ts1);

		/* ts2 = when did pulse start [eventlib sec] */
		ts2 = evSubTime(evNowTime(), ts2);

		/* ts2 = when should we look again [eventlib sec] */
		ts2 = evAddTime(ip->sample_zero, ts2);
		ts2.tv_sec++;

		/* ts2 = how long from our due is that [sec] */
		ts2 = evSubTime(ts2, due);

		/* Look for short pulse next */
		ip->timer_state = SHORT;

		ip->ts[2] = ip->ts[0];

		TsNormalize(&ip->ts[0]);
		delta = (int64_t)ip->ts[0].ts.bt.frac;
		delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
		delta -= ip->delay;
		if (ip->state == DCF_LOCK) {
			SourceDelta(ip->sid, &ip->ts[0],
			    delta - ip->delta, SRC_UTC);
		} else {
			SourceDelta(ip->sid, &ip->ts[0],
			    delta, SRC_FRACTION);
		}

		/* Round second based on fraction */
		i = ip->ts[0].ts.bt.sec;
		if ((int64_t)ip->ts[0].ts.bt.frac < 0)
			i++;
		i %= 60;

		while (ip->lasti != -1) {
			/* Add missing pulses */
			j = (ip->lasti + 1) % 60;
			if (j == i)
				break;
			dcf77_addpulse(ip, j, 0);
			ip->missing++;
		}
		dcf77_addpulse(ip, i, class);
	}
	i = evResetTimer(ctx, ip->tid, dcf77_timer, ip, due, ts2);
	if (i < 0)
		LogErr(Lerror, "evResetTimer(dcf77) failed %s:%d",
		    __FILE__, __LINE__);
}

/**********************************************************************/

struct ck_instance *
dcf77_new(CK_ARG, u_int unit)
{
	struct dcf77 *ip;
	char *p;
	int i;

	ip = calloc(sizeof *ip, 1);
	if (ip == NULL)
		return (NULL);
	*ip = defaults;
	ip->magic = DCF77_MAGIC;
	ip->unit = unit;
	ip->sid = SourceAdd("dcf77_%u", ip->unit);
	SourceRefid(ip->sid, "DCF\0");
	SourceParam(ip->sid, SRC_UPDATE_RATE, 1.0);
	SourceParam(ip->sid, SRC_LOS_TIMEOUT, 15.0);
	SourceParam(ip->sid, SRC_VOTES, 1.0);
	PpsCreate(&ip->pps, 2);
	ip->sample_zero = evConsTime(0, ip->ts_zero * 1e9);
	ip->sample_one = evConsTime(0, ip->ts_one * 1e9);
	ip->minute = 0;
	ip->target = -1;
	ip->lasti = -1;
	ip->ts[2].ts.ts = evUTCTime();
	ip->ts[2].fmt = TS_TIMESPEC;
	i = evSetTimer(C, dcf77_timer, ip,
	     evConsTime(0, 0), evConsTime(1, 0), &ip->tid);
	if (i < 0)
		err(1, "evSetTimer() failed %s:%d", __FILE__, __LINE__);
	i = evConfigTimer(C, ip->tid, "rate", 0);
	if (i < 0)
		err(1, "evConfigTimer() failed %s:%d", __FILE__, __LINE__);
	p = getenv("TZ");
	if (p == NULL || strcmp(p, "Europe/Berlin")) {
		Log(Linfo, "setting TZ for dcf77");
		putenv("TZ=Europe/Berlin");
		tzset();
	}

	dcf77_change_state(ip, DCF_INIT);
	ip->cip.obj = ip;
	return (&ip->cip);
}

/**********************************************************************/

static void
cfg_dcf77_var(CK_ARG, const char **key, const char *name, double *val, double *nval, double def)
{
	char *p;

        if (key == NULL || !strcmp(*key, name)) {
		p = cfa->tree;
                asprintf(&cfa->tree, "%s %s", p, name);
                ck_RealVar(CK_PASS, nval, val, def, "%g");
		free(cfa->tree);
		cfa->tree = p;
        }
}

int
cfg_dcf77_tweak(CK_ARG, const char **key, double *val)
{
	struct dcf77 *ip;

	ip = cfa->priv;
	Assert(ip->magic == DCF77_MAGIC, "Wrong DCF77 magic");

#define TWEAK(name) \
	cfg_dcf77_var(CK_PASS, key, #name, &ip->name, val, defaults.name)

	TWEAK(max_gap);
	TWEAK(max_jitter);
	TWEAK(low_zero);
	TWEAK(high_zero);
	TWEAK(low_one);
	TWEAK(high_one);
	TWEAK(disproof);
	TWEAK(timeout);
	TWEAK(squelch);
	TWEAK(ts_zero);
	TWEAK(ts_one);
	TWEAK(lockbits);
	TWEAK(untrack);
	TWEAK(delay);
#undef TWEAK
	return (0);
}

int
cfg_dcf77_preamble(CK_ARG, const char **arg)
{
	int i;
	struct dcf77 *ip;
	const char *p;

	ip = cfa->priv;
	Assert(ip->magic == DCF77_MAGIC, "Wrong DCF77 magic");

	if (cfa->flag & CK_NO) {
		for (i = 0; i < NPREAMBLE; i++)
			ip->preamble[i] = 0;
		return (0);
	}
	if (cfa->flag & CK_CONF) {
		for (i = 0; i < NPREAMBLE; i++)
			if (ip->preamble[i])
				break;
		if (i == NPREAMBLE)
			return (0);
	}
	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		ck_printf(CK_PASS, "%s ", cfa->tree);
		for (i = 0; i < NPREAMBLE; i++) {
			if (ip->preamble[i] == -1)
				ck_printf(CK_PASS, "0");
			else if (ip->preamble[i] == 1)
				ck_printf(CK_PASS, "1");
			else
				ck_printf(CK_PASS, "x");
		}
		ck_printf(CK_PASS, "\n");
		return (0);
	}
	p = *arg;
	for (i = 0; i < NPREAMBLE && *p; i++, p++) {
		if (*p == '0')
			ip->preamble[i] = -1;
		else if (*p == '1')
			ip->preamble[i] = 1;
		else if (*p == 'x' || *p == 'X')
			ip->preamble[i] = 0;
		else {
			ck_printf(CK_PASS,
			    "Unknown character in preamble ignored\n");
		}
	}
	return (0);
}

int
dcf77_histogram(CK_ARG, const char **arg)
{
	struct dcf77 *ip;
	int i, j, k, l, b[56];

	ip = cfa->priv;

	if (cfa->flag & (CK_NO | CK_CONF))
		return (0);
	if (!(cfa->flag & CK_SHOW)) {
		if (!strcmp(*arg, "zero")) {
			memset(ip->dwhist, 0, sizeof ip->dwhist);
			memset(ip->dthist, 0, sizeof ip->dthist);
		}
		return (0);
	}

	if (arg != NULL && !strcmp(*arg, "dw")) {
		b[0] = 0;
		b[1] = ip->low_zero / NDWBUCKET;
		b[2] = ip->high_zero / NDWBUCKET;
		b[3] = ip->low_one / NDWBUCKET;
		b[4] = ip->high_one / NDWBUCKET;
		b[5] = NDWHIST;
		ck_printf(CK_PASS, "dW [s] histogram\n");
		ck_printf(CK_PASS, "|%-14s|%-14s|%-14s|%-14s|%-14s|\n",
		    " too short",
		    " zero bit",
		    " in between",
		    " one bit",
		    " too long");
		for (i = 0; i < NDWHIST; i++) {
			j = 0;
			for (k = 0; k < 5; k++) {
				l = b[k] + i;
				if (l >= b[k + 1]) {
					ck_printf(CK_PASS, "|              ");
				} else {
					ck_printf(CK_PASS, "|%.3f %8u",
					     l * NDWBUCKET, ip->dwhist[l]);
					j++;
				}
			}
			ck_printf(CK_PASS, "|\n");
			if (!j)
				break;
		}
		ck_printf(CK_PASS, "\n");
	}
	if (arg != NULL && !strcmp(*arg, "dt")) {
		ck_printf(CK_PASS, "dT [s] histogram\n");
		j = NDTHIST / 5;
		for (i = 0; i < j; i++) {
			for (k = 0; k < 5; k++) {
				l = i + j * k;
				if (l < NDTHIST)
					ck_printf(CK_PASS, "|%2d %9u",
					    l, ip->dthist[l]);
			}
			ck_printf(CK_PASS, "|\n");
		}
		ck_printf(CK_PASS, "\n");
	}
	return (0);
}

int
cfg_dcf77_pps_setup(CK_ARG)
{
	struct dcf77 *ip;

	ip = cfa->priv;
	Assert(ip->magic == DCF77_MAGIC, "Wrong DCF77 magic");
	cfa->priv = ip->pps;
	return (0);
}

int
cfg_dcf77_source_setup(CK_ARG)
{
	struct dcf77 *ip;

	ip = cfa->priv;
	Assert(ip->magic == DCF77_MAGIC, "Wrong DCF77 magic");
	cfa->priv = ip->sid;
	return (0);
}

mod_init_t dcf77_init;

const char *
dcf77_init(void)
{
	static int once = 0;


	if (once) { 
		warnx("dcf77 module already loaded");
		return (NULL);
	} else {
		once = 1;
	}
	ck_TreeAdd(cft, &dcf77_config);

	l_filter = LogNewHandle("dcf77_filter");
	l_msg = LogNewHandle("dcf77_msg");
	return (MAGIC);
}
