/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: ntpns.h,v 1.80 2007/02/07 07:26:33 phk Exp $
 *
 * This file exposes the API for refclocks and other plugins.
 */

#include <sys/queue.h>

#include <timestamp.h>

extern int dopt, Dopt;
extern evContext C;

#define debug(level, ...) \
	do { if (level >= dopt) fprintf(stderr, ## __VA_ARGS__); } while (0)

void fhexdump(FILE *f, u_int len, const void *p);

#define nomem() warn("Could not allocate memory.")

typedef const char * mod_init_t(void);
/* Declare the init functions of loadable modules */

extern struct ck_tree *cft;

/* -------- */
/* config.c */
/* -------- */

#define CONFIG_PROMPT	0x0001

#ifdef CK_ARG
void ConfigSource(FILE *fo, FILE *fi, int fd, int flags);
#endif

/* ----- */
/* log.c */
/* ----- */


struct logh;

struct logh *LogNewHandle(const char *fmt, ...) __printflike(1, 2);
int LogActive(struct logh *lh);
void LogDeleteHandle(struct logh *);
void Log(struct logh *, const char *fmt, ...) __printflike(2, 3);
void LogHex(struct logh *, void *ptr, u_int len, const char *fmt, ...) __printflike(4, 5);
void LogErr(struct logh *, const char *fmt, ...) __printflike(2, 3);
void LogAssert(const char *file, const char *func, int line, const char *fmt, ...) __printflike(4, 5) __dead2;

extern struct logh *Lerror, *Lwarn, *Linfo, *Ldebug;

#define Assert(e, ...) do { if (!(e)) \
    LogAssert(__FILE__, __func__, __LINE__, __VA_ARGS__); \
    } while (0)

#define Abort(...) LogAssert(__FILE__, __func__, __LINE__, __VA_ARGS__)

/* mumble.c */
int Warn(FILE *f, int error, const char *fmt, ...);
int Err(FILE *f, int error, const char *fmt, ...);
int update_string(FILE *f, char **p, const char *q);
#ifdef CK_ARG
int GetBoundSocket(CK_ARG, int af, int fam, int proto, void *adr, int adrlen);
#endif
const char *FormatRefid(unsigned char *u, int stratum);

/* --------- */
/* pll.c */
/* --------- */

extern int clock_step;		/* increments when clock is stepped */

/* --------- */
/* pps-api.c */
/* --------- */

struct ppsdev;

int PpsCreate(struct ppsdev **, int flanks);
#ifdef CK_ARG
void PpsDefaultFd(CK_ARG, struct ppsdev *, int fd);
#endif
void PpsDestroy(struct ppsdev **);
int PpsPickup(struct ppsdev **p, struct ts *ts, int flanks);


/* ---------- */
/* resolver.c */
/* ---------- */
struct addrinfo;
#ifdef CK_ARG
typedef void resfunc_t(CK_ARG, void *priv, int error, struct addrinfo *ai);
int resolver(CK_ARG, const char *host, const char *port, int protofam, int socktype, int flags, resfunc_t *func, void *priv);
#endif

/* -------- */
/* record.c */
/* -------- */

int rrrdupdate(const char *file, const char *template, const char *fmt, ...);
void record(const char *ident, int unit, const void *p, int len);

/* ------- */
/* shmem.c */
/* ------- */
void *shmem_alloc(u_int length, u_int tag);
void shmem_free(void *ptr);

/* -------- */
/* source.c */
/* -------- */

struct source;

typedef enum {
	SRC_INVALID = 1,
	SRC_LOS = 2,
	SRC_FREQUENCY = 3,
	SRC_FRACTION = 4,
	SRC_UTC = 5
} src_state_e;

typedef enum {
	SRC_UPDATE_RATE,
	SRC_LOS_TIMEOUT,
	SRC_VOTES,
	SRC_DISPERSION,
	SRC_DELAY,
	SRC_LIMIT,
	SRC_STRATUM
} src_param_e;

typedef enum {
	SRC_LEAP_UNKNOWN,
	SRC_LEAP_NO,
	SRC_LEAP_INSERT,
	SRC_LEAP_DELETE
} src_leap_e;


struct source *SourceAdd(const char *fmt, ...) __printflike(1, 2);
void SourceDel(struct source *sid);
void SourceDelta(struct source *sid, struct ts *ts, double delta, src_state_e state);
void SourceTick(struct source *sid, struct ts *ts, src_state_e state);
void SourceParam(struct source *sid, src_param_e param, double val);
void SourceRefid(struct source *sid, const void *refid);
void SourceLeap(struct source *sid, src_leap_e leap);
#ifdef CK_HELP
ck_func_t SourceConfig;
#endif

struct sys_state {
	u_char	__sys_refid[4];
	double	__sys_delay;
	double	__sys_dispersion;
	int	__sys_precision;
	int	__sys_stratum;
	uint	__sys_able;
	src_leap_e	__sys_leap;
	char	__selsource[32];
	char	__clkstate[32];
};

extern struct sys_state sys_state;
#define	sys_refid	sys_state.__sys_refid
#define	sys_delay	sys_state.__sys_delay
#define	sys_dispersion	sys_state.__sys_dispersion
#define	sys_precision	sys_state.__sys_precision
#define	sys_stratum	sys_state.__sys_stratum
#define sys_able	sys_state.__sys_able
#define sys_leap	sys_state.__sys_leap

#define ABLE_STARTUP			(1 << 0)
#define ABLE_PLL_UNLOCKED		(1 << 1)
#define ABLE_NO_SOURCE			(1 << 2)
#define ABLE_SOURCE_OFFSET		(1 << 3)
