/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: source.h,v 1.19 2005/09/12 12:36:17 phk Exp $
 *
 * NB: This file is only for inclusion on clock selectors and the source 
 * NB: management code.  The interface for refclocks is in ntpns.h
 *
 * A time source has the following properties:
 *
 *	Identity
 *		char *name;
 *		The name of the clock source, the name is decided by the code
 *		that offers the clock source and is typically type/instance
 *		like "m12/0"
 *
 *	Update rate
 *		double update_rate; 	[sec]
 *		How often we expect to receive an independent measurement from
 *		this source.  Note well: "independent measurement"
 *
 *	LOS timeout
 *		double los_timeout;	[sec]
 *		How long time before we vote this source out.  Configurable.
 *
 *	Votes
 *		double votes;
 *		How many independent votes does this source provide.  For a
 *		GPS source this should be number of tracked non-removed
 *		satellites.  For a VLF receiver it should be one, for Loran-C
 *		number of LORSTA tracked, etc etc.  Fractional part can be used
 *		to tweak ordering.
 *
 *	Source dispersion
 *		double dispersion;	[sec]
 *		Dispersion as reported by time source.
 *
 *	State
 *		enum {...} state;
 *		OFFLINE -- no signal available
 *		FREQ -- non-UTC locked frequency standard;
 *		FRAC -- UTC locked fractional seconds.
 *		UTC -- UTC locked.
 *
 */

#define NTIMELINE	8		/*
					 * Number of samples at each averaging
					 * time.  Cannot be lower than the 8
					 * positions specified for reachability
					 * shift-reg in RFC1305. Can be higher.
					 */

#define NTIMEHIST	20		/*
					 * How many averaging times do we keep
					 * track off.  Power of two buckets
					 * starting with 1 second.
					 * 2^19 sec ~= 1 week
					 */

struct timeline {
	int	v[NTIMELINE];		/* Valid flag */
	double	d[NTIMELINE];		/* Delta time [sec] */
	double	s;			/* Sum of valid samples [sec] */
	int 	n;			/* Number of valid samples */
	double	a;			/* MVAR of valid samples */
};

/*
 * The clock selection code gets to define this one, but we name it
 * rather than use a void* for typesafety and coding convenience.
 */

struct clockselect;

#define SOURCE_TAG		0x53524300U

struct source {
	u_int			magic;
#define SOURCE_MAGIC	(~SOURCE_TAG)
	u_int			idx;
	char 			name[32];
	TAILQ_ENTRY(source)	list;
	struct ck_instance	cip;

	u_int			flags;
#define SRC_F_DATA		(1<<0)
#define SRC_F_PARAM		(1<<1)
#define SRC_F_NEW		(1<<2)
#define SRC_F_DELETE		(1<<3)
#define SRC_F_SELECTABLE	(1<<4)
#define SRC_F_RRRD		(1<<5)

	src_state_e		state;
	int			update_rate;
	int			los_timeout;
	double			votes;
	double			dispersion;
	double			delay;
	double			limit;
	int			last_sample;
	int			stratum;
	u_char			refid[4];
	src_leap_e		leap;

	u_int			tick;

	double			jitter;
	double			last_delta;
	struct ts		last_ts;
	int			los;	/* los of signal */

	int			hist0;

	struct timeline		t[NTIMEHIST];

	struct clockselect	*clk_sel;
};

TAILQ_HEAD(sourcehd, source);

extern struct sourcehd		sources;

/*
 * A clock selection algorithm must supply just one function which will
 * be called with three different enum arguments to notify of "start to
 * use", "do your job" and "end of use".  After "end of use" the routine
 * can be unloaded.
 */

typedef enum {INIT, TICK, DEINIT} clk_sel_e;

typedef struct source *clk_sel_t(struct source *current, clk_sel_e action);

/* Register and start to use this clock selection function */
void ClockSelect(clk_sel_t *func);
