/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: log.c,v 1.26 2005/10/04 06:44:22 phk Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <configkit.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <isc/eventlib.h>

#include "ntpns.h"
#include "ntpns_cfg.h"
#include "internal.h"

struct logh *Lerror, *Lwarn, *Linfo, *Ldebug;

static struct cfgsrc *term_fd;

static struct ck_instancehead instance_head =
    TAILQ_HEAD_INITIALIZER(instance_head);

struct logh {
	LIST_ENTRY(logh)	next;
	struct ck_instance	cip;
	const char		*name;
	u_int			flags;
	u_int			std_flags;
};

/* Bottom three bits are for syslog priority */
#define LF_ENABLE		(1U<<3)
#define LF_SYSLOG		(1U<<4)
#define LF_STDOUT		(1U<<5)
#define LF_STDERR		(1U<<6)
#define LF_TERM			(1U<<7)
#define LF_DISABLE		(1U<<31) 	/* pseudo bit ! */

static LIST_HEAD(, logh)	top = LIST_HEAD_INITIALIZER(top);

struct logh *
LogNewHandle(const char *fmt, ...)
{
	struct logh *hp;
	va_list ap;
	char *p;

	va_start(ap, fmt);
	vasprintf(&p, fmt, ap);
	va_end(ap);

	hp = calloc(sizeof *hp, 1);
	if (hp != NULL) {
		hp->name = p;
		hp->cip.obj = hp;
		hp->cip.idx.word = strdup(p);
		TAILQ_INSERT_TAIL(&instance_head, &hp->cip, next);
		LIST_INSERT_HEAD(&top, hp, next);
	}
	return (hp);
}

#if 0
/* XXX: we need to rip out from configkit */
void
LogDeleteHandle(struct logh *hp)
{
}
#endif

void
LogErr(struct logh *lh, const char *fmt, ...)
{
	va_list ap;
	char *p, *e;

	e = strerror(errno);

	if (!(lh->flags & LF_ENABLE))
		return;
	va_start(ap, fmt);
	vasprintf(&p, fmt, ap);
	va_end(ap);
	if (p == NULL)
		return;		/* XXX: count one missing */
	if (lh->flags & LF_SYSLOG)
		syslog(lh->flags & LOG_PRIMASK, "%s: %s", p, e);
	if (lh->flags & LF_STDOUT) {
		fprintf(stdout, "%s: %s\n", p, e);
		fflush(stdout);
	}
	if (lh->flags & LF_STDERR) {
		fprintf(stderr, "%s: %s\n", p, e);
		fflush(stderr);
	}
	free(p);
}

int
LogActive(struct logh *lh)
{

	return (lh->flags & LF_ENABLE);
}

static void
LogS(struct logh *lh, const char *s)
{

	if (!(lh->flags & LF_ENABLE))
		return;
	if (lh->flags & LF_SYSLOG)
		syslog(lh->flags & LOG_PRIMASK, s);
	if (lh->flags & LF_STDOUT) {
		fprintf(stdout, "%s", s);
		fflush(stdout);
	}
	if (lh->flags & LF_STDERR) {
		fprintf(stderr, "%s", s);
		fflush(stderr);
	}
	if ((lh->flags & LF_TERM))
		ConfigLog(term_fd, s);
}

static void
LogV(struct logh *lh, const char *fmt, va_list ap)
{
	char *p;

	if (!(lh->flags & LF_ENABLE))
		return;
	vasprintf(&p, fmt, ap);
	if (p == NULL)
		return;		/* XXX: count one missing */
	LogS(lh, p);
	free(p);
}

static int
ascii(int x)
{
	if (x < ' ')
		return '.';
	if (x > '~')
		return '.';
	return (x);
}

void
LogHex(struct logh *lh, void *ptr, u_int len, const char *fmt, ...)
{
	va_list ap;
	u_char *p;
	u_int i, j;
	char cb[2];
	char hb[4];

	if (!(lh->flags & LF_ENABLE))
		return;
	va_start(ap, fmt);
	LogV(lh, fmt, ap);
	va_end(ap);
	p = ptr;
	cb[1] = '\0';
	if (len == 0)
		return;
	for (i = 0; i < len; i += 16) {
		for (j = 0; j < 16 && i + j < len; j++) {
			sprintf(hb, " %02x", p[i + j]);
			LogS(lh, hb);
		}
		for (; j < 16; j++)
			sprintf(hb, "   ");
		sprintf(hb, "  ");
		for (j = 0; j < 16 && i + j < len; j++) {
			cb[0] = ascii(p[i + j]);
			LogS(lh, cb);
		}
		LogS(lh, "\n");
	}
}

void
Log(struct logh *lh, const char *fmt, ...)
{
	va_list ap;

	if (!(lh->flags & LF_ENABLE))
		return;
	va_start(ap, fmt);
	LogV(lh, fmt, ap);
	LogS(lh, "\n");
	va_end(ap);
}

void
LogAssert(const char *file, const char *func, int line, const char *fmt, ...)
{
	va_list ap;
	char *p, *q, *e;
	int i;

	i = errno;
	e = strerror(i);
	va_start(ap, fmt);
	vasprintf(&p, fmt, ap);
	va_end(ap);

	asprintf(&q, "Assertion failed\n%s:%d %s(): %s\nerrno = %d (%s)",
	     file, line, func, p, i, e);
	syslog(LOG_EMERG, "%s", q);
	fprintf(stderr, "%s", q);
	fflush(stderr);
	fprintf(stdout, "%s", q);
	fflush(stdout);
	abort();
}

struct ck_instancehead *
cfg_msg_head(CK_ARG)
{

	return (&instance_head);
}

struct ck_instance *
cfg_msg_new(CK_ARG, const char *name __unused)
{

	return (NULL);
}

int
cfg_msg_show(CK_ARG)
{
	struct logh *lh;

	lh = cfa->priv;
	ck_printf(CK_PASS, "%-16s", lh->name);
	if (lh->flags & LF_STDOUT)
		ck_printf(CK_PASS, " stdout");
	if (lh->flags & LF_STDERR)
		ck_printf(CK_PASS, " stderr");
	if (lh->flags & LF_SYSLOG)
		ck_printf(CK_PASS, " syslog");
	if (lh->flags & LF_TERM)
		ck_printf(CK_PASS, " term");
	if (lh->flags & LF_ENABLE)
		ck_printf(CK_PASS, " enable");
	ck_printf(CK_PASS, "\n");
	return (0);
}

static int
cfg_msg_bitflip(CK_ARG, u_int bit)
{
	struct logh *lh;

	lh = cfa->priv;
	if (cfa->flag & CK_CONF) {
		/* We never put these bits in the config */
		if (bit & (LF_DISABLE|LF_TERM))
			return (0);
		if ((lh->flags & bit) != (lh->std_flags & bit)) {
			if (lh->flags & bit)
				ck_printf(CK_PASS, "%s\n", cfa->tree);
			else
				ck_printf(CK_PASS, "no %s\n", cfa->tree);
		}
		return (0);
	}
	if (bit == LF_DISABLE) {
		if (cfa->flag & CK_NO)
			lh->flags |= LF_ENABLE;
		else
			lh->flags &= ~LF_ENABLE;
	} else {
		if (cfa->flag & CK_NO)
			lh->flags &= ~bit;
		else
			lh->flags |= bit;
	}
	return (0);
}

int
cfg_msg_enable(CK_ARG)
{

	/*
	 * We want to always generate some (empty) output for each
	 * instance, in order to prevent configkit from outputting
	 * the "msg foo" instance marker in the conf
	 */
	ck_printf(CK_PASS, "");
	return cfg_msg_bitflip(CK_PASS, LF_ENABLE);
}

int
cfg_msg_disable(CK_ARG)
{

	return cfg_msg_bitflip(CK_PASS, LF_DISABLE);
}

int
cfg_msg_stdout(CK_ARG)
{

	return cfg_msg_bitflip(CK_PASS, LF_STDOUT);
}

int
cfg_msg_stderr(CK_ARG)
{

	return cfg_msg_bitflip(CK_PASS, LF_STDERR);
}

int
cfg_msg_term(CK_ARG)
{

	if (!(cfa->flag & (CK_CONF|CK_SHOW|CK_SHOW|CK_NO)))
		term_fd = cfa->src;
	return cfg_msg_bitflip(CK_PASS, LF_TERM);
}

int
cfg_msg_syslog(CK_ARG, const char **arg __unused)
{

	return cfg_msg_bitflip(CK_PASS, LF_SYSLOG);
}

void
log_close(struct cfgsrc *fd)
{
	struct logh *lh;

	if (fd != term_fd)
		return;

	term_fd = NULL;

	LIST_FOREACH(lh, &top, next) 
		lh->flags &= ~LF_TERM;
}

const char *
init_log(void)
{

	openlog("NTPns", LOG_NDELAY, LOG_DAEMON);

	Lerror = LogNewHandle("error");
	Lerror->flags |= LF_ENABLE | LF_SYSLOG | LF_STDERR | LOG_ERR;
	Lerror->std_flags = Lerror->flags;

	Lwarn = LogNewHandle("warning");
	Lwarn->flags |= LF_ENABLE | LF_SYSLOG | LF_STDERR | LOG_WARNING;
	Lwarn->std_flags = Lwarn->flags;

	Linfo = LogNewHandle("info");
	Linfo->flags |= LF_ENABLE | LF_SYSLOG | LF_STDOUT | LOG_INFO;
	Linfo->std_flags = Linfo->flags;

	Ldebug = LogNewHandle("debug");
	Ldebug->flags |= LF_SYSLOG | LF_STDOUT | LOG_DEBUG;
	Ldebug->std_flags = Ldebug->flags;

	return (MAGIC);
}
