/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: mumble.c,v 1.6 2005/09/30 11:03:07 phk Exp $
 *
 */

#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <ntpns.h>

static void
WarnV(FILE *f, int error, const char *fmt, va_list ap)
{

	if (fmt != NULL)
		vfprintf(f, fmt, ap);
	if (error != 0)
		fprintf(f, ": %s", strerror(error));
	fprintf(f, "\n");
}

int
Warn(FILE *f, int error, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	WarnV(f, error, fmt, ap);
	va_end(ap);
	return (-1);
}

int
Err(FILE *f, int error, const char *fmt, ...)
{
	va_list ap;

	fprintf(f, "%s: ", getprogname());
	va_start(ap, fmt);
	WarnV(f, error, fmt, ap);
	va_end(ap);
	exit (1);
}

int
update_string(FILE *f, char **p, const char *q)
{

	*p = realloc(*p, strlen(q) + 1);
	if (*p == NULL)
		return Warn(f, errno, "malloc failed");
	strcpy(*p, q);
	return (0);
}

int
GetBoundSocket(CK_ARG, int af, int fam, int proto, void *adr, int adrlen)
{
	int s, i, j;

	s = socket(af, fam, proto); 
	if (s < 0) {
		ck_printf(CK_PASS,
		    "socket failed for record: %s", strerror(errno));
		return (-1);
	}
	i = fcntl(s, F_GETFL, &j);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "fcntl F_GETFL record socket: %s", strerror(errno));
		close (s);
		return (-1);
	}
	j |= O_NONBLOCK;
	i = fcntl(s, F_SETFL, &j);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "fcntl F_SETFL record socket: %s", strerror(errno));
		close (s);
		return (-1);
	}
	
	i = bind(s, adr, adrlen);
	if (i != 0) {
		ck_printf(CK_PASS, "bind failed on record socket: %s",
		    strerror(errno));
		close (s);
		return (-1);
	}
	return (s);
}

static const char *refclk[] = {
	[0] =	"NONE",
	[1] =	"LOCALCLOCK",
	[2] =	"GPS_TRAK",
	[3] =	"WWV_PST",
	[4] =	"SPECTRACOM",
	[5] =	"TRUETIME",
	[6] =	"IRIG_AUDIO",
	[7] =	"CHU_AUDIO",
	[8] =	"PARSE",
	[9] =	"GPS_MX4200",
	[10] =	"GPS_AS2201",
	[11] =	"GPS_ARBITER",
	[12] =	"IRIG_TPRO",
	[13] =	"ATOM_LEITCH",
	[14] =	"MSF_EES",
	[15] =	"GPSTM_TRUE",
	[16] =	"IRIG_BANCOMM",
	[17] =	"GPS_DATUM",
	[18] =	"NIST_ACTS",
	[19] =	"WWV_HEATH",
	[20] =	"GPS_NMEA",
	[21] =	"GPS_VME",
	[22] =	"ATOM_PPS",
	[23] =	"PTB_ACTS",
	[24] =	"USNO",
	[26] =	"GPS_HP",
	[27] =	"ARCRON_MSF",
	[28] =	"SHM",
	[29] =	"PALISADE",
	[30] =	"ONCORE",
	[31] =	"GPS_JUPITER",
	[32] =	"CHRONOLOG",
	[33] =	"DUMBCLOCK",
	[34] =	"ULINK",
	[35] =	"PCF",
	[36] =	"WWV_AUDIO",
	[37] =	"FG",
	[38] =	"HOPF_SERIAL",
	[39] =	"HOPF_PCI",
	[40] =	"JJY",
	[41] =	"TT560",
	[42] =	"ZYFER",
	[43] =	"RIPENCC",
	[44] =	"NEOCLOCK4X",
};

static const int nrefclk = sizeof(refclk) / sizeof(refclk[0]);

const char *
FormatRefid(unsigned char *u, int stratum)
{
	static char buf[30];
	int i;

	if (u[0] == 127 && u[1] == 127 && u[2] < nrefclk) {
		sprintf(buf, "[%s#%u]", refclk[u[2]], u[3]);
		return (buf);
	}
	if (u[0] == 0 && u[1] == 0 && u[2] == 0 && u[3] == 0)
		return ("[]");
	for (i = 3; i > 0; i--)
		if (u[i] != 0)
			break;
	for (; i >= 0; i--)
		if (u[i] < ' ' || u[i] > '~')
			stratum = 2;
	if (stratum > 1) {
		sprintf(buf, "[%d.%d.%d.%d]", u[0], u[1], u[2], u[3]);
		return (buf);
	}
	sprintf(buf, "[%.4s]", u);
	return (buf);
}

