/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pps-api.c,v 1.25 2005/09/25 21:52:54 phk Exp $
 *
 * This is the main entrance where we set up the event engine, parse the
 * options and set the ball rolling by starting events to read the config
 * file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <configkit.h>
#include <isc/eventlib.h>

#include "ntpns.h"
#include "ntpns_cfg.h"
#include "sys/timepps.h"

struct ppsdev {
	int		magic;
#define PPS_MAGIC	0x2783

	/* Configuration values */
	int		default_fd;
	char		*device;
	int		flanks;
	int		cf_flank;
	double		offset;

	/* Running values */
	struct timespec	ts_offset;
	int		fd;
	int		flank;
	pps_handle_t	handle;
	int		mode;
	pps_params_t	params;
	pps_info_t	event;
	FILE		*simulate;
};

int
PpsCreate(struct ppsdev **ppp, int flanks)
{
	struct ppsdev *pp;

	if (flanks < 1 || flanks > 2)
		err(1, "Wrong number of flanks to PpsCreate: %d", flanks);

	if (*ppp != NULL)
		return (0);

	pp = calloc(sizeof *pp, 1);
	if (pp == NULL)
		return (ENOMEM);

	pp->default_fd = -1;
	pp->fd = -1;
	pp->flanks = flanks;
	pp->magic = PPS_MAGIC;
	*ppp = pp;
	return (0);
}

static void
PpsClose(struct ppsdev *pp)
{

	if (pp->fd >= 0 && pp->fd != pp->default_fd)
		close(pp->fd);
	pp->fd = -1;
}

static void
PpsOpen(CK_ARG, struct ppsdev *pp)
{
	int i;
	const char *ed;
	struct timespec to;

	if (pp->fd >= 0) 
		PpsClose(pp);
	pp->fd = pp->default_fd;
	ed = "(default)";
	if (pp->device != NULL) {
		pp->fd = open(pp->device, O_RDONLY | O_NDELAY);
		if (pp->fd < 0)
			ck_printf(CK_PASS, "Could not open PPS device %s: %s\n",
			    pp->device, strerror(errno));
		ed = pp->device;
	}
	if (pp->fd < 0)
		return;

	i = time_pps_create(pp->fd, &pp->handle);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "PPS device %s does not support PPS-API: %s\n",
		    ed, strerror(errno));
		PpsClose(pp);
		return;
	}
	i = time_pps_getcap(pp->handle, &pp->mode);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "PPS device %s does not properly support PPS-API (getcap): %s\n",
		    ed, strerror(errno));
		PpsClose(pp);
		return;
	}
	i = time_pps_getparams(pp->handle, &pp->params);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "PPS device %s does not properly support PPS-API (getparams): %s\n",
		    ed, strerror(errno));
		PpsClose(pp);
		return;
	}
	if (pp->flanks == 2 && !(pp->mode & PPS_CAPTUREASSERT)) {
		ck_printf(CK_PASS,
		    "PPS device %s does not support two flanks (only clear)\n",
		    ed);
		PpsClose(pp);
		return;
	}
	if (pp->flanks == 2 && !(pp->mode & PPS_CAPTURECLEAR)) {
		ck_printf(CK_PASS, 
		    "PPS device %s does not support two flanks (only assert)\n",
		    ed);
		PpsClose(pp);
		return;
	}
	if (pp->flank != 0 && !(pp->mode & pp->flank)) {
		ck_printf(CK_PASS,
		    "PPS device %s does not support configured flank\n", ed);
		PpsClose(pp);
		return;
	}

	pp->flank = pp->cf_flank;
	if (pp->flanks == 2)
		pp->params.mode |= PPS_CAPTUREBOTH;
	else if (pp->flank == PPS_CAPTUREASSERT)
		pp->params.mode |= PPS_CAPTUREASSERT;
	else if (pp->flank == PPS_CAPTURECLEAR)
		pp->params.mode |= PPS_CAPTURECLEAR;
	else if (pp->mode & PPS_CAPTUREASSERT) {
		pp->params.mode |= PPS_CAPTUREASSERT;
		pp->flank = PPS_CAPTUREASSERT;
	} else {
		pp->params.mode |= PPS_CAPTURECLEAR;
		pp->flank = PPS_CAPTURECLEAR;
	}

	if (pp->offset != 0.0) {
		pp->params.assert_offset = pp->ts_offset;
		pp->params.clear_offset = pp->ts_offset;
	}

	i = time_pps_setparams(pp->handle, &pp->params);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "PPS device %s does not properly support PPS-API (setparams): %s\n",
		    ed, strerror(errno));
		PpsClose(pp);
		return;
	}
	to = evConsTime(0, 0);
	i = time_pps_fetch(pp->handle, PPS_TSFMT_TSPEC, &pp->event, &to);
	if (i < 0) {
		ck_printf(CK_PASS,
		    "PPS device %s does not properly support PPS-API (getch): %s\n",
		    ed, strerror(errno));
		PpsClose(pp);
		return;
	}
	return;
}

void
PpsDefaultFd(CK_ARG, struct ppsdev *pp, int fd)
{

	Assert(pp->magic == PPS_MAGIC, "Wrong PPS struct magic");
	if (pp->fd == pp->default_fd)
		PpsClose(pp);
	pp->default_fd = fd;
	if (pp->fd == -1 && pp->default_fd != -1)
		PpsOpen(CK_PASS, pp);
}

int
config_pps_flank(CK_ARG, const char **flank)
{
	struct ppsdev *pp;

	pp = cfa->priv;
	Assert(pp->magic == PPS_MAGIC, "Wrong PPS struct magic");

	if (cfa->flag & CK_NO) {
		pp->cf_flank = 0;
		return (0);
	}
	if (cfa->flag & (CK_CONF | CK_SHOW)) {
		if (pp->cf_flank == PPS_CAPTUREASSERT)
			ck_printf(CK_PASS, "%s assert\n", cfa->tree);
		if (pp->cf_flank == PPS_CAPTURECLEAR)
			ck_printf(CK_PASS, "%s clear\n", cfa->tree);
		return (0);
	}
	if (!strcmp(*flank, "clear"))
		pp->cf_flank = PPS_CAPTURECLEAR;
	else if (!strcmp(*flank, "assert"))
		pp->cf_flank = PPS_CAPTUREASSERT;
	PpsClose(pp);
	PpsOpen(CK_PASS, pp);
	return (0);
}

int
config_pps_offset(CK_ARG, double *offset)
{
	struct ppsdev *pp;
	int64_t off;
	int i;

	pp = cfa->priv;
	Assert(pp->magic == PPS_MAGIC, "Wrong PPS struct magic");

	if (!ck_RealVar(CK_PASS, offset, &pp->offset, 0.0, "%.9f"))
		return (0);

	pp->offset = *offset;
	off = pp->offset * 1e9;
	pp->ts_offset.tv_sec = off / 1000000000;
	pp->ts_offset.tv_nsec = off % 1000000000;
	if (pp->fd < 0)
		return (0);
	i = time_pps_getparams(pp->handle, &pp->params);
	if (i < 0) {
		ck_printf(CK_PASS, "PPS device does not properly support PPS-API (getparams): %s\n",
		    strerror(errno));
		PpsClose(pp);
		return (0);
	}
	pp->params.assert_offset.tv_sec = off / 1000000000;
	pp->params.assert_offset.tv_nsec = off % 1000000000;
	pp->params.clear_offset = pp->params.assert_offset;
	i = time_pps_setparams(pp->handle, &pp->params);
	if (i < 0) {
		ck_printf(CK_PASS, "PPS device does not properly support PPS-API (setparams): %s\n",
		    strerror(errno));
		PpsClose(pp);
		return (0);
	}
	/* XXX ioctl */
	return (0);
}

int
config_pps_device(CK_ARG, const char **device)
{
	struct ppsdev *pp;

	pp = cfa->priv;
	Assert(pp->magic == PPS_MAGIC, "Wrong PPS struct magic");

	if (!ck_WordVar(CK_PASS, device, &pp->device))
		return (0);
	PpsClose(pp);
	if (pp->device == NULL)
		return (0);
	PpsOpen(CK_PASS, pp);
	return (0);
}

/*
 * Pick up a PPS-API measurement.
 *
 * The flank argument determines how many flanks we want to see change,
 * this us useful for things like DCF77/MSF/WWWB receivers which want
 * to measure the duration of the pulses.
 */

int
PpsPickup(struct ppsdev **pp, struct ts *ts, int flanks)
{
	pps_info_t ev;
	struct ppsdev *p;
	struct timespec to;
	struct ts *t1;
	int i;
	char simbuf[200];

	p = *pp;
	ts->fmt = TS_INVALID;
	if (p == NULL) {
		Log(Lerror, "PpsPickup(NULL)");
		return (0);
	}
	if (p->simulate != NULL) {
		if (NULL == fgets(simbuf, sizeof simbuf, p->simulate)) {
			Log(Lerror, "EOF on pps simulate file");
			exit (1);
		}
		i = sscanf(simbuf, "%u .%ld %u %u .%ld %u",
		    &ev.assert_timestamp.tv_sec,
		    &ev.assert_timestamp.tv_nsec,
		    &ev.assert_sequence,
		    &ev.clear_timestamp.tv_sec,
		    &ev.clear_timestamp.tv_nsec,
		    &ev.clear_sequence);
		if (i != 6) {
			Log(Lerror, "format error in pps simulate file");
			return (-1);
		}
		i = 0;
	} else {
		to = evConsTime(0, 0);
		i = time_pps_fetch(p->handle, PPS_TSFMT_TSPEC, &ev, &to);
	}
	if (0)
		Log(Ldebug, "PPS Raw: %d %d.%09ld %u -> %d.%09ld %u", i,
		    ev.assert_timestamp.tv_sec, ev.assert_timestamp.tv_nsec,
		    ev.assert_sequence,
		    ev.clear_timestamp.tv_sec, ev.clear_timestamp.tv_nsec,
		    ev.clear_sequence);
	if (i < 0) {
		warn("time_pps_fetch() failed");
		return (-1);
	}
	if (flanks == 1 && p->flank == PPS_CAPTUREASSERT) {
		if (ev.assert_sequence == p->event.assert_sequence)
			return (0);
		ts->ts.ts = ev.assert_timestamp;
		ts->fmt = TS_TIMESPEC;
		p->event = ev;
		return (p->flanks);
	}
	if (flanks == 1 && p->flank == PPS_CAPTURECLEAR) {
		if (ev.clear_sequence == p->event.clear_sequence)
			return (0);
		ts->ts.ts = ev.clear_timestamp;
		ts->fmt = TS_TIMESPEC;
		p->event = ev;
		return (p->flanks);
	}
	t1 = ts + 1;
	if (p->flank == PPS_CAPTUREASSERT) {
		if (ev.clear_sequence == p->event.clear_sequence)
			return (0);
		ts->ts.ts = ev.assert_timestamp;
		ts->fmt = TS_TIMESPEC;
		t1->ts.ts = ev.clear_timestamp;
		t1->fmt = TS_TIMESPEC;
	} else {
		if (ev.assert_sequence == p->event.assert_sequence)
			return (0);
		ts->ts.ts = ev.clear_timestamp;
		ts->fmt = TS_TIMESPEC;
		t1->ts.ts = ev.assert_timestamp;
		t1->fmt = TS_TIMESPEC;
	}
	if (TsDiff(t1, ts) < .0)
		return (0);
	if (0)
		Log(Ldebug, "PPS  %d.%09ld %u -> %d.%09ld %u",
		    ev.assert_timestamp.tv_sec, ev.assert_timestamp.tv_nsec,
		    ev.assert_sequence,
		    ev.clear_timestamp.tv_sec, ev.clear_timestamp.tv_nsec,
		    ev.clear_sequence);
	p->event = ev;
	return (p->flanks);
}
