/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: record.c,v 1.29 2005/12/30 10:07:41 phk Exp $
 *
 * "Record" is for dumping diagnostic and debugging info to a remote machine.
 * The way it works is that we wrap the data in a UDP packet and kicks it
 * off without looking back.  The receiving gets to have fun decoding things.
 *
 * Packet format:
 *	BE32		sequence number
 *	C[8]		ident
 *	BE32		timestamp (time_t)
 *	BE16		unitno
 *	BE16		length;
 *	C[...]		data
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <configkit.h>
#include <stdarg.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <math.h>
#include <netdb.h>

#include "ntpns.h"
#include "ntpns_cfg.h"
#include "internal.h"

static uint32_t seq = 0;
static struct sockaddr record_sa;
static int record_salen;

static int record_sock = -1;
static int record_enable;
static int record_resolved;
static char *record_host;
static char *record_port;

void
record(const char *ident, int unit, const void *p, int len)
{
	u_char hd[20];
	time_t t;
	struct msghdr h;
	struct iovec v[2];
	int i;

	seq++;
	if (record_enable == 0)
		return;
	if (record_resolved == 0)
		return;

	if (len > 20000)
		len = 20000;
	memset(&h, 0, sizeof h);

	time(&t);
	be32enc(hd, seq);
	strncpy(hd + 4, ident, 8);
	be32enc(hd + 12, t);
	be16enc(hd + 16, unit);
	be16enc(hd + 18, len);

	v[0].iov_base = hd;
	v[0].iov_len = 20;
	v[1].iov_base = __DECONST(void *, p);
	v[1].iov_len = len;

	h.msg_name = &record_sa;
	h.msg_namelen = record_salen;
	h.msg_iov = v;
	h.msg_iovlen = 2;

	i = sendmsg(record_sock, &h, 0);
	if (i < 0)
		warn("sendmsg (%d, %p, %d)", record_sock, &h, 0);
}

static void
resolved_record(CK_ARG, void *priv __unused, int error, struct addrinfo *ai)
{

	if (error) {
		ck_printf(CK_PASS,
		    "failed to resolve address for record (%d %s)\n", 
		    error, gai_strerror(error));
		return;
	}

	record_sa = *ai->ai_addr;
	record_salen = ai->ai_addrlen;
	record_resolved = 1;
}

static int
conf_record(CK_ARG)
{
	int i;
	const char *mport;
	struct sockaddr_in sins;

	if (!record_enable)
		return (0);
	if (record_host == NULL)
		return (0);
	if (record_port == NULL)
		mport = "4444";
	else
		mport = record_port;
	if (record_sock < 0) {
		memset(&sins, 0, sizeof sins);
		sins.sin_len = sizeof sins;
		sins.sin_family = AF_INET;
		record_sock = GetBoundSocket(CK_PASS, AF_INET, SOCK_DGRAM, 0,
		    &sins, sizeof sins);
	}
	if (record_sock < 0)
		return (-1);
	record_resolved = 0;
	i = resolver(CK_PASS, record_host, mport,
	     PF_INET, SOCK_DGRAM, AI_CANONNAME, resolved_record, NULL);
	if (i < 0)
		ck_printf(CK_PASS, "Failed...\n");

	return (0);
}

int
cfg_record_host(CK_ARG, const char **host)
{

	if (!ck_WordVar(CK_PASS, host, &record_host))
		return (0);
	conf_record(CK_PASS);
	return (0);
}

int
cfg_record_port(CK_ARG, const char **port)
{

	if (!ck_WordVar(CK_PASS, port, &record_port))
		return (0);
	conf_record(CK_PASS);
	return (0);
}

int
cfg_record_enable(CK_ARG)
{

	if (cfa->flag & CK_SHOW) {
		ck_printf(CK_PASS, "%s is %s\n", cfa->tree,
			record_enable ? "enabled" : "disabled");
		return (0);
	}
	if (cfa->flag & CK_CONF) {
		if (record_enable)
			ck_printf(CK_PASS, "%s\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_NO)
		record_enable = 0;
	else
		record_enable = 1;
	conf_record(CK_PASS);
	return (0);
}
