/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: resolver.c,v 1.23 2007/01/18 09:42:16 phk Exp $
 *
 * Gack!  Nobody seems to have offered us an asynchronous resolver yet,
 * so build our own from a pthread.  I could have used rfork() as well,
 * but since I expect other pthread uses, we might as well bite the
 * bullet up front.  Use a pipe to signal when the thread has finished
 * the lookup.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <configkit.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <math.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pthread.h>

#include "ntpns.h"
#include "internal.h"

struct reswork {
	char	 	*host;
	char 		*port;
	int		protofam;
	int		socktype;
	struct addrinfo	hints;
	struct addrinfo	*res;
	int 		error;
	int		pipe[2];
	evFileID	fid;
	resfunc_t	*func;
	void		*priv;
	struct ck_args	*cfa;
	pthread_t	thread;
};

static void *
res_thread(void *p)
{
	struct reswork *rw;

	rw = p;
	rw->hints.ai_family = rw->protofam;
	rw->hints.ai_socktype = rw->socktype;
	rw->error = getaddrinfo(rw->host, rw->port, &rw->hints, &rw->res);
	write(rw->pipe[1], "", 1);
	pthread_exit(NULL);
}

static void
resolved(struct reswork *rw)
{
	struct addrinfo *a;

	fprintf(stderr, "Resolved(%s, %s, %d, %d, %d): Error %d\n",
	     rw->host, rw->port, rw->protofam, rw->socktype,
	     rw->hints.ai_flags, rw->error);
	for (a = rw->res; a != NULL; a = a->ai_next) {
		if (a->ai_canonname != NULL)
			fprintf(stderr,
			    "Canonical \"%s\"\n", a->ai_canonname);
		fprintf(stderr, "flag %d fam %d sock %d proto %d\n",
		    a->ai_flags, a->ai_family, a->ai_socktype, a->ai_protocol);
	}
}

static void
EvRes(evContext ctx, void *uap, int fd __unused, int eventmask __unused)
{
	struct reswork *rw;
	char b[1];

	rw = uap;
	evDeselectFD(ctx, rw->fid);
	read(rw->pipe[0], b, 1);
	close(rw->pipe[0]);
	close(rw->pipe[1]);
	pthread_join(rw->thread, NULL);
	if (dopt && 0)
		resolved(rw);
	rw->func(rw->cfa, rw->priv, rw->error, rw->res);
	if (rw->res != NULL)
		freeaddrinfo(rw->res);
	if (rw->host != NULL)
		free(rw->host);
	free(rw->port);
	free(rw);
}

int
resolver(CK_ARG, const char *host, const char *port, int protofam, int socktype, int flags, resfunc_t *func, void *priv)
{
	struct reswork *rw;
	int i;

	rw = calloc(sizeof *rw, 1);
	if (rw == NULL) {
		ck_printf(CK_PASS, "malloc failed: %s\n", strerror(errno));
		return (-1);
	}
	i = pipe(rw->pipe);
	if (i != 0) {
		ck_printf(CK_PASS, "pipe(2) failed: %s\n", strerror(errno));
		free(rw);
		return (-1);
	}
	if (dopt && 0)
		fprintf(stderr, "Resolve \"%s\" \"%s\" fd %d %d\n",
		    host, port, rw->pipe[0], rw->pipe[1]);
	if (host == NULL) {
		rw->host = NULL;
	} else {
		rw->host = strdup(host);
		if (rw->host == NULL) {
			ck_printf(CK_PASS, "malloc failed: %s\n", strerror(errno));
			close(rw->pipe[0]);
			close(rw->pipe[1]);
			free(rw);
			return (-1);
		}
	}
	rw->port = strdup(port);
	if (rw->port == NULL) {
		ck_printf(CK_PASS, "malloc failed: %s\n", strerror(errno));
		close(rw->pipe[0]);
		close(rw->pipe[1]);
		if (rw->host != NULL)
			free(rw->host);
		free(rw);
		return (-1);
	}
	rw->protofam = protofam;
	rw->socktype = socktype;
	rw->func = func;
	rw->priv = priv;
	rw->hints.ai_flags = flags;
	rw->cfa = cfa;
	i = evSelectFD(C, rw->pipe[0], EV_READ | EV_EXCEPT, EvRes,
	     rw, &rw->fid);
	if (i)
		err(1, "evSelectFD() failed in resolver");
	i = pthread_create(&rw->thread, NULL, res_thread, rw);
	if (i != 0) {
		ck_printf(CK_PASS, "pthread_create failed: %s\n", strerror(errno));
		close(rw->pipe[0]);
		close(rw->pipe[1]);
		free(rw->port);
		if (rw->host != NULL)
			free(rw->host);
		free(rw);
		return (-1);
	}
	return (1);
}
