/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: rrrd.c,v 1.11 2005/09/25 22:12:54 phk Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <configkit.h>
#include <stdarg.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <sys/resource.h>
#include <math.h>
#include <netinet/in.h>
#include <netdb.h>

#include "ntpns.h"
#include "ntpns_cfg.h"
#include "internal.h"

static int rrrd_sock = -1;

static struct sockaddr rrrd_sa;
static int rrrd_salen;

static int rrrd_enable;
static int rrrd_resolved;
static int rrrd_resources;
static char *rrrd_host;
static char *rrrd_port;
static char *rrrd_prefix;

static evTimerID resource_tid;

static void
resource_timer(evContext ctx __unused, void *uap __unused, struct timespec t0 __unused, struct timespec dt __unused)
{
	int i;
	struct rusage r;
	uintptr_t up;

	up = (uintptr_t)sbrk(0);
	i = getrusage(0, &r);
	Assert(i == 0, "getrusage()");
	rrrdupdate("rusage", "brk:sbrk:utime:stime:maxrss:ixrss:idrss:isrss:minflt:majflt:nswap:inblock:oublock:msgsnd:msgrcv:nsignals:nvcsw:nivcsw",
	    "N:%ju:%ju:%u:%u:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu",
		(uintmax_t)up, (uintmax_t)up,
		r.ru_utime.tv_sec,
		r.ru_stime.tv_sec,
		r.ru_maxrss, r.ru_ixrss, r.ru_idrss,
		r.ru_isrss, r.ru_minflt, r.ru_majflt,
		r.ru_nswap, r.ru_inblock, r.ru_oublock,
		r.ru_msgsnd, r.ru_msgrcv, r.ru_nsignals,
		r.ru_nvcsw, r.ru_nivcsw);
}

int
rrrdupdate(const char *file, const char *template, const char *fmt, ...)
{
	int error;
	char *buf1, *buf2;
	va_list ap;

	if (!rrrd_enable)
		return (-1);
	if (!rrrd_resolved)
		return (-1);
	if (rrrd_salen == 0)
		return (-1);
	if (fmt == NULL) 
		return (-1);

	va_start(ap, fmt);
	error = vasprintf(&buf1, fmt, ap);
	va_end(ap);
	if (error == -1 || buf1 == NULL)
		return (-1);
	error = asprintf(&buf2, "00 %s%s %s %s",
	    rrrd_prefix, file, template, buf1);
	free(buf1);
	if (error == -1 || buf2 == NULL)
		return (-1);
	error = sendto(rrrd_sock, buf2, strlen(buf2), 0, &rrrd_sa, rrrd_salen);
	free(buf2);
	return (error);
}

static void
resolved_rrrd(CK_ARG, void *priv __unused, int error, struct addrinfo *ai)
{

	if (error) {
		ck_printf(CK_PASS,
		    "failed to resolve address for record (%d %s)\n", 
		    error, gai_strerror(error));
		return;
	}

	rrrd_sa = *ai->ai_addr;
	rrrd_salen = ai->ai_addrlen;
	rrrd_resolved = 1;
}

static int
conf_rrrd(CK_ARG)
{
	int i;
	const char *mport;
	struct sockaddr_in sins;

	if (!rrrd_enable)
		return (0);
	if (rrrd_host == NULL)
		return (0);
	if (rrrd_port == NULL)
		mport = "4444";
	else
		mport = rrrd_port;
	if (rrrd_sock < 0) {
		memset(&sins, 0, sizeof sins);
		sins.sin_len = sizeof sins;
		sins.sin_family = AF_INET;
		rrrd_sock = GetBoundSocket(CK_PASS, AF_INET, SOCK_DGRAM, 0,
		    &sins, sizeof sins);
	}
	if (rrrd_sock < 0)
		return (-1);
	rrrd_resolved = 0;
	i = resolver(CK_PASS, rrrd_host, mport,
	     PF_INET, SOCK_DGRAM, AI_CANONNAME, resolved_rrrd, NULL);
	if (i < 0)
		ck_printf(CK_PASS, "Failed...\n");

	return (0);
}

int
cfg_rrrd_host(CK_ARG, const char **host)
{

	if (!ck_WordVar(CK_PASS, host, &rrrd_host))
		return (0);
	conf_rrrd(CK_PASS);
	return (0);
}

int
cfg_rrrd_port(CK_ARG, const char **port)
{

	if (!ck_WordVar(CK_PASS, port, &rrrd_port))
		return (0);
	conf_rrrd(CK_PASS);
	return (0);
}

int
cfg_rrrd_prefix(CK_ARG, const char **prefix)
{

	if (!ck_WordVar(CK_PASS, prefix, &rrrd_prefix))
		return (0);
	return (0);
}

int
cfg_rrrd_resources(CK_ARG)
{
	int i;

	if (cfa->flag & CK_SHOW) {
		ck_printf(CK_PASS, "%s is %s\n", cfa->tree,
			rrrd_resources ? "enabled" : "disabled");
		return (0);
	}
	if (cfa->flag & CK_CONF) {
		if (rrrd_resources)
			ck_printf(CK_PASS, "%s\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_NO) {
		if (rrrd_resources)
			evClearTimer(C, resource_tid);
		rrrd_resources = 0;
		return (0);
	} else {
		if (!rrrd_resources) {
			i = evSetTimer(C, resource_timer, &resource_tid,
			    evConsTime(0, 0), evConsTime(60, 0), &resource_tid);
			Assert(i == 0, "evSetTimer(resource_timer)");
			i = evConfigTimer(C, resource_tid, "rate", 0);
			Assert(i == 0, "evConfigTimer(source_timer)");
		}
		rrrd_resources = 1;
		return (0);
	}
}

int
cfg_rrrd_enable(CK_ARG)
{

	if (cfa->flag & CK_SHOW) {
		ck_printf(CK_PASS, "%s is %s\n", cfa->tree,
			rrrd_enable ? "enabled" : "disabled");
		return (0);
	}
	if (cfa->flag & CK_CONF) {
		if (rrrd_enable)
			ck_printf(CK_PASS, "%s\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_NO)
		rrrd_enable = 0;
	else
		rrrd_enable = 1;
	conf_rrrd(CK_PASS);
	return (0);
}
