/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: source.c,v 1.87 2005/12/30 10:27:45 phk Exp $
 *
 * Herding time sources.
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <math.h>

#include "ntpns.h"
#include "ntpns_cfg.h"
#include "shmem.h"
#include "source.h"
#include "pll.h"
#include "internal.h"

struct sys_state sys_state;

static struct ntpns_shmem *ntp_shmem;

static struct logh *Lsource;

static evTimerID	source_tid;
static int myclock_step = 0;

static struct ck_instancehead source_head =
	TAILQ_HEAD_INITIALIZER(source_head);

struct sourcehd sources = TAILQ_HEAD_INITIALIZER(sources);

static struct source *sync_source;

static double source_leap[] = {
	[SRC_LEAP_UNKNOWN]	=	0.0,
	[SRC_LEAP_NO]		=	0.0,
	[SRC_LEAP_INSERT]	=	0.0,
	[SRC_LEAP_DELETE]	=	0.0
};

clk_state_e clk_state = NOSRC;
struct source *selsrc;
static clk_state_e next_state = NOSRC;
static clk_sel_t *clk_sel;

static const char *
state_name(clk_state_e st)
{
	switch (st) {
        case NOSRC:		return "<nosrc>";
        case FRAC_SWITCH:	return "<frac_switch>";
        case FRAC_TRACK:	return "<frac_track>";
        case UTC_STEP:		return "<utc_step>";
        case SRC_SWITCH:	return "<src_switch>";
        case UTC_LOCK:		return "<utc_lock>";
        case COAST:		return "<coast>";
        case FRAC_LOCK:		return "<frac_lock>";
	default:		return "<invalid>";
	}
}

static const char *
ref_name(src_state_e st)
{
	switch (st) {
	case SRC_INVALID:	return "<invalid>";
	case SRC_LOS:		return "<LOS>";
	case SRC_FREQUENCY:	return "<frequency>";
	case SRC_FRACTION:	return "<fraction>";
	case SRC_UTC:		return "<UTC>";
	default:		return "<?>";
	}
}

static void
timer_kick(struct source *sid)
{
	int i, k, l, tv;
	u_int u;
	double td, a, b;
	struct timeline *tl;

	/* XXX: check sid->state changes */
	if (sid->flags & SRC_F_DATA) {
		tv = 1;
		td = sid->last_delta;
		sid->flags &= ~SRC_F_DATA;
		sid->los = 0;
	} else {
		tv = 0;
		td = -20011966.999;	/* We should be able to spot this */
		sid->los++;
	}
	u = sid->tick;
	sid->tick += 2;
	for (i = 0; i < NTIMEHIST; i++) {

		/* We only react on every 2^i'th sample */
		if (u & 1)
			break;
		u >>= 1;

		/* update sum fields with the sample we discard */
		tl = &sid->t[i];
		if (tl->v[NTIMELINE - 1]) {
			tl->n--;
			tl->s -= tl->d[NTIMELINE - 1];
		}

		/* Shift the register */
		memmove(&tl->v[1], &tl->v[0],
		     sizeof(tl->v[0]) * (NTIMELINE - 1));
		memmove(&tl->d[1], &tl->d[0],
		     sizeof(tl->d[0]) * (NTIMELINE - 1));

		/* Insert new sample */
		tl->v[0] = tv;
		tl->d[0] = td;
		if (tv) {
			tl->n++;
			tl->s += td;
		}

		/*
		 * If this sample were not valid, we will use the previous
		 * sample to update the next row nomatter what validity
		 * it has.
		 */
		if (!tv) {
			tv = tl->v[1];
			td = tl->d[1];
		}
	
		/* Calculate modified allan variance, if we can. */
		if (tl->n < 3) {
			tl->a = 1.0;
		} else {
			a = 0;
			l = 0;
			for (k = 0; k < NTIMELINE - 3; k++) {
				if (!(tl->v[k] && tl->v[k + 1] && tl->v[k + 2]))
					continue;
				b = tl->d[k + 2] + tl->d[k] - 2 * tl->d[k + 1];
				b /= (1 << i);
				a += b * b;
				l++;
			}
			if (l == 0)
				tl->a = 1.0;
			else
				tl->a = sqrt((1.0 / (2.0 * l)) * a);
		}
	}
	record("SRC", 0, sid, sizeof *sid);
}

static void
source_leapstatus(void)
{
	struct source *sp;

	/* Update collective leap second status */
	source_leap[SRC_LEAP_UNKNOWN] = 0.0;
	source_leap[SRC_LEAP_NO] = 0.0;
	source_leap[SRC_LEAP_INSERT] = 0.0;
	source_leap[SRC_LEAP_DELETE] = 0.0;
	TAILQ_FOREACH(sp, &sources, list)
		if (sp->state > SRC_LOS)
			source_leap[sp->leap] += sp->votes;

	/* Go with any qualified majority */
	if (source_leap[SRC_LEAP_NO] >
	    (source_leap[SRC_LEAP_INSERT] + source_leap[SRC_LEAP_DELETE]))
		sys_leap = SRC_LEAP_NO;
	else if (source_leap[SRC_LEAP_INSERT] >
	    (source_leap[SRC_LEAP_NO] + source_leap[SRC_LEAP_DELETE]))
		sys_leap = SRC_LEAP_INSERT;
	else if (source_leap[SRC_LEAP_DELETE] >
	    (source_leap[SRC_LEAP_NO] + source_leap[SRC_LEAP_INSERT]))
		sys_leap = SRC_LEAP_DELETE;
	else
		sys_leap = SRC_LEAP_NO;
}

static void
source_timer(evContext ctx, void *uap, struct timespec t0, struct timespec dt)
{
	struct timespec t1, t2;
	struct source *sp, *sp2;
	double a;
	int i, j;

	/* Make sure the source timer runs at 100msec before each UTC sec */
	t1 = evUTCTime();
	j = t1.tv_nsec - 900000000;
	t0 = evSubTime(t0, evConsTime(1, j));
	t2 = evAddTime(t0, dt);
	i = evResetTimer(ctx, source_tid, source_timer, uap,
		t2, evConsTime(1, 0));
	Assert(i == 0, "evResetTimer(source_timer)");

	source_leapstatus();

	TAILQ_FOREACH(sp, &sources, list)
		timer_kick(sp);

	if (sync_source != NULL && sync_source->flags & SRC_F_DELETE)
		sync_source->los = sync_source->los_timeout + 1;
	next_state = clk_state;
	selsrc = clk_sel(sync_source, TICK);

	if (selsrc != NULL)
		strcpy(sys_state.__selsource, selsrc->name);
	else
		strcpy(sys_state.__selsource, "(null)");

	strcpy(sys_state.__clkstate, state_name(clk_state));

	record("SYS", 0, &sys_state, sizeof sys_state);

	switch(clk_state) {
	case NOSRC:
		/*
		 * This is the state we start in, and where we end up whenever
		 * we loose track of time.  Hang around until a usable source
		 * gets selected.
		 */
		if (selsrc == NULL)
			next_state = NOSRC;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_STEP;
		else if (selsrc->state == SRC_FRACTION)
			next_state = FRAC_SWITCH;
		break;
	case FRAC_SWITCH:
		/*
		 * Switch to a fractional source.
		 * If we are too far off, step the clock, and go to NOSRC.
		 * Otherwise, pass into tracking.
		 */
		if (selsrc == NULL)
			next_state = NOSRC;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_STEP;
		else if (selsrc != sync_source)
			next_state = NOSRC;
		else if (selsrc->state != SRC_FRACTION) 
			next_state = NOSRC;
		else if (sync_source->t[sync_source->hist0].v[0]) {
			if (a < -.050 || a > .050) {
				pllstep(a);
				next_state = NOSRC;
			} else {
				next_state = FRAC_TRACK;
			}
		}
		break;
	case FRAC_TRACK:
		/*
		 * Track a fractional source as long as we can or until
		 * we get a better offer.
		 */
		if (sync_source->los > sync_source->los_timeout)
			next_state = NOSRC;
		else if (selsrc == NULL)
			next_state = NOSRC;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_STEP;
		else if (selsrc != sync_source)
			next_state = NOSRC;
		break;
	case UTC_STEP:
		/*
		 * We have seen our first UTC source, and may have to
		 * step the clock. 
		 */
		if (selsrc != sync_source)
			next_state = NOSRC;
		else if (sync_source->state != SRC_UTC)
			next_state = NOSRC;
		else if (sync_source->los > sync_source->los_timeout)
			next_state = NOSRC;
		else if (sync_source->t[sync_source->hist0].v[0]) {
			a = sync_source->t[sync_source->hist0].d[0];
			if (a < -.020 || a > .020) {
				pllstep(a);
				next_state = NOSRC;
			} else {
				next_state = SRC_SWITCH;
			}
		}
		break;
	case SRC_SWITCH:
		/*
		 * Locking down on a new synchronized source.
		 */
		if (selsrc == NULL)
			next_state = COAST;
		else if (selsrc->los > selsrc->los_timeout)
			next_state = COAST;
		else if (selsrc->state == SRC_FRACTION)
			next_state = FRAC_LOCK;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_LOCK;
		else
			next_state = COAST;
		break;
	case UTC_LOCK:
		if (sync_source->los > sync_source->los_timeout)
			next_state = SRC_SWITCH;
		else if (sync_source->state != SRC_UTC)
			next_state = SRC_SWITCH;
		else if (selsrc != sync_source)
			next_state = SRC_SWITCH;
		break;
	case COAST:
		/* XXX: timeout */
		if (selsrc == NULL)
			;
		else if (selsrc->state == SRC_UTC)
			next_state = SRC_SWITCH;
		else if (selsrc->state == SRC_FRACTION)
			next_state = SRC_SWITCH;
		break;
	case FRAC_LOCK:
		if (sync_source->los > sync_source->los_timeout)
			next_state = SRC_SWITCH;
		else if (selsrc != sync_source)
			next_state = SRC_SWITCH;
		else if (selsrc->state != SRC_FRACTION)
			next_state = SRC_SWITCH;
		break;
	default:
		Abort("Illegal clock state %d", clk_state);
	}

	if (clk_state != next_state)
		Log(Lsource, "Clock state %s -> %s",
		    state_name(clk_state), state_name(next_state));

	switch(next_state) {
	case NOSRC:
		sync_source = NULL;
		sys_able |= ABLE_NO_SOURCE;
		memset(sys_refid, 0, sizeof sys_refid);
		i = 0;
		break;
	case FRAC_SWITCH:
		sync_source = selsrc;
		sys_able |= ABLE_NO_SOURCE;
		pll_reset();
		i = 0;
		break;
	case FRAC_TRACK:
		sync_source = selsrc;
		sys_able |= ABLE_NO_SOURCE;
		i = 1;
		break;
	case UTC_STEP:
		sync_source = selsrc;
		sys_able |= ABLE_NO_SOURCE;
		i = 0;
		break;
	case SRC_SWITCH:
		sync_source = selsrc;
		sys_able |= ABLE_NO_SOURCE;
		pll_reset();	/* XXX: really ? */
		i = 0;
		break;
	case UTC_LOCK:
		sync_source = selsrc;
		sys_able &= ~ABLE_NO_SOURCE;
		memcpy(sys_refid,  sync_source->refid, sizeof sys_refid);
		i = 1;
		break;
	case COAST:
		sync_source = NULL;
		sys_able &= ~ABLE_NO_SOURCE;
		memcpy(sys_refid, "LOCL", 4);
		i = 0;
		break;
	case FRAC_LOCK:
		i = 1;
		sync_source = selsrc;
		sys_able &= ~ABLE_NO_SOURCE;
		memcpy(sys_refid,  sync_source->refid, sizeof sys_refid);
		break;
	default:
		Abort("Illegal new clock state %d", next_state);
	}

	if (sync_source != NULL) {
		if (fabs(sync_source->t[sync_source->hist0].d[0]) >
		    sync_source->limit)
			sys_able |= ABLE_SOURCE_OFFSET;
		else
			sys_able &= ~ABLE_SOURCE_OFFSET;
	}

	clk_state = next_state;
	ntp_shmem->clockstate = (int)next_state;
	if (sync_source == NULL)
		ntp_shmem->clocksource = 0;
	else
		ntp_shmem->clocksource = sync_source->idx;

	if (i && sync_source->t[sync_source->hist0].v[0]) {
		pll(sync_source->t[sync_source->hist0].d[0]);
		sys_delay = sync_source->delay;
		sys_dispersion = sync_source->dispersion +
			sync_source->t[sync_source->hist0].a;
		sys_stratum = sync_source->stratum + 1;
	} else {
		pllnone();
	}

	Assert((sync_source == NULL) || !(sync_source->flags & SRC_F_DELETE),
	    "deleting sync_source");

	TAILQ_FOREACH_SAFE(sp, &sources, list, sp2) {
		if (clock_step != myclock_step)
			memset(sp->t, 0, sizeof sp->t);
		sp->flags &= ~(SRC_F_PARAM | SRC_F_NEW);
		if (sp->flags & SRC_F_DELETE) {
			TAILQ_REMOVE(&source_head, &sp->cip, next);
			TAILQ_REMOVE(&sources, sp, list);
			shmem_free(sp);
		}
	}
	myclock_step = clock_step;
}

/*
 * Default clock selector
 *
 * Simply pick the source with the highest votes in the highest state.
 */
static struct source *
default_clock_selector(struct source *cur, clk_sel_e action)
{
	struct source *sp;

	if (action != TICK)
		return (NULL);

	TAILQ_FOREACH(sp, &sources, list) {
		if (sp->los > sp->los_timeout)
			continue;
		if (sp->votes <= 0)
			continue;
		if (cur == NULL) {
			cur = sp;
			continue;
		}
		if (sp->state < cur->state)
			continue;
		if (sp->state > cur->state) {
			cur = sp;
			continue;
		}
		if (sp->votes > cur->votes) {
			cur = sp;
			continue;
		}
		if (sp->update_rate < cur->update_rate) {
			cur = sp;
			continue;
		}
	}
	if (cur != NULL && cur->los > cur->los_timeout)
		cur = NULL;
	return (cur);
}

/*
 * Public interface to sources
 */

void
ClockSelect(clk_sel_t *func)
{

	if (clk_sel != NULL)
		(void)clk_sel(NULL, DEINIT);
	if (func == NULL)
		clk_sel = default_clock_selector;
	else
		clk_sel = func;
	(void)clk_sel(NULL, INIT);
}

struct source *
SourceAdd(const char *fmt, ...)
{
	struct source *sp;
	va_list ap;
	static u_int idx = 10000;


	sp = shmem_alloc(sizeof *sp, SOURCE_TAG);
	if (sp == NULL)
		return (NULL);
	sp->idx = idx++;
	va_start(ap, fmt);
	vsnprintf(sp->name, sizeof sp->name, fmt, ap);
	va_end(ap);
	TAILQ_INSERT_TAIL(&sources, sp, list);
	sp->flags |= SRC_F_NEW;
	sp->los_timeout = 60;
	sp->los = sp->los_timeout + 1;
	sp->state = SRC_INVALID;
	sp->limit = .128;
	sp->magic = SOURCE_MAGIC;
	sp->cip.obj = sp;
	sp->cip.idx.word = sp->name;
	TAILQ_INSERT_TAIL(&source_head, &sp->cip, next);
	return (sp);
}

void
SourceDel(struct source *sid)
{

	sid->flags |= SRC_F_DELETE;
}

void
SourceTick(struct source *sid, struct ts *ts, src_state_e state)
{
	double delta;
	struct ts mts;

	if (ts == NULL)
		return;
	if (state == SRC_FRACTION) {
		TsNormalize(ts);
		delta = ts->ts.bt.frac;
		delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
		if (delta > .5)
			delta -= 1.0;
	} else if (state == SRC_UTC) {
		mts.ts.ts = evUTCTime();
		mts.ts.ts.tv_nsec = 0;
		mts.fmt = TS_TIMESPEC;
		delta = TsDiff(&mts, ts);
	} else {
		delta = 0;
	}

	SourceDelta(sid, ts, delta, state);
}

void
SourceDelta(struct source *sid, struct ts *ts, double delta, src_state_e state)
{
	char buf[30];

	if (ts == NULL)
		return;

	TsFormat(ts, buf);
	if (state == SRC_UTC)
		sid->last_delta = delta;
	else {
		sid->last_delta = delta - (int)delta;
		if (sid->last_delta > .5)
			sid->last_delta -= 1.0;
	}
	Log(Lsource, "SourceDelta(%s), %s %12.9f (%12.9f) %s",
	    sid->name, buf, delta, sid->last_delta, ref_name(state));
	sid->last_ts = *ts;
	TsNormalize(&sid->last_ts);
	sid->state = state;
	sid->flags |= SRC_F_DATA;
	if (sid->flags & SRC_F_RRRD) {
		rrrdupdate(sid->name, "offset:votes:state:dispersion:delay",
		    "N:%.9f:%.3f:%d:%.9f:%.9f",
		    sid->last_delta, sid->votes, sid->state,
		    sid->dispersion, sid->delay);
	}
}

void
SourceRefid(struct source *sid, const void *refid)
{

	memmove(sid->refid, refid, sizeof sid->refid);
	sid->flags |= SRC_F_PARAM;
}

void
SourceParam(struct source *sid, src_param_e param, double val)
{
	int i;

	switch (param) {
	case SRC_UPDATE_RATE:
		sid->update_rate = val;
		for (i = 0; i < NTIMEHIST; i++)
			if ((1 << i) >= val)
				break;
		sid->hist0 = i;
		break;
	case SRC_LOS_TIMEOUT:	sid->los_timeout = val;	break;
	case SRC_VOTES:		sid->votes = val; 	break;
	case SRC_DISPERSION:	sid->dispersion = val;	break;
	case SRC_STRATUM:	sid->stratum = val;	break;
	case SRC_DELAY:		sid->delay = val;	break;
	case SRC_LIMIT:		sid->limit = val;	break;
	default:
		Abort("Unknown source property %d", param);
		break;
	}
	sid->flags |= SRC_F_PARAM;
}

void
SourceLeap(struct source *sid, src_leap_e leap)
{

	sid->leap = leap;
}

struct ck_instancehead *
cfg_source_head(CK_ARG)
{

	return (&source_head);
}

struct ck_instance *
cfg_source_new(CK_ARG, const char *name __unused)
{

	return (NULL);
}

static void
source_show_allan_data(CK_ARG, struct source *sp, int data)
{
	int i, j;
			
	for (i = 0; i < NTIMEHIST; i++) {
		if (sp->t[i].n < 3)
			continue;
		ck_printf(CK_PASS, "%6d: ", 1 << i);
		for (j = 0; j < NTIMELINE; j++)
			ck_printf(CK_PASS, "%d", sp->t[i].v[j]);
		if (sp->t[i].n) {
			ck_printf(CK_PASS, " %12.9f", sp->t[i].s / sp->t[i].n);
			ck_printf(CK_PASS, " %.3e", sp->t[i].a);
		}
		if (data) {
			for (j = 0; j < NTIMELINE; j++) {
				if (sp->t[i].v[j])
					ck_printf(CK_PASS,
					    " %12.9f", sp->t[i].d[j]);
				else
					ck_printf(CK_PASS, " %12s", "");
			}
		}
		ck_printf(CK_PASS, "\n");
	}
}

int
config_source_limit(CK_ARG, double *limit)
{
	struct source *sid;

	sid = cfa->priv;
	if (sid == NULL)
		return (0);
	Assert(sid->magic == SOURCE_MAGIC, "Wrong SOURCE struct magic");

	ck_RealVar(CK_PASS, limit, &sid->limit, .128, "%.9f");
	return (0);
}

int
config_source_rrrd(CK_ARG)
{
	struct source *sid;

	sid = cfa->priv;
	if (sid == NULL)
		return (0);
	Assert(sid->magic == SOURCE_MAGIC, "Wrong SOURCE struct magic");

	if (cfa->flag & (CK_CONF | CK_SHOW)) {
		if (sid->flags & SRC_F_RRRD)
			ck_printf(CK_PASS, "%s\n", cfa->tree);
		else if (cfa->flag & CK_SHOW)
			ck_printf(CK_PASS, "%s not set\n", cfa->tree);
		return (0);
	}
	if (cfa->flag & CK_NO)
		sid->flags &= ~SRC_F_RRRD;
	else
		sid->flags |= SRC_F_RRRD;
	return (0);
}

static void
source_show(CK_ARG, struct source *sp)
{
	char buf[30];

	ck_printf(CK_PASS, "Source %s: votes %f flags %s",
	    sp->name, sp->votes, ref_name(sp->state));
	ck_printf(CK_PASS, " los %d/%d update %d",
		sp->los, sp->los_timeout, sp->update_rate);
	if (sp == selsrc)
		ck_printf(CK_PASS, " SELECTED");
	ck_printf(CK_PASS, "\n\tlimit %e\t\t", sp->limit);
	switch(sp->leap) {
	case SRC_LEAP_UNKNOWN:
		ck_printf(CK_PASS, "Unknown leap second status");
		break;
	case SRC_LEAP_NO:
		ck_printf(CK_PASS, "No leapsecond at end of today");
		break;
	case SRC_LEAP_INSERT:
		ck_printf(CK_PASS, "Insert leapsecond at end of today");
		break;
	case SRC_LEAP_DELETE:
		ck_printf(CK_PASS, "Delete leapsecond at end of today");
		break;
	}
	ck_printf(CK_PASS, "\n\tstratum %d", sp->stratum);
	ck_printf(CK_PASS, "\t\t\trefid %s\n",
	    FormatRefid(sp->refid, sp->stratum + 1));
	ck_printf(CK_PASS, "\tdelay %.09f\t\tdispersion %.09f\n",
	    sp->delay, sp->dispersion);
	TsFormat(&sp->last_ts, buf);
	ck_printf(CK_PASS, "\tlast_ts %s\tlast_delta %.9f\n",
	    buf, sp->last_delta);
}

int
show_source(CK_ARG)
{
	struct source *sid, *sp;

	sid = cfa->priv;
	if (sid != NULL) {
		Assert(sid->magic == SOURCE_MAGIC,
		    "Wrong SOURCE struct magic (0x%x)", sid->magic);
		source_show(CK_PASS, sid);
		return (0);
		
	}
	TAILQ_FOREACH(sp, &sources, list) 
		source_show(CK_PASS, sp);
	return (0);
}

int
show_source_allan(CK_ARG)
{
	struct source *sid, *sp;

	sid = cfa->priv;
	if (sid != NULL) {
		Assert(sid->magic == SOURCE_MAGIC, "Wrong SOURCE struct magic");
		source_show(CK_PASS, sid);
		source_show_allan_data(CK_PASS, sid, 0);
		return (0);
		
	}
	TAILQ_FOREACH(sp, &sources, list) {
		source_show(CK_PASS, sp);
		source_show_allan_data(CK_PASS, sp, 0);
	}
	return (0);
}

int
show_source_data(CK_ARG)
{
	struct source *sid, *sp;

	sid = cfa->priv;
	if (sid != NULL) {
		Assert(sid->magic == SOURCE_MAGIC, "Wrong SOURCE struct magic");
		source_show(CK_PASS, sid);
		source_show_allan_data(CK_PASS, sid, 1);
		return (0);
		
	}
	TAILQ_FOREACH(sp, &sources, list) {
		source_show(CK_PASS, sp);
		source_show_allan_data(CK_PASS, sp, 1);
	}
	return (0);
}

int
cfg_source_show(CK_ARG)
{
	struct source *sid;

	sid = cfa->priv;
	source_show(CK_PASS, sid);
	return (0);
}

int
cfg_leap(CK_ARG)
{
	struct source *sp;

	if (!(cfa->flag & CK_SHOW))
		return (0);

	/* Update collective leap second status */
	source_leap[SRC_LEAP_UNKNOWN] = 0.0;
	source_leap[SRC_LEAP_NO] = 0.0;
	source_leap[SRC_LEAP_INSERT] = 0.0;
	source_leap[SRC_LEAP_DELETE] = 0.0;
	ck_printf(CK_PASS, "%-20s %-11s  %7s  %7s  %7s  %7s\n",
	    "Source", "State", "Unknown", "No", "Insert", "Delete");
	TAILQ_FOREACH(sp, &sources, list) {
		if (sp->state > SRC_LOS)
			source_leap[sp->leap] += sp->votes;
		ck_printf(CK_PASS, "%-20s %-11s",
			sp->name, ref_name(sp->state));
		if (sp->state <= SRC_LOS) {
			ck_printf(CK_PASS, "\n");
			continue;
		}
		if (sp->leap == SRC_LEAP_UNKNOWN)
			ck_printf(CK_PASS, "  %7.3f", sp->votes);
		else
			ck_printf(CK_PASS, "  %7s", "-    ");

		if (sp->leap == SRC_LEAP_NO)
			ck_printf(CK_PASS, "  %7.3f", sp->votes);
		else
			ck_printf(CK_PASS, "  %7s", "-    ");

		if (sp->leap == SRC_LEAP_INSERT)
			ck_printf(CK_PASS, "  %7.3f", sp->votes);
		else
			ck_printf(CK_PASS, "  %7s", "-    ");

		if (sp->leap == SRC_LEAP_DELETE)
			ck_printf(CK_PASS, "  %7.3f\n", sp->votes);
		else
			ck_printf(CK_PASS, "  %7s\n", "-    ");
	}
	ck_printf(CK_PASS, "%-20s %11s  %7.3f  %7.3f  %7.3f  %7.3f\n",
	    "Total", "",
	    source_leap[SRC_LEAP_UNKNOWN], source_leap[SRC_LEAP_NO],
	    source_leap[SRC_LEAP_INSERT], source_leap[SRC_LEAP_DELETE]);

	/* Go with any qualified majority */
	ck_printf(CK_PASS, "System Leap state:");
	if (source_leap[SRC_LEAP_NO] >
	    (source_leap[SRC_LEAP_INSERT] + source_leap[SRC_LEAP_DELETE]))
		ck_printf(CK_PASS, " No Leap second today\n");
	else if (source_leap[SRC_LEAP_INSERT] >
	    (source_leap[SRC_LEAP_NO] + source_leap[SRC_LEAP_DELETE]))
		ck_printf(CK_PASS, " Insert Leap second today\n");
	else if (source_leap[SRC_LEAP_DELETE] >
	    (source_leap[SRC_LEAP_NO] + source_leap[SRC_LEAP_INSERT]))
		ck_printf(CK_PASS, " Delete Leap second today\n");
	else
		ck_printf(CK_PASS, " Unknown\n");
	return (0);
}

const char *
init_source(void)
{
	int i;

	ClockSelect(NULL);

	Lsource = LogNewHandle("source");

	i = evSetTimer(C, source_timer, &source_tid,
	     evConsTime(0, 0), evConsTime(1, 0), &source_tid);
	Assert(i == 0, "evSetTimer(source_timer)");
	i = evConfigTimer(C, source_tid, "rate", 0);
	Assert(i == 0, "evConfigTimer(source_timer)");
	sys_precision = -29;
	ntp_shmem = shmem_alloc(sizeof *ntp_shmem, NTP_TAG);
	ck_TreeAdd(cft, &config_source);
	ck_TreeAdd(cft, &config_leap);
	return (MAGIC);
}

