/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: timestamp.c,v 1.13 2005/09/09 20:41:55 phk Exp $
 *
 * Diddle timestamps
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <isc/eventlib.h>
#include <math.h>

#include "ntpns.h"

/*
 * Seconds between 1900 (NTP epoch) and 1970 (UNIX epoch).
 * 17 is the number of leapdays.
 */
#define NTP_UNIX	(((1970U - 1900U) * 365U + 17U) * 24U * 60U * 60U)

/*
 * Our best timestamp format is bintime, so we normalize to that.
 */
void
TsNormalize(struct ts *t)
{
	struct bintime bt;

	if (t->fmt == TS_BINTIME \
	    && t->ts.bt.sec == 0 && t->ts.bt.frac == 0)
		t->fmt = TS_INVALID;
	else if (t->fmt == TS_TIMEVAL \
	    && t->ts.tv.tv_sec == 0 && t->ts.tv.tv_usec == 0)
		t->fmt = TS_INVALID;
	else if (t->fmt == TS_TIMESPEC \
	    && t->ts.ts.tv_sec == 0 && t->ts.ts.tv_nsec == 0)
		t->fmt = TS_INVALID;
	else if (t->fmt == TS_NTPTIME \
	    && t->ts.ntp == 0)
		t->fmt = TS_INVALID;

	switch(t->fmt) {
	case TS_BINTIME:
		return;
	case TS_INVALID:
		memset(&t->ts, 0, sizeof t->ts);
		return;
	case TS_TIMEVAL:
		timeval2bintime(&t->ts.tv, &bt);
		break;
	case TS_TIMESPEC:
		timespec2bintime(&t->ts.ts, &bt);
		break;
	case TS_NTPTIME:
		bt.sec = (t->ts.ntp >> 32) - NTP_UNIX;
		bt.frac = t->ts.ntp << 32;
		break;
	default:
		fprintf(stderr, "timestamp format %d", t->fmt);
		abort();
	};
	t->fmt = TS_BINTIME;
	t->ts.bt = bt;
}

void
TsTimeval(struct ts *t, struct timeval *tv)
{

	if (t->fmt != TS_BINTIME)
		TsNormalize(t);
	bintime2timeval(&t->ts.bt, tv);
}

void
TsTimespec(struct ts *t, struct timespec *ts)
{

	if (t->fmt != TS_BINTIME)
		TsNormalize(t);
	bintime2timespec(&t->ts.bt, ts);
}

void
TsNTPtime(struct ts *t, uint64_t *nt)
{

	if (t->fmt == TS_INVALID) {
		*nt = 0;
	} else {
		TsNormalize(t);
		*nt = (uint64_t)(t->ts.bt.sec + NTP_UNIX) << 32;
		*nt |= t->ts.bt.frac >> 32;
	}
}


char *
TsFormat(struct ts *t, char *buf)
{
	struct timespec ts;

	if (buf == NULL)
		buf = malloc(50);
	if (buf != NULL) {
		if (t->fmt != TS_BINTIME)
			TsNormalize(t);
		if (t->fmt == TS_INVALID) {
			sprintf(buf, "%20s", "*invalid_timestamp*");
		} else {
			/* Nanosecond precision is OK for now */
			TsTimespec(t, &ts);
			sprintf(buf, "%10d.%09ld", ts.tv_sec, ts.tv_nsec);
		}
	}
	return (buf);
}

double
TsDiff(struct ts *t1, struct ts *t2)
{
	double a, b;

	if (t1->fmt != TS_BINTIME)
		TsNormalize(t1);
	if (t2->fmt != TS_BINTIME)
		TsNormalize(t2);

	if (t1->fmt == TS_INVALID || t2->fmt == TS_INVALID)
		return (NAN);

	a = t1->ts.bt.frac;
	a /= 18446744073709551616.0;		/* 2^64 */
	b = t2->ts.bt.frac;
	b /= 18446744073709551616.0;		/* 2^64 */
	a -= b;
	a += (t1->ts.bt.sec - t2->ts.bt.sec);
	return (a);
}

uint64_t
TsNTPNow(void)
{
	uint64_t	u;
	struct timespec tsp;

	clock_gettime(CLOCK_REALTIME, &tsp);
	u = (uint64_t)(tsp.tv_sec + NTP_UNIX) << 32LL;
	u |= ((uint64_t)tsp.tv_nsec << 32LL) / 1000000000;
	return (u);
}
