/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: watch.c,v 1.11 2007/05/08 07:36:09 phk Exp $
 *
 * Watchdog stuff.
 *
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/types.h>
#include <sys/watchdog.h>
#include "ntpns.h"
#include "ntpns_cfg.h"
#include "internal.h"
#include "pll.h"

static evTimerID watch_tid;

static int fd_utc_second = -1;
static char *fn_utc_second;
static int fd_running = -1;
static char *fn_running;
static int fd_no_source = -1;
static char *fn_no_source;
static int fd_no_lock = -1;
static char *fn_no_lock;
static int fd_time_sync = -1;
static char *fn_time_sync;

static double t_watchdog;
static u_int  bone;
static int fd_fido = -1;

static int
pat_fido(void)
{
	u_int b;
	int i;

	if (fd_fido < 0) 
		return (0);
	b = WD_ACTIVE | bone;
	i = ioctl(fd_fido, WDIOCPATPAT, &b);
	if (i)
		LogErr(Linfo, "fido barked at 0x%x", b);
	return  (i);
}

static void
watch_timer(evContext ctx __unused, void *uap __unused, struct timespec t0 __unused, struct timespec dt __unused)
{
	struct timespec t1;
	int i, j;

	/* Get the UTC SECOND led into state right away */
	if (fd_utc_second >= 0) 
		write(fd_utc_second, sys_able ? "0" : "1", 1);

	pat_fido();

	/* Aim the timer for the next second */
	t1 = evUTCTime();
	j = 1000000000 - t1.tv_nsec;
	if (j < 1000000)
		j += 1000000000;
	i = evResetTimer(ctx, watch_tid, watch_timer, uap,
		t0, evConsTime(0, j));
	Assert(i == 0, "evResetTimer(watch_timer, %d)", j);

	if (fd_utc_second >= 0) {
		/* Figure out how long time the UTC LED should be on */
		if (clk_state > UTC_STEP) {
			i = t1.tv_sec;
			if (t1.tv_nsec > 500000000)
				i++;
			if ((i % 60) == 0)
				write(fd_utc_second, "sDa.", 5);
			else if ((i % 60) == 59)
				write(fd_utc_second, "sAa.", 5);
			else if ((i % 60) == 58)
				write(fd_utc_second, "sAaAa.", 7);
			else if ((i % 60) == 57)
				write(fd_utc_second, "sAaAaAa.", 9);
			else
				write(fd_utc_second, "sAa.", 5);
		}
	}

	if (fd_running >= 0) {
		/*
		 * Give the RUN Led a poke long enough to keep it 
		 * lit until we come around again, taking into account
		 * possible steps.
		 */
		write(fd_running, "sJJj.", 5);
	}

	if (fd_no_source >= 0) {
		/* Give the source LED a poke */
		if (selsrc == NULL)
			write(fd_no_source, "1", 1);
		else if (sys_able & ABLE_SOURCE_OFFSET)
			write(fd_no_source, "sAjaA.", 6);
		else
			write(fd_no_source, "sjaA.", 5);
	}

	if (fd_time_sync >= 0) {
		/* Give the time LED a poke */
		if (sys_able != 0) {
			write(fd_time_sync, "0", 1);
		} else if (clk_state == UTC_LOCK) {
			write(fd_time_sync, "sJAa.", 5);
		} else if (clk_state == FRAC_LOCK) {
			write(fd_time_sync, "sJAa.", 5);
		} else if (clk_state == COAST) {
			write(fd_time_sync, "saJa.", 5);
		} else {
			write(fd_time_sync, "0", 1);
		}
	}

	if (fd_no_lock >= 0) {
		/* And the lock LED */
		if (pll_state.st >= 4 && clk_state >= UTC_LOCK)
			write(fd_no_lock, "sajA.", 5);
		else if (pll_state.st >= 4)
			write(fd_no_lock, "saA.", 4);
		else if (pll_state.st >= 3)
			write(fd_no_lock, "saAaA.", 6);
		else if (pll_state.st >= 2)
			write(fd_no_lock, "saAaAaA.", 8);
		else
			write(fd_no_lock, "1", 1);
	}
}

static int
do_cfg_led(CK_ARG, const char **word, char **fn, int *fd)
{

	if (!ck_WordVar(CK_PASS, word, fn))
		return (0);
	if (*fd >= 0)
		close(*fd);
	*fd = -1;
	if (*fn == NULL)
		return (0);
	*fd = open(*fn, O_RDWR);
	if (*fd < 0)
		ck_printf(CK_PASS, "Could not open led (%s): %s\n",
		    *fn, strerror(errno));
	return (0);
}

int
cfg_led_running(CK_ARG, const char **word)
{

	return (do_cfg_led(CK_PASS, word, &fn_running, &fd_running));
}

int
cfg_led_no_source(CK_ARG, const char **word)
{

	return (do_cfg_led(CK_PASS, word, &fn_no_source, &fd_no_source));
}

int
cfg_led_no_lock(CK_ARG, const char **word)
{

	return (do_cfg_led(CK_PASS, word, &fn_no_lock, &fd_no_lock));
}

int
cfg_led_time_sync(CK_ARG, const char **word)
{

	return (do_cfg_led(CK_PASS, word, &fn_time_sync, &fd_time_sync));
}

int
cfg_led_utc_second(CK_ARG, const char **word)
{

	return (do_cfg_led(CK_PASS, word, &fn_utc_second, &fd_utc_second));
}

int
cfg_watchdog(CK_ARG, double *real)
{
	u_int new_bone;
	double tout;

	tout = t_watchdog;
	if (!ck_RealVar(CK_PASS, real, &tout, 0.0, "%.0f"))
		return (0);

	if (dopt)
		return (0);

	if (tout < 2.0) {
		if (fd_fido >= 0) {
			bone = 0;
			pat_fido();
			close(fd_fido);
			fd_fido = -1;
		}
		t_watchdog = 0.0;
		return (0);
	}

	if (fd_fido < 0) {
		fd_fido = open("/dev/" _PATH_WATCHDOG, O_RDWR);
		if (fd_fido < 0) {
			ck_printf(CK_PASS,
			    "Can not open watchdog device: %s\n",
			    strerror(errno));
			return (0);
		}
		bone = 0;
	}

	new_bone = ceil(log(tout * 1e9) / log(2.0));
	if (new_bone > WD_INTERVAL) {
		ck_printf(CK_PASS,
		    "watchdog does not support timeout this long\n");
		return (0);
	}
	new_bone |= WD_ACTIVE;
	if (ioctl(fd_fido, WDIOCPATPAT, &new_bone)) {
		if (errno == EOPNOTSUPP)
			ck_printf(CK_PASS,
			    "watchdog does not support this timeout\n");
		else
			ck_printf(CK_PASS,
			    "Fido barked at 0x%x: %s\n",
			    new_bone, strerror(errno));
		pat_fido();
		return (0);
	} else {
		bone = new_bone;
		t_watchdog = tout;
	}
	return (0);
}

const char *
init_watch(void)
{
	int i;

	ck_TreeAdd(cft, &config_led);
	ck_TreeAdd(cft, &config_watchdog);

	i = evSetTimer(C, watch_timer, &watch_tid,
		evConsTime(0, 0), evConsTime(1, 0), &watch_tid);
	Assert(i == 0, "evSetTimer(watch_timer)");
	i = evConfigTimer(C, watch_tid, "rate", 0);
	Assert(i == 0, "evConfigTimer(watch_timer)");
	return (MAGIC);
}
