/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: ntp_mib.c,v 1.32 2005/08/21 16:45:17 phk Exp $
 *
 * BSNMP module for NTP related mibs.
 *
 */

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <err.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <bsnmp/asn1.h>
#include <math.h>
#include <configkit.h>
#include <bsnmp/snmp.h>
#include <bsnmp/snmpmod.h>
#include <isc/eventlib.h>

#include "ntp_tree.h"
#include "ntp_oid.h"
#include "shmem.h"
#include "ntpns.h"
#include "source.h"
#include <sys/endian.h>
#include "../ntpv4/ntpv4.h"
#include "oids.h"

/*
 * Parameters settable with SNMP
 */
static int traps_per_timer = 0;			/* count */
static int timer_interval = 10;			/* seconds */
static double partner_threshold_offset = .100;	/* seconds */
static int partner_threshold_los = 1800;	/* half an hour */
static int shmem_update = 10;			/* shmem cache timeout */

static int shmem_fd = -1;
static struct stat shmem_stat;
static void *shmem_map;
static struct ntpns_id *shmem_id;
static struct ntpns_shmem *ntpns_shmem;

static u_int trap_seq;
static struct lmodule *lmmod;
static int o_clockstate, o_clocksource;


/*
 * Increment the trap-count and return a snmp_value for it.
 */
static struct snmp_value *
trap_seq_oid(void)
{
	static struct snmp_value trapseq;

	trapseq.var = oid_ntpNsSystemTrapSeq;
	trapseq.syntax = SNMP_SYNTAX_GAUGE;
	trapseq.v.integer = ++trap_seq;

	return(&trapseq);
}

static void
DoNano(struct snmp_value *v, double a)
{
	if (a > 2.0)
		a = 2.0;
	if (a < -2.0)
		a = -2.0;
	v->v.integer = (int)(1e9 * a);
}

#define ALARM_LOS	1
#define ALARM_OFFSET	2
#define ALARM_STATE	4

struct mypartner {
	u_int			generation;
	int			idx;
	int			alarms;
	TAILQ_ENTRY(mypartner)	iplist;
};
TAILQ_HEAD(, mypartner) iplist = TAILQ_HEAD_INITIALIZER(iplist);

struct mysource {
	struct source		*src;
	u_int			index;
	int			found;
	src_state_e		last_state;
	int			alarms;
	TAILQ_ENTRY(mysource)	link;
};

static TAILQ_HEAD(, mysource) srclist = TAILQ_HEAD_INITIALIZER(srclist);

static struct partner *partners;
static struct mypartner *mypartners;
static int npartners, maxpartner = 10;
static u_int npartner, npartneralarm;

static void
UpdPartner(int idx)
{
	struct partner *pp;
	struct mypartner *mp, *mp2;

	pp = &partners[idx];
	mp = &mypartners[idx];

	if (mp->generation != 0) {
		TAILQ_REMOVE(&iplist, mp, iplist);
		mp->generation = 0;
	}
	if (pp->ip != 0) {
		mp->generation = pp->generation;
		TAILQ_FOREACH(mp2, &iplist, iplist) {
			if (partners[mp2->idx].ip > pp->ip) {
				TAILQ_INSERT_BEFORE(mp2, mp, iplist);
				return;
			}
		}
		TAILQ_INSERT_TAIL(&iplist, mp, iplist);
		return;
	}
}

static void
UpdatePartners(void)
{
	int i;

	if (mypartners == NULL) {
		mypartners = calloc(sizeof *mypartners, npartners);
		if (mypartners == NULL)
			err(1, "malloc");
	}
	for (i = 0; i < npartners; i++) {
		if (partners[i].flag & P_F_NEVER) {
			maxpartner = i;
			break;
		}
		mypartners[i].idx = i;
		if (partners[i].generation != mypartners[i].generation)
			UpdPartner(i);
	}
}

static void
AddSource(struct source *s)
{
	struct mysource *msrc, *ms2;

	/*
	 * If we already had it, just mark it and return.
	 */
	TAILQ_FOREACH(ms2, &srclist, link) {
		if (ms2->src->idx > s->idx)
			break;
		if (ms2->src == s) {
			ms2->found = 1;
			return;
		}
	}
	/*
	 * Add new source.
 	 */
	msrc = calloc(1, sizeof *msrc);
	if (msrc == NULL)
		err(1, "malloc failed");
	msrc->index = s->idx;
	msrc->src = s;
	msrc->found = 1;
	msrc->last_state = s->state;
	TAILQ_FOREACH(ms2, &srclist, link) {
		if (ms2->src->idx > msrc->src->idx) {
			TAILQ_INSERT_BEFORE(ms2, msrc, link);
			return;
		}
	}
	TAILQ_INSERT_TAIL(&srclist, msrc, link);
}

static void
WalkShmem(void)
{
	struct shmem *sp;
	u_char *shp;
	struct mysource *msrc, *m2;

	shmem_id = NULL;
	shp = shmem_map;
	ntpns_shmem = NULL;
	partners = NULL;

	/* Mark all sources as "not found" */
	TAILQ_FOREACH(msrc, &srclist, link)
		msrc->found = 0;
	while (shp < (u_char *)shmem_map + shmem_stat.st_size) {
		sp = (void*)shp;
		if (sp->tag == 0)
			break;
		if (sp->tag == ID_TAG)
			shmem_id = (void *)(shp + sizeof *sp);
		if (sp->tag == TAG_PARTNER) {
			partners = (void *)(shp + sizeof *sp);
			npartners = sp->length / sizeof (*partners);
			UpdatePartners();
		}
		if (sp->tag == NTP_TAG)
			ntpns_shmem = (void *)(shp + sizeof *sp);
		if (sp->tag == SOURCE_TAG)  {
			AddSource((void *)(shp + sizeof *sp));
		}
		shp += sp->length;
		if (sp->length == 0)
			break;
	}
	/* Delete any sources we did not find */
	TAILQ_FOREACH_SAFE(msrc, &srclist, link, m2) {
		if (msrc->found == 0) {
			TAILQ_REMOVE(&srclist, msrc, link);
			free(msrc);
		}
	}
}

static int
CheckShmem(void)
{
	struct stat s1;
	static time_t t1, t2;

	/* Check timeout before doing anything */

	time(&t2);
	if (t2 < t1 + shmem_update) 
		return (0);
	t1 = t2;
	

	/* If we cannot stat the shmem file, bail */
	if (stat(SHMEM_FILENAME, &s1) < 0) {
		warn("stat(SHMEM_FILENAME");
		if (shmem_fd >= 0)
			close(shmem_fd);
		shmem_fd = -1;
		return (-1);
	}

	/* Check that the shmem file have not changed under us */
	if (shmem_fd >= 0) {
		if (shmem_stat.st_dev != s1.st_dev)
			warnx("changed SHMEM_FILENAME (dev)");
		if (shmem_stat.st_ino != s1.st_ino)
			warnx("changed SHMEM_FILENAME (ino)");
		if (shmem_stat.st_size != s1.st_size)
			warnx("changed SHMEM_FILENAME (size)");
		if (shmem_stat.st_dev != s1.st_dev ||
		    shmem_stat.st_ino != s1.st_ino ||
		    shmem_stat.st_size != s1.st_size) {
			close(shmem_fd);
			shmem_fd = -1;
		}
	}

	/* If necessary, open and map file */
	if (shmem_fd < 0) {
		shmem_fd = open(SHMEM_FILENAME, O_RDONLY);
		if (shmem_fd < 0) {
			warn("open SHMEM_FILENAME");
			return (-1);
		}
		fstat(shmem_fd, &shmem_stat);
		/*
		 * Once we have a shmem_map, we stick to it and map new
		 * instances of our file the same place.
		 */
		shmem_map = mmap(shmem_map, shmem_stat.st_size, PROT_READ,
		    MAP_PRIVATE, shmem_fd, 0);
		if (shmem_map == MAP_FAILED) {
			warn("mmap SHMEM_FILENAME");
			close(shmem_fd);
			shmem_fd = -1;
			return (-1);
		}

		/* Look for ID block */
		WalkShmem();
		time(&t1);
		if (shmem_id == NULL || strcmp(shmem_id->magic, MAGIC)) {
			warn("no (valid) ID in shmem");
			close(shmem_fd);
			shmem_fd = -1;
			return (-1);
		}
	}

	/* Check that NTPns is still running */
	if (kill(shmem_id->pid, 0)) {
		warnx("could not ping pid=%d with signal", shmem_id->pid);
		close(shmem_fd);
		shmem_fd = -1;
		return (-1);
	}

	WalkShmem();

	/* Be happy */
	return (0);
}

static void
ntp_timer(void *arg __unused)
{
	struct mysource *msrc;
	struct source *src;
	struct mypartner *mpp;
	struct partner *pp;
	int alarms, ip[4], i, j, ntraps;
	time_t t;
	double d;
	static struct snmp_value trapoid[5];
	const struct asn_oid *trap;

	ntraps = 0;


	CheckShmem();

	if (o_clockstate != ntpns_shmem->clockstate ||
	    o_clocksource != ntpns_shmem->clocksource) {
		trapoid[0].var = oid_ntpNsSystemClockState;
		trapoid[0].syntax = SNMP_SYNTAX_INTEGER;
		trapoid[0].v.integer = ntpns_shmem->clockstate;

		trapoid[1].var = oid_ntpNsSystemClockSource;
		trapoid[1].syntax = SNMP_SYNTAX_INTEGER;
		trapoid[1].v.integer = ntpns_shmem->clocksource;

		snmp_send_trap(&oid_ntpNsSystemSyncTrap,
		    trap_seq_oid(),
		    &trapoid[0],
		    &trapoid[1],
		    NULL);
		o_clockstate = ntpns_shmem->clockstate;
		o_clocksource = ntpns_shmem->clocksource;
	}


	TAILQ_FOREACH(msrc, &srclist, link) {
		alarms = 0;
		src = msrc->src;
		if (src->los > src->los_timeout)
			alarms |= ALARM_LOS;
		if (src->state != msrc->last_state) {
			alarms |= ALARM_STATE;
			msrc->last_state = src->state;
		}
		if (src->limit > 0 && fabs(src->last_delta) > src->limit)
			alarms |= ALARM_OFFSET;
		if (msrc->alarms == alarms)
			continue;
		if ((msrc->alarms ^ alarms) & ALARM_LOS)
			trap = &oid_ntpNsSourceLOSTrap;
		else if ((msrc->alarms ^ alarms) & ALARM_STATE)
			trap = &oid_ntpNsSourceStateTrap;
		else
			trap = &oid_ntpNsSourceOffsetTrap;

		trapoid[0].var = oid_ntpNsSourceIndex;
		trapoid[0].syntax = SNMP_SYNTAX_GAUGE;
		trapoid[0].v.integer = src->idx;

		trapoid[1].var = oid_ntpNsSourceName;
		trapoid[1].syntax = SNMP_SYNTAX_OCTETSTRING;
		trapoid[1].v.octetstring.octets = src->name;
		trapoid[1].v.octetstring.len = strlen(src->name);

		trapoid[2].var = oid_ntpNsSourceState;
		trapoid[2].syntax = SNMP_SYNTAX_INTEGER;
		trapoid[2].v.integer = src->state;

		trapoid[3].var = oid_ntpNsSourceLastDelta;
		trapoid[3].syntax = SNMP_SYNTAX_INTEGER;
		DoNano(&trapoid[3],  src->last_delta);

		trapoid[4].var = oid_ntpNsSourceLastSample;
		trapoid[4].syntax = SNMP_SYNTAX_GAUGE;
		time(&t);
		trapoid[4].v.integer = t - msrc->src->last_ts.ts.bt.sec;

		/* fix up the index */
		for (j = 0; j < 5; j++)
			trapoid[j].var.subs[trapoid[j].var.len++] = src->idx;

		snmp_send_trap(trap, trap_seq_oid(),
		    &trapoid[0],
		    &trapoid[1],
		    &trapoid[2],
		    &trapoid[3],
		    &trapoid[4],
		    NULL);

		msrc->alarms = alarms;
	}

	time(&t);
	npartner = 0;
	npartneralarm = 0;
	TAILQ_FOREACH(mpp, &iplist, iplist) {
		alarms = 0;
		npartner++;
		pp = &partners[mpp->idx];
		d = pp->delta;
		if (fabs(d) > partner_threshold_offset)
			alarms |= ALARM_OFFSET;
		if (pp->last + partner_threshold_los < t)
			alarms |= ALARM_LOS;
		if (alarms)
			npartneralarm++;
		if (mpp->alarms == alarms)
			continue;
		if (++ntraps > traps_per_timer)
			continue;

		trapoid[0].var = oid_ntpNsPartnerSrcIP;
		trapoid[0].syntax = SNMP_SYNTAX_IPADDRESS;
		trapoid[0].v.integer = ntohl(pp->ip);

		trapoid[1].var = oid_ntpNsPartnerSrcPort;
		trapoid[1].syntax = SNMP_SYNTAX_INTEGER;
		trapoid[1].v.integer = ntohs(pp->addr.sin_port);

		trapoid[2].var = oid_ntpNsPartnerOffset;
		trapoid[2].syntax = SNMP_SYNTAX_INTEGER;
		trapoid[2].v.integer = d * 1e9;

		trapoid[3].var = oid_ntpNsPartnerLastPacket;
		trapoid[3].v.integer = t - pp->last;
		trapoid[3].syntax = SNMP_SYNTAX_INTEGER;

		if ((mpp->alarms ^ alarms) & ALARM_LOS)
			trap = &oid_ntpNsPartnerLOSTrap;
		else
			trap = &oid_ntpNsPartnerOffsetTrap;

		/* plug in index */
		ip[0] = (pp->ip >> 24) & 0xff;
		ip[1] = (pp->ip >> 16) & 0xff;
		ip[2] = (pp->ip >> 8) & 0xff;
		ip[3] = (pp->ip >> 0) & 0xff;
		for (i = 0; i < 4; i++)
			for (j = 0; j < 4; j++)
				trapoid[j].var.subs[trapoid[j].var.len++] =
				     ip[i];

		snmp_send_trap(trap, trap_seq_oid(),
		    &trapoid[0],
		    &trapoid[1],
		    &trapoid[2],
		    &trapoid[3],
		    NULL);

		mpp->alarms = alarms;
	}

	timer_start(timer_interval * 100, ntp_timer, NULL, lmmod);
}

static int
ntp_init(struct lmodule *mod, int argc __unused, char *argv[] __unused)
{
	lmmod = mod;
	return (0);
}

static void
ntp_start(void)
{

	snmp_send_trap(&oid_ntpNsSystemRestartTrap, trap_seq_oid(), NULL);
	ntp_timer(NULL);
}

const struct snmp_module config = {
	.comment = 	"This module implements the NTP MIB(s).",
	.init = 	ntp_init,
	.start =	ntp_start,
	.tree =		ntp_ctree,
	.tree_size =	ntp_CTREE_SIZE,
};

static int
DoTS(struct snmp_value *v, struct ts *ts)
{
	uint64_t ntpts;
	struct ts t;

	/*
	 * The timestamp may need normalizing and we cannot do that if it
	 * resides in R/O shmem.
	 */
	t = *ts;
	TsNTPtime(&t, &ntpts);
	be64enc(&ntpts, ntpts);
	return (string_get(v, (void*)&ntpts, 8));
}

static int
fooM(int n, struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{

	u_int u;
	const char *sep;
	int32_t idx;
	int i;

	fprintf(stderr, "foo%02d c %p v %p sub %d iidx %d o %x {", n, c, v, sub, iidx, o);
	
	sep = "";
	for (u = 0; u < sub; u++) {
		fprintf(stderr, "%s%d", sep, v->var.subs[u]);
		sep = ".";
	}
	if (o == SNMP_OP_GETNEXT) {
		idx = 0;
		i = index_decode(&v->var, sub, iidx, &idx);
		fprintf(stderr, "[%d (%d)]", idx, i);
	}
	fprintf(stderr, "}\n");

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	return (0);
}

static int
fooN(int n, struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{

	fooM(n, c, v, sub, iidx, o);
	return (SNMP_ERR_NOSUCHNAME);
}

int
foo00(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx __unused, enum snmp_op o)
{

	if (o == SNMP_OP_GETNEXT || o == SNMP_OP_GET) {
	} else if (o == SNMP_OP_GET) {
	} else {
		return (SNMP_ERR_NOT_WRITEABLE);
	}
	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (ntpns_shmem == NULL)
		return (SNMP_ERR_NOSUCHNAME);

	return (SNMP_ERR_NOSUCHNAME);

	switch (v->var.subs[sub - 1]) {
	case LEAF_cntpSysLeap:
		v->v.integer = ntpns_shmem->ntpSysLeap;
		break;
	case LEAF_cntpSysStratum:
		v->v.integer = ntpns_shmem->ntpSysStratum;
		break;
	case LEAF_cntpSysPrecision:
		v->v.integer = ntpns_shmem->ntpSysPrecision;
		break;
	case LEAF_cntpSysRootDelay:
		v->v.integer = ntpns_shmem->ntpSysRootDelay;
		break;
	case LEAF_cntpSysRootDispersion:
		v->v.integer = ntpns_shmem->ntpSysRootDispersion;
		break;
	case LEAF_cntpSysRefId:
		string_get(v, ntpns_shmem->ntpRefId, 4);
		break;
	case LEAF_cntpSysRefTime:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpSysPoll:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpSysPeer:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpSysClock:
		return (SNMP_ERR_NOSUCHNAME);
	default:
		return (SNMP_ERR_NOSUCHNAME);
	}
	return (SNMP_ERR_NOERROR);
}

int
foo01(struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{
	int32_t idx;
	int i;
	struct asn_oid ao;
	struct partner *pp;
	u_char buf[32];

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (partners == NULL)
		return (SNMP_ERR_NOSUCHNAME);
	if (o == SNMP_OP_GETNEXT) {
		idx = 0;
		i = index_decode(&v->var, sub, iidx, &idx);
		if (idx >= maxpartner) 
			return (SNMP_ERR_NOSUCHNAME);
		ao.len = 1;
		ao.subs[0] = idx + 1;
		index_append(&v->var, sub, &ao);
	} else if (o == SNMP_OP_GET) {
		idx = 0;
		i = index_decode(&v->var, sub, iidx, &idx);
		if (idx >= maxpartner) 
			return (SNMP_ERR_NOSUCHNAME);
	} else {
		return (SNMP_ERR_NOT_WRITEABLE);
	}
	pp = &partners[idx];
	switch (v->var.subs[sub - 1]) {
	case LEAF_cntpPeersAssocId:
		v->v.integer = idx + 1;
		break;
	case LEAF_cntpPeersConfigured:
		v->v.integer = pp->configured == NULL ? 2 : 1;
		break;
	case LEAF_cntpPeersPeerAddress:
		be32enc(v->v.ipaddress, pp->ip);
		break;
	case LEAF_cntpPeersPeerPort:
		v->v.integer = ntohs(pp->addr.sin_port);
		break;
	case LEAF_cntpPeersHostAddress:
	case LEAF_cntpPeersHostPort:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersLeap:
		v->v.integer = pp->packet.leap;
		break;
	case LEAF_cntpPeersMode:
		v->v.integer = pp->packet.mode;
		break;
	case LEAF_cntpPeersStratum:
		v->v.integer = pp->packet.stratum;
		break;
	case LEAF_cntpPeersPeerPoll:
		v->v.integer = pp->packet.poll;
		break;
	case LEAF_cntpPeersHostPoll:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersPrecision:
		v->v.integer = pp->packet.prec;
		break;
	case LEAF_cntpPeersRootDelay:
		be32enc(buf, (int)(pp->packet.delay * 65536.0));
		string_get(v, buf, 4);
		break;
	case LEAF_cntpPeersRootDispersion:
		be32enc(buf, (unsigned)(pp->packet.disp * 65536.0));
		string_get(v, buf, 4);
		break;
	case LEAF_cntpPeersRefId:
		string_get(v, pp->packet.refid, 4);
		break;
	case LEAF_cntpPeersRefTime:
		DoTS(v, &pp->packet.ts[0]);
		break;
	case LEAF_cntpPeersOrgTime:
		DoTS(v, &pp->packet.ts[1]);
		break;
	case LEAF_cntpPeersReceiveTime:
		DoTS(v, &pp->packet.ts[4]);
		break;
	case LEAF_cntpPeersTransmitTime:
		DoTS(v, &pp->packet.ts[4]);	/* XXX: Not quite */
		break;
	case LEAF_cntpPeersUpdateTime:
	case LEAF_cntpPeersUpdateTimeRev1:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersReach:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersTimer:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersOffset:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersDelay:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersDispersion:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersFilterValidEntries:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersEntryStatus:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_cntpPeersPrefPeer:
		return (SNMP_ERR_NOSUCHNAME);
	default:
		i = fooN(1, c, v, sub, iidx, o);
		return (i);
	}
	return (SNMP_ERR_NOERROR);
}

int
foo02(struct snmp_context *c, struct snmp_value *v, u_int x, u_int y, enum snmp_op o)
{

	/*
	 * XXX: Filter Group
	 */
	return (SNMP_ERR_NOSUCHNAME);
	return(fooN(2, c, v, x, y, o));
}

/*
 * XXX: Udel MIB.  Not yet implemented.
 */
int
foo10(struct snmp_context *c, struct snmp_value *v, u_int x, u_int y, enum snmp_op o)
{

	return (SNMP_ERR_NOSUCHNAME);
	return(fooN(10, c, v, x, y, o));
}

int
foo11(struct snmp_context *c, struct snmp_value *v, u_int x, u_int y, enum snmp_op o)
{

	return (SNMP_ERR_NOSUCHNAME);
	return(fooN(11, c, v, x, y, o));
}

int
foo12(struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{

	return (SNMP_ERR_NOSUCHNAME);
	return(fooN(12, c, v, sub, iidx, o));
}


int
foo13(struct snmp_context *c, struct snmp_value *v, u_int x, u_int y, enum snmp_op o)
{

	return (SNMP_ERR_NOSUCHNAME);
	return(fooN(13, c, v, x, y, o));
}

int
NTPnsSystem(struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (o != SNMP_OP_GET)
		return (SNMP_ERR_NOT_WRITEABLE);
	switch (v->var.subs[sub - 1]) {
	case LEAF_ntpNsSystemMajor:
		v->v.integer = shmem_id->major;
		break;
	case LEAF_ntpNsSystemMinor:
		v->v.integer = shmem_id->minor;
		break;
	case LEAF_ntpNsSystemMagic:
		return (string_get(v, (void*)shmem_id->magic, -1));
	case LEAF_ntpNsSystemTrapSeq:
		v->v.integer = trap_seq;
		break;
	case LEAF_ntpNsSystemClockState:
		v->v.integer = ntpns_shmem->clockstate;
		break;
	case LEAF_ntpNsSystemClockSource:
		v->v.integer = ntpns_shmem->clocksource;
		break;
	default:
		return(fooN(99, c, v, sub, iidx, o));
	}
	return (SNMP_ERR_NOERROR);
}


int
NTPnsSourceTable(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx __unused, enum snmp_op o)
{
	struct mysource *msrc;
	time_t t;

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	switch(o) {
	case SNMP_OP_GETNEXT:
		msrc = NEXT_OBJECT_INT(&srclist, &v->var, sub);
		if (msrc == NULL)
			return (SNMP_ERR_NOSUCHNAME);
		v->var.len = sub + 1;
		v->var.subs[sub] = msrc->index;
		break;
	case SNMP_OP_GET:
		if (v->var.len - sub != 1)
			return (SNMP_ERR_NOSUCHNAME);
		msrc = FIND_OBJECT_INT(&srclist, &v->var, sub);
		if (msrc == NULL)
			return (SNMP_ERR_NOSUCHNAME);
		break;
	default:
		return (SNMP_ERR_NOT_WRITEABLE);
	}
	switch (v->var.subs[sub - 1]) {
	case LEAF_ntpNsSourceIndex:
		v->v.integer = msrc->index;
		break;
	case LEAF_ntpNsSourceName:
		return (string_get(v, (void*)msrc->src->name, -1));
	case LEAF_ntpNsSourceState:
		v->v.integer = msrc->src->state;
		break;
	case LEAF_ntpNsSourceUpdateRate:
		v->v.integer = msrc->src->update_rate;
		break;
	case LEAF_ntpNsSourceLosTimeout:
		v->v.integer = msrc->src->los_timeout;
		break;
	case LEAF_ntpNsSourceVotes:
		v->v.integer = msrc->src->votes;
		break;
	case LEAF_ntpNsSourceLastSample:
		time(&t);
		v->v.integer = t - msrc->src->last_ts.ts.bt.sec;
		break;
	case LEAF_ntpNsSourceLastDelta:
		DoNano(v,  msrc->src->last_delta);
		break;
	case LEAF_ntpNsSourceLimit:
		DoNano(v,  msrc->src->limit);
		break;
	default:
		fooM(23, c, v, sub, iidx, o);
		return (SNMP_ERR_NOSUCHNAME);
	}
	return (SNMP_ERR_NOERROR);
}

int
NTPnsSourceStatTable(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{
	return (SNMP_ERR_NOSUCHNAME);
	return(fooN(13, c, v, sub, iidx, o));
}

int
NTPnsPartner(struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (o != SNMP_OP_GET)
		return (SNMP_ERR_NOT_WRITEABLE);
	switch (v->var.subs[sub - 1]) {
	case LEAF_ntpNsPartnerCount:
		v->v.integer = npartner;
		break;
	case LEAF_ntpNsPartnerMaxCount:
		v->v.integer = npartners;
		break;
	case LEAF_ntpNsPartnerAlarm:
		v->v.integer = npartneralarm;
		break;
	default:
		return(fooN(99, c, v, sub, iidx, o));
	}
	return (SNMP_ERR_NOERROR);
}

int
NTPnsPartnerTable(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{
	uint32_t idx;
	int i;
	struct asn_oid ao;
	u_char buf[32];
	struct mypartner *mp;
	struct partner *pp;
	struct ts ts1, ts2, ts3, ts4;
	double a;
	time_t t;

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (partners == NULL)
		return (SNMP_ERR_NOSUCHNAME);
	if (o == SNMP_OP_GETNEXT || o == SNMP_OP_GET) {
		i = index_decode(&v->var, sub, iidx, buf);
		if (i && o == SNMP_OP_GET)
			return (SNMP_ERR_NOSUCHNAME);
		if (i) {
			mp = TAILQ_FIRST(&iplist);
			pp = &partners[mp->idx];
		} else {
			idx = be32dec(buf);
			TAILQ_FOREACH(mp, &iplist, iplist) {
				pp = &partners[mp->idx];
				if (o == SNMP_OP_GET && pp->ip == idx)
					break;
				if (o == SNMP_OP_GETNEXT && pp->ip > idx)
					break;
			}
			if (mp == NULL)
				return (SNMP_ERR_NOSUCHNAME);
		}
		if (o == SNMP_OP_GETNEXT) {
			be32enc(buf, partners[mp->idx].ip);
			ao.len = 4;
			ao.subs[0] = buf[0];
			ao.subs[1] = buf[1];
			ao.subs[2] = buf[2];
			ao.subs[3] = buf[3];
			index_append(&v->var, sub, &ao);
		}
	} else {
		return (SNMP_ERR_NOT_WRITEABLE);
	}
	time(&t);
	switch (v->var.subs[sub - 1]) {
	case LEAF_ntpNsPartnerSrcIP:
		be32enc(v->v.ipaddress, pp->ip);
		break;
	case LEAF_ntpNsPartnerSrcPort:
		v->v.integer = ntohs(pp->addr.sin_port);
		break;
	case LEAF_ntpNsPartnerDstIP:
	case LEAF_ntpNsPartnerDstPort:
		return (SNMP_ERR_NOSUCHNAME);
	case LEAF_ntpNsPartnerOffset:
	case LEAF_ntpNsPartnerAbsOffset:
		ts1 = pp->packet.ts[1];
		ts2 = pp->packet.ts[2];
		ts3 = pp->packet.ts[3];
		ts4 = pp->packet.ts[4];
		TsNormalize(&ts1);
		TsNormalize(&ts2);
		TsNormalize(&ts3);
		TsNormalize(&ts4);
		if (ts1.fmt != TS_INVALID && ts2.fmt != TS_INVALID && 
		    ts3.fmt != TS_INVALID && ts4.fmt != TS_INVALID)
			a = .5 * (TsDiff(&ts2, &ts1) + TsDiff(&ts3, &ts4));
		else if (ts3.fmt != TS_INVALID && ts4.fmt != TS_INVALID)
			a = TsDiff(&ts3, &ts4);
		else
			return (SNMP_ERR_NOSUCHNAME);
		if (v->var.subs[sub - 1] == LEAF_ntpNsPartnerAbsOffset)
			a = fabs(a);
		DoNano(v, a);
		break;
	case LEAF_ntpNsPartnerPeerAssocId:
		v->v.integer = mp->idx + 1;
		break;
	case LEAF_ntpNsPartnerLastPacket:
		v->v.integer = t - pp->last;
		break;
	default:
		return(fooN(98, c, v, sub, iidx, o));
	}
	return (SNMP_ERR_NOERROR);
}

int
NTPnsTest(struct snmp_context *c, struct snmp_value *v, u_int sub, u_int iidx, enum snmp_op o)
{

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (o != SNMP_OP_GET)
		return (SNMP_ERR_NOT_WRITEABLE);
	switch (v->var.subs[sub - 1]) {
	case LEAF_ntpNsTest1ns:
		v->v.integer = 1;
		break;
	case LEAF_ntpNsTest1us:
		v->v.integer = 1000;
		break;
	case LEAF_ntpNsTest1ms:
		v->v.integer = 1000000;
		break;
	case LEAF_ntpNsTest1s:
		v->v.integer = 1000000000;
		break;
	case LEAF_ntpNsTestn1ns:
		v->v.integer = -1;
		break;
	case LEAF_ntpNsTestn1us:
		v->v.integer = -1000;
		break;
	case LEAF_ntpNsTestn1ms:
		v->v.integer = -1000000;
		break;
	case LEAF_ntpNsTestn1s:
		v->v.integer = -1000000000;
		break;
	default:
		return(fooN(99, c, v, sub, iidx, o));
	}
	return (SNMP_ERR_NOERROR);
}

int
NTPnsConfig(struct snmp_context *c __unused, struct snmp_value *v, u_int sub, u_int iidx __unused, enum snmp_op o)
{

	if (CheckShmem())
		return (SNMP_ERR_NOSUCHNAME);
	if (o == SNMP_OP_SET) {
		switch (v->var.subs[sub - 1]) {
			case LEAF_ntpNsConfigTimerPeriod:
				if (v->v.integer < 1)
					return (SNMP_ERR_BADVALUE);
				timer_interval = v->v.integer;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigTrapsPerRun:
				if (v->v.integer < 0)
					return (SNMP_ERR_BADVALUE);
				traps_per_timer = v->v.integer;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigPartnerThresholdLOS:
				if (v->v.integer < 0)
					return (SNMP_ERR_BADVALUE);
				partner_threshold_los = v->v.integer;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigPartnerThresholdOffset:
				if (v->v.integer < 0)
					return (SNMP_ERR_BADVALUE);
				partner_threshold_offset = v->v.integer / 1e9;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigCacheShmem:
				if (v->v.integer < 0)
					return (SNMP_ERR_BADVALUE);
				shmem_update = v->v.integer;
				return (SNMP_ERR_NOERROR);
			default:
				return (SNMP_ERR_NOT_WRITEABLE);
		}
	} else if (o == SNMP_OP_GET) {
		switch (v->var.subs[sub - 1]) {
			case LEAF_ntpNsConfigTimerPeriod:
				v->v.integer = timer_interval;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigTrapsPerRun:
				v->v.integer = traps_per_timer;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigPartnerThresholdLOS:
				v->v.integer = partner_threshold_los;
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigPartnerThresholdOffset:
				DoNano(v, partner_threshold_offset);
				return (SNMP_ERR_NOERROR);
			case LEAF_ntpNsConfigCacheShmem:
				v->v.integer = shmem_update;
				return (SNMP_ERR_NOERROR);
			default:
				return (SNMP_ERR_NOSUCHNAME);
		}
	} else {
		return (SNMP_ERR_GENERR);
	}
}
