/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: packet.c,v 1.24 2007/01/18 09:49:43 phk Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/endian.h>
#include <netinet/in.h>
#include <netdb.h>

#include <ntpns.h>
#include <ntpv4.h>
#include <ntpv4_cfg.h>

struct msghdr *
socket_prepare(struct partner *p, void *buf, u_int len)
{
	u_char *ap;
	struct msghdr *msg;
	struct iovec *iov;

	ap = calloc(sizeof *msg + sizeof *iov + p->cookielen, 1);
	if (ap == NULL)
		return ((void *)ap);
	msg = (void *)ap; ap += sizeof *msg;
	msg->msg_name = &p->addr;
	msg->msg_namelen = sizeof(p->addr);
	iov = (void *)ap; ap += sizeof *iov;
	iov->iov_base = buf;
	iov->iov_len = len;
	msg->msg_iov = iov;
	msg->msg_iovlen = 1;
	if (p->cookie != NULL) {
		memcpy(ap, p->cookie, p->cookielen);
		msg->msg_control = ap;
		msg->msg_controllen = p->cookielen;
	}
	return (msg);
}

static void
socket_input(evContext ctx __unused, void *uap, int fd __unused, int eventmask __unused)
{
	struct instance *ip;
	int len;
	u_char cbuf[BUFSIZ];
	u_char ctrl[BUFSIZ];
	struct sockaddr_in sind;
	struct msghdr msg;
	struct iovec iov;
	struct cmsghdr *cmsg;
	struct ts ts;
	void *cookie;
	int cookielen;
	in_addr_t rip;
	u_char rref[4];

	ip = uap;
	for (;;) {
		ts.fmt = TS_INVALID;
		if (ip->userland_ts) {
#if defined(CLOCK_REALTIME)
			ts.fmt = TS_TIMESPEC;
			clock_gettime(CLOCK_REALTIME, &ts.ts.ts);
#else
			ts.fmt = TS_TIMEVAL;
			gettimeofday(&bt.ts.tv, NULL);
#endif
		}
		memset(&msg, 0, sizeof msg);
		msg.msg_name = &sind;
		msg.msg_namelen = sizeof(sind);
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = ctrl;
		msg.msg_controllen = sizeof ctrl;
		iov.iov_base = cbuf;
		iov.iov_len = sizeof cbuf;
		cmsg = (void*)ctrl;
		cookie = NULL;
		cookielen = 0;
		memset(rref, 0, sizeof rref);

		len = recvmsg(ip->socket, &msg, 0);
		if (len < 0 && errno == EAGAIN)	
			return;
		if (len < 0) /* XXX */
			return;

		LogHex(lpacket, cbuf, len, "RCV\n");

		for(;cmsg != NULL; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
#if 0
LogHex(Ldebug, CMSG_DATA(cmsg), cmsg->cmsg_len, "%x %x %x\n", cmsg->cmsg_level, cmsg->cmsg_type, cmsg->cmsg_len);
#endif
#ifdef SCM_TIMESTAMP
			if (cmsg->cmsg_level == SOL_SOCKET &&
			    cmsg->cmsg_type == SCM_TIMESTAMP &&
			    cmsg->cmsg_len == CMSG_LEN(sizeof ts.ts.tv)) {
				memcpy(&ts.ts.tv, CMSG_DATA(cmsg),
				    sizeof ts.ts.tv);
				ts.fmt = TS_TIMEVAL;
				continue;
			}
#endif
#ifdef SCM_BINTIME
			if (cmsg->cmsg_level == SOL_SOCKET &&
			    cmsg->cmsg_type == SCM_BINTIME &&
			    cmsg->cmsg_len == CMSG_LEN(sizeof ts.ts.bt)) {
				memcpy(&ts.ts.bt, CMSG_DATA(cmsg),
			 	    sizeof ts.ts.bt);
				ts.fmt = TS_BINTIME;
				continue;
			}
#endif
#ifdef IP_RECVDSTADDR
			if (cmsg->cmsg_level == IPPROTO_IP &&
			    cmsg->cmsg_type == IP_RECVDSTADDR &&
			    cmsg->cmsg_len == CMSG_LEN(sizeof(in_addr_t))) {
				cookie = cmsg;
				cookielen = CMSG_LEN(sizeof(in_addr_t));
				memcpy(&rip, CMSG_DATA(cmsg), sizeof rip);
				rip = ntohl(rip);
				rref[0] = (rip >> 24);
				rref[1] = (rip >> 16);
				rref[2] = (rip >> 8);
				rref[3] = (rip >> 0);
				continue;
			}
#endif
			warnx("option {level: %5d type %d len %d}\n",
			    cmsg->cmsg_level, cmsg->cmsg_type,
			    cmsg->cmsg_len);
		}
		if (ts.fmt == TS_INVALID) {
			/*
			 * We expected timestamp from the kernel and didn't
			 * get one, enable userland timestamping.  If the
			 * kernel provides (some) timestamps we will still
			 * use those.  Drop this packet.
			 */
			ip->userland_ts = 1;
			continue;
		}
		RcvPacket(ip, cbuf, len, (void*)&sind, &ts,
		    cookie, cookielen, rref);
	}
}

int
cfg_ntpv4_listen(CK_ARG, const char **adr)
{
	struct instance *ip;
	struct sockaddr_in sins;
	int s, i, error;

	ip = cfa->priv;	
	Assert(ip->magic == NTPV4_MAGIC, ("Wrong NTPv4 magic"));
	if (!ck_WordVar(CK_PASS, adr, &ip->listen))
		return (0);

	memset(&sins, 0, sizeof sins);
	sins.sin_len = sizeof(sins);
	sins.sin_family = AF_INET;
	sins.sin_addr.s_addr = INADDR_ANY;
	sins.sin_port = htons(WKP);

	s = GetBoundSocket(NULL, AF_INET, SOCK_DGRAM, 0,
	    &sins, sizeof sins);
	if (s < 0)
		return(-1);

#ifdef IP_RECVDSTADDR
	/*
	 * Set option to receive destination IP from packets (ie: our own
	 * IP.  This makes it possible to send back packets with correct
	 * source IP if we are multihomed
	 */
	i = 1;
	error = setsockopt(s, IPPROTO_IP, IP_RECVDSTADDR, &i, sizeof i);
	if (error < 0) {
		LogErr(Lwarn,
		   "Could not set IP_RECVDSTADDR.  No worries (%s)\n",
		   "(unless we are multihomed)");
	}
#endif

	error = -1;
#ifdef SO_BINTIME
	/*
	 * Try to timestamp packets in high resolution format at the
	 * entrance of the IP stack in the kernel.
	 */
	i = 1;
	error = setsockopt(s, SOL_SOCKET, SO_BINTIME, &i, sizeof i);
#endif
#ifdef SO_TIMESTAMP
	/*
	 * Otherwise try with lower resolution.
 	 */
	i = 1;
	if (error)
		error = setsockopt(s, SOL_SOCKET, SO_TIMESTAMP, &i, sizeof i);
#endif
	/*
	 * Otherwise, timestamp in userland.
	 */
	if (error)
		ip->userland_ts = 1;

	ip->socket = s;

	/*
	 * Set up events to get packets
	 */
	i = evSelectFD(C, ip->socket, EV_READ | EV_EXCEPT, socket_input,
	    ip, &ip->socket_fid);
	if (i)
		err(1, "evSelectFD(%s)", "socket");
	return (0);
}
