/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: partner.c,v 1.40 2007/05/08 08:59:28 phk Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/endian.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <ntpns.h>
#include <ntpv4.h>
#include <ntpv4_cfg.h>

LIST_HEAD(phead, partner);

static int nhash = 53;
static int timeout = 86400;
static int npartner = 0;

static int maxpartner = 10000;
static struct partner *partner;
static LIST_HEAD(, partner) freelist = LIST_HEAD_INITIALIZER(freelist);
static TAILQ_HEAD(, partner) lrulist = TAILQ_HEAD_INITIALIZER(lrulist);

static void config_timer(evContext ctx, void *uap, struct timespec due, struct timespec inter);
static struct phead *hash;

static int
InitPartner(void)
{
	int i;

	hash = calloc(sizeof *hash, nhash);
	if (hash == NULL) {
		nomem();
		return (-1);
	}
	partner = shmem_alloc(sizeof(struct partner) * maxpartner, TAG_PARTNER);
	if (partner == NULL) {
		nomem();
		return (-1);
	}
	/* Put them on the freelist in reverse order */
	for (i = maxpartner - 1; i >= 0; i--) {
		partner[i].generation = 0;
		partner[i].magic = PARTNER_MAGIC;
		partner[i].flag = P_F_NEVER;
		LIST_INSERT_HEAD(&freelist, &partner[i], hash);
	}
	for (i = 0; i < nhash; i++)
		LIST_INIT(&hash[i]);
	return (0);
}

/*
 * Lookup an existing, or attempt to create a new, partner structure.
 */
struct partner *
LookupPartner(time_t sec, uint32_t ip, int iswkp)
{
	int h;
	struct partner *p, *pp;

	h = ip % nhash;
	if (hash != NULL) {
		LIST_FOREACH_SAFE(p, &hash[h], hash, pp) {
			if (p->ip == ip && p->iswkp == iswkp) {
				LIST_REMOVE(p, hash);
				LIST_INSERT_HEAD(&hash[h], p, hash);
				if (p->configured == NULL) {
					TAILQ_REMOVE(&lrulist, p, lru);
					TAILQ_INSERT_TAIL(&lrulist, p, lru);
				}
				if (sec)
					p->last = sec;
				return (p);
			}
			if (p->configured == NULL && p->last + timeout < sec)
				DeletePartner( p);
		}
	} else if (InitPartner())
		return (NULL);

	if (LIST_EMPTY(&freelist) && !TAILQ_EMPTY(&lrulist))
		DeletePartner(TAILQ_FIRST(&lrulist));
	if (LIST_EMPTY(&freelist))
		return (NULL);

	p = LIST_FIRST(&freelist);
	LIST_REMOVE(p, hash);
	p->flag &= ~P_F_NEVER;
	p->ip = ip;
	p->iswkp = iswkp;
	p->generation++;
	npartner++;
	LIST_INSERT_HEAD(&hash[h], p, hash);
	TAILQ_INSERT_TAIL(&lrulist, p, lru);
	p->last = sec;
	return (p);
}

void
DeletePartner(struct partner *pp)
{
	u_int g;

	LIST_REMOVE(pp, hash);
	if (pp->configured) {
		evClearTimer(C, pp->tid);
		LIST_REMOVE(pp, config);
	} else {
		TAILQ_REMOVE(&lrulist, pp, lru);
	}
	if (pp->cookie != NULL)
		free(pp->cookie);
	g = pp->generation;
	memset(pp, 0, sizeof *pp);
	pp->generation = g + 1;
	LIST_INSERT_HEAD(&freelist, pp, hash);
	npartner--;
}

struct ck_instancehead *
cfg_ntpv4_server_head(CK_ARG)
{
	struct instance *ip;

	ip = cfa->priv;
	Assert(ip->magic == NTPV4_MAGIC, ("Wrong NTPv4 magic"));
	return (&ip->servers);
}

int
cfg_ntpv4_server_poll(CK_ARG, u_int *arg)
{
	struct configured *cf;

	cf = cfa->priv;
	Assert(cf->magic == CONF_MAGIC, ("Wrong configured magic"));
	ck_UintVar(CK_PASS, arg, &cf->poll, 64);
	return (0);
}

int
cfg_ntpv4_server_votes(CK_ARG, double *arg)
{
	struct configured *cf;

	cf = cfa->priv;
	Assert(cf->magic == CONF_MAGIC, ("Wrong configured magic"));
	ck_RealVar(CK_PASS, arg, &cf->votes, 0, "%g");
	return (0);
}

struct ck_instance *
cfg_ntpv4_server_new(CK_ARG, const char *word)
{
	struct configured *cp;
	struct instance *ip;
	int i;

	ip = cfa->priv;
	Assert(ip->magic == NTPV4_MAGIC, ("Wrong NTPv4 magic"));

	cp = calloc(sizeof *cp, 1);
	if (cp == NULL) {
		nomem();
		return (NULL);
	}
	cp->magic = CONF_MAGIC;
	cp->poll = 64;
	cp->votes = 0;
	cp->name = strdup(word);
	cp->inp = ip;
	if (cp->name == NULL) {
		nomem();
		free(cp);
		return (NULL);
	}
	LIST_INIT(&cp->partner);
	i = evSetTimer(C, config_timer, cp,
	    evConsTime(0, 0), evConsTime(0,0), &cp->tid);
	Assert(i == 0, ("EvSetTimer(config_timer)"));
	cp->cki.obj = cp;
	return (&cp->cki);
}

/*
 * XXX: this is suboptimal, we can have multiple partners and sources
 */
int
cfg_ntpv4_server_setup(CK_ARG)
{
	struct configured *cp;
	struct partner *pp;

	cp = cfa->priv;
	Assert(cp->magic == CONF_MAGIC, ("Wrong configured magic"));
	pp = LIST_FIRST(&cp->partner);
	if (pp != NULL)
		cfa->priv = pp->sid;
	else
		cfa->priv = NULL;
	return (0);
}

void
DeleteConfigured(struct configured *cp)
{
	struct partner *p;

	Assert(cp->magic == CONF_MAGIC, ("Wrong configured magic"));
	evClearTimer(C, cp->tid);
	LIST_REMOVE(cp, next);
	free(cp->name);
	for (;;) {
		p = LIST_FIRST(&cp->partner);
		if (p == NULL)
			break;
		TAILQ_INSERT_TAIL(&lrulist, p, lru);
		DeletePartner(p);
	}
	free(cp);
}

/*
 * This timer lingers for all configured peers.  It is responsible for
 * DNS resolution and generally figuring out that things have changed.
 */

static void
config_resolved(CK_ARG, void *priv, int error, struct addrinfo *ai)
{
	struct addrinfo *a;
	struct partner *p, *pt;
	struct configured *cf;
	struct sockaddr_in *sa;
	char hbuf[NI_MAXHOST];
	int i;
	uint32_t nip;

	cf = priv;
	Assert(cf->magic == CONF_MAGIC, ("Wrong configured magic"));

	if (error) {
		/* Check again not quite five minutes from now */
		i = evSetTimer(C, config_timer, cf,
		    evAddTime(evNowTime(),
		    evConsTime(5 * 60 + (random() & 0x3f), 0)), /* 5m...6m */
		    evConsTime(0, 0), &cf->tid);
		if (i < 0)
			err(1, "evSetTimer() failed");
		return;
	}

	LIST_FOREACH(p, &cf->partner, config)
		p->flag &= ~P_F_UPDATE;

	for (a = ai; a != NULL; a = a->ai_next) {
		if (a->ai_addrlen != sizeof p->addr) {
			warnx("ai_addlen=%d, expected %d",
				a->ai_addrlen, sizeof p->addr);
			continue;
		}
		sa = (void*)a->ai_addr;
		nip = ntohl(sa->sin_addr.s_addr);
		LIST_FOREACH(p, &cf->partner, config) {
			if (p->ip != nip)
				continue;
			p->flag |= P_F_UPDATE;
			break;
		}
		if (p != NULL) 
			continue;

		p = LookupPartner(0, nip, WKP == WKP);
		if (p == NULL)
			err(1, "XXX: malloc failed");
		if (p->configured != NULL)	/* belongs elsewhere */
			continue;
		p->flag |= P_F_UPDATE;
		LIST_INSERT_HEAD(&cf->partner, p, config);
		memcpy(&p->addr, a->ai_addr, a->ai_addrlen);
		p->configured = cf;
		TAILQ_REMOVE(&lrulist, p, lru);
		if (p->sid == NULL) {
			i = getnameinfo(a->ai_addr, a->ai_addr->sa_len,
			    hbuf, sizeof(hbuf), NULL, 0, NI_NUMERICHOST);
			if (i != 0)
				err(1, "could not get numeric hostname");
			p->sid = SourceAdd(hbuf);
			SourceParam(p->sid, SRC_UPDATE_RATE, cf->poll);
			SourceParam(p->sid, SRC_LOS_TIMEOUT, 3 * cf->poll);
		}
		i = evSetTimer(C, poll_timer, p,
		     evConsTime(0, 0), evConsTime(cf->poll, 0), &p->tid);
		if (i < 0)
			err(1, "evSetTimer() failed");
		p->inp = cf->inp;
	}
	LIST_FOREACH_SAFE(p, &cf->partner, config, pt) {
		if (p->flag & P_F_UPDATE)
			continue;
		if (p->sid != NULL)
			SourceDel(p->sid);
		DeletePartner(p);
	}

	/* Check again not quite an hour from now */
	i = evSetTimer(C, config_timer, cf,
	     evAddTime(evNowTime(),
	     evConsTime((random() & 0x3f) + 59 * 60 , 0)), /* roughly 1h */
	     evConsTime(0, 0), &cf->tid);
	if (i < 0)
		err(1, "evSetTimer() failed");
}

static void
config_timer(evContext ctx __unused, void *uap, struct timespec due __unused, struct timespec inter __unused)
{
	struct configured *cf;
	int i;

	cf = uap;
	Assert(cf->magic == CONF_MAGIC, ("Wrong configured magic"));

	Log(lpartner, "config_timer(%p, %s)", cf, cf->name);
	i = resolver(CK_NOARG, cf->name, "ntp", PF_UNSPEC, SOCK_DGRAM,
	    AI_CANONNAME, config_resolved, cf);
	if (i < 0)
		Log(Lerror, "ntpv4 config_timer(%p, %s): resolver failed",
		    cf, cf->name);
}

static void
show_partner_head(CK_ARG)
{

	ck_printf(CK_PASS, "%-16s  port ", "IP number");
	PrintPacketHeader(CK_PASS);
	ck_printf(CK_PASS, "\n");
}

static void
show_partner(CK_ARG, struct partner *p, int show_packet)
{
	char buf[20];
	uint32_t ip;
	int i;

	ip = ntohl(p->addr.sin_addr.s_addr);
	i = ntohs(p->addr.sin_port);
	sprintf(buf, "%d.%d.%d.%d",
	    (ip >> 24) & 0xff, (ip >> 16) & 0xff,
	    (ip >> 8) & 0xff, (ip >> 0) & 0xff);
	ck_printf(CK_PASS, "%-16s %5u ", buf, i);
	PrintPacket(CK_PASS, p, show_packet);
	ck_printf(CK_PASS, "\n");
}

int
cfg_ntpv4_partner(CK_ARG, const char **word)
{
	struct instance *ip;
	struct ck_instance *cki;
	struct configured *cf;
	struct partner *p, *p2;
	time_t sec;
	int n_total, n_1000, n_100, n_10, n_1, n_0, n_good, n_bad;
	int o_total, o_1000, o_100, o_10, o_1, o_0, o_good, o_bad;
	int bits;
	uint32_t mask;
        struct in_addr		ina;
	double delta;

	if (!(cfa->flag & CK_SHOW))
		return (0);
	ip = cfa->priv;
	Assert(ip->magic == NTPV4_MAGIC, ("Wrong NTPv4 magic"));

	time(&sec);

	/* calculate summary */
	n_good = n_bad = n_total = n_1000 = n_100 = n_10 = n_1 = n_0 = 0;
	o_good = o_bad = o_total = o_1000 = o_100 = o_10 = o_1 = o_0 = 0;
	TAILQ_FOREACH_SAFE(p, &lrulist, lru, p2) {
		if (p->last + timeout < sec) {
			DeletePartner(p);
			continue;
		}
		n_total++;
		if (p->packet.ours)
			o_total++;
		delta = fabs(p->delta);
		if (delta >= .128) {
			n_bad++;
			if (p->packet.ours)
				o_bad++;
		} else {
			n_good++;
			if (p->packet.ours)
				o_good++;
		}
		if (delta >= 1.0) {
			n_1000++;
			if (p->packet.ours)
				o_1000++;
		} else if (delta >= .1) {
			n_100++;
			if (p->packet.ours)
				o_100++;
		} else if (delta >= .01) {
			n_10++;
			if (p->packet.ours)
				o_10++;
		} else if (delta >= .001) {
			n_1++;
			if (p->packet.ours)
				o_1++;
		} else {
			n_0++;
			if (p->packet.ours)
				o_0++;
		}
	}

	if (word == NULL)  {
		show_partner_head(CK_PASS);
		TAILQ_FOREACH(cki, &ip->servers, next) {
			cf = cki->obj;
			Assert(cf->magic == CONF_MAGIC, ("Wrong configured magic"));
			ck_printf(CK_PASS, "NAME %s\n", cf->name);
			LIST_FOREACH(p, &cf->partner, config)
				show_partner(CK_PASS, p, 0);
		}
		ck_printf(CK_PASS, "Max partners:\t\t%d\n", maxpartner);
		ck_printf(CK_PASS, "%-17s %6s %6s %6s\n", "",
		    "total", "ours", "others");
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners",
		    n_total, o_total, n_total - o_total);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners good",
		    n_good, o_good, n_good - o_good);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners bad",
		    n_bad, o_bad, n_bad - o_bad);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners > 1s",
		    n_1000, o_1000, n_1000 - o_1000);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners < 1s",
		    n_100, o_100, n_100 - o_100);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners < 100ms",
		    n_10, o_10, n_10 - o_10);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners < 10ms",
		    n_1, o_1, n_1 - o_1);
		ck_printf(CK_PASS, "%-17s %6d %6d %6d\n", "partners < 1ms",
		    n_0, o_0, n_0 - o_0);
		return (0);
	}

	n_total = o_total = o_bad = mask = 0;

	bits = inet_net_pton(AF_INET, *word, &ina, sizeof(ina));
	show_partner_head(CK_PASS);
	if (bits >= 0) {
		mask = 0xffffffff;
		if (bits < 32)
			mask <<= (32 - bits);
		ina.s_addr = ntohl(ina.s_addr);
		ina.s_addr &= mask;
	} else if (!strcmp(*word, "ours")) {
		o_total = 1;
	} else if (!strcmp(*word, "bad")) {
		o_bad = 1;
	} else if (!strcmp(*word, "all")) {
		;
	} else {
		TAILQ_FOREACH(cki, &ip->servers, next) {
			cf = cki->obj;
			Assert(cf->magic == CONF_MAGIC, ("Wrong configured magic"));
			if (!strcmp(cf->name, *word)) {
				LIST_FOREACH(p, &cf->partner, config)
					show_partner(CK_PASS, p, 1);
				return (0);
			}
		}
		ck_printf(CK_PASS, "Unknown argument\n");
		return (0);
	}

	TAILQ_FOREACH_SAFE(p, &lrulist, lru, p2) {
		if (p->last + timeout < sec) {
			DeletePartner(p);
			continue;
		}
		if (o_bad && fabs(p->delta) < .128)
			continue;
		if (o_total && !p->packet.ours)
			continue;
		if (mask &&
		    (ina.s_addr != (mask & ntohl(p->addr.sin_addr.s_addr))))
			continue;
		if (mask == 0xffffffff) 
			show_partner(CK_PASS, p, 1);
		else
			show_partner(CK_PASS, p, 0);
	}
	return (0);
}
