/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: proto.c,v 1.42 2007/01/18 09:49:44 phk Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/endian.h>
#include <netinet/in.h>

#include <ntpns.h>
#include <ntpv4.h>
#include <ntpv4_cfg.h>

/**********************************************************************/

#define MODE_0		0
#define MODE_SYMACT	1
#define MODE_SYMPAS	2
#define MODE_CLIENT	3
#define MODE_SERVER	4
#define MODE_BCAST	5
#define MODE_CONTROL	6
#define MODE_7		7

#if 0
/*
 * Number of Days since 1 BC Gregorian to 1 January of given year
 */
#define julian0(year)	(((year) * 365 ) + ((year) > 0 ? (((year) + 3) \
			/ 4 - ((year - 1) / 100) + ((year - 1) / \
			400)) : 0))

/*
 * Number of days since start of NTP time to 1 January of given year
 */
#define ntp0(year)	(julian0(year) - julian0(1900))

/*
 * Number of days since start of UNIX time to 1 January of given year
 */
#define unix0(year)	(julian0(year) - julian0(1970))

#define	JAN_1970	0x83aa7e80	/* 2208988800 1970 - 1900 in seconds */

#endif

/**********************************************************************/

void
PrintPacketHeader(CK_ARG)
{

	ck_printf(CK_PASS, "err leap v m  s  p   P         offset refid");
}

void
PrintPacket(CK_ARG, struct partner *pr, int how)
{
	char buf[50];
	struct ntp_packet *p;
	time_t l;

	ck_printf(CK_PASS, "%3d ", pr->err);
	p = &pr->packet;
	switch (p->leap) {
	case 0: ck_printf(CK_PASS, "no  "); break;
	case 1: ck_printf(CK_PASS, "ins "); break;
	case 2: ck_printf(CK_PASS, "del "); break;
	case 3: ck_printf(CK_PASS, "unk "); break;
	}
	ck_printf(CK_PASS, " %1d", p->version);
	ck_printf(CK_PASS, " %1d", p->mode);
	ck_printf(CK_PASS, " %2d", p->stratum);
	ck_printf(CK_PASS, " %2d", p->poll);
	ck_printf(CK_PASS, " %3d", p->prec);
	if (fabs(pr->delta) < 1000.0) {
		ck_printf(CK_PASS, " %14.9f", pr->delta);
	} else {
		ck_printf(CK_PASS, " %14g", pr->delta);
	}
	ck_printf(CK_PASS, " %s", FormatRefid(p->refid, p->stratum));
	if (p->ours)
		ck_printf(CK_PASS, " = ours");
	if (!how)
		return;
	ck_printf(CK_PASS, "\n\tDelay		%9.2e", p->delay);
	ck_printf(CK_PASS, "\n\tDispersion 	%9.2e", p->disp);
	time(&l);
	TsFormat(&p->ts[4], buf);
	ck_printf(CK_PASS, "\n\tOur Rx		%s (%ds ago)",
	    buf, l - p->ts[4].ts.bt.sec);
	TsFormat(&p->ts[0], buf);
	ck_printf(CK_PASS, "\n\tReference	%s", buf);
	TsFormat(&p->ts[1], buf);
	ck_printf(CK_PASS, "\n\tOriginate	%s", buf);
	TsFormat(&p->ts[2], buf);
	ck_printf(CK_PASS, "\n\tReceive		%s", buf);
	TsFormat(&p->ts[3], buf);
	ck_printf(CK_PASS, "\n\tTransmit	%s", buf);
}

static uint
LeapBits(void)
{

	if (sys_leap == SRC_LEAP_INSERT)
		return (0x1 << 6);
	if (sys_leap == SRC_LEAP_DELETE)
		return (0x2 << 6);
	if (sys_leap == SRC_LEAP_NO)
		return (0x0 << 6);
	return (0x3 << 6);
}

static void
SendPacket(struct partner *pr, struct msghdr *mh, u_char *buf)
{
	uint64_t u;
	int i;

	u = TsNTPNow();
	be64enc(buf + 40, u);
	be64enc(buf + 16, u);

	Assert(pr->inp != NULL, ("Null inp on partner"));

	i = sendmsg(pr->inp->socket, mh, 0);
	if (i == 48)
		pr->err = 0;
	else if (i < 0)
		pr->err = errno;
	else
		pr->err = -i;
	free(mh);
	pr->originate = u;
}

static void
ReplyNtp(struct partner *pr)
{
	u_char buf[48];
	struct msghdr *mh;
	uint64_t u;
	struct ntp_packet *p;

	if (sys_able != 0)
		return;
	p = &pr->packet;
	if (p->mode != 1 && p->mode != 3)
		return;
	mh = socket_prepare(pr, buf, 48);

	buf[0] = LeapBits();
	buf[0] |= p->version << 3;
	if (p->mode == 1)
		buf[0] |= 2;
	else if (p->mode == 3)
		buf[0] |= 4;
	buf[1] = sys_stratum;
	buf[2] = p->poll;
	buf[3] = sys_precision;
	be32enc(buf + 4, sys_delay * 65536);
	be32enc(buf + 8, sys_dispersion * 65536);
	memcpy(buf + 12, sys_refid, sizeof sys_refid);

	TsNTPtime(&p->ts[3], &u);
	be64enc(buf + 24, u);

	TsNTPtime(&p->ts[4], &u);
	be64enc(buf + 32, u);

	SendPacket(pr, mh, buf);
}

static void
SendNtp(struct partner *pr)
{
	u_char buf[48];
	struct msghdr *mh;
	uint64_t u;
	struct ntp_packet *p;
	mh = socket_prepare(pr, buf, 48);

	p = &pr->packet;

	buf[0] = LeapBits();
	buf[0] |= 4 << 3;
	buf[0] |= 3;
	buf[1] = sys_stratum;
	buf[2] = 4; /* XXX */
	buf[3] = sys_precision;
	be32enc(buf + 4, sys_delay * 65536);
	be32enc(buf + 8, sys_dispersion * 65536);
	memcpy(buf + 12, sys_refid, sizeof sys_refid);
	TsNTPtime(&p->ts[3], &u);
	be64enc(buf + 24, u);

	TsNTPtime(&p->ts[4], &u);
	be64enc(buf + 32, u);

	SendPacket(pr, mh, buf);
	Log(lsocket, "SendNtp(%p, %08x)", pr, pr->ip);
}


/*
 * XXX: I'm not happy with the ip+port argument.  Ideally it should be an
 * XXX: opaque chunk of data, but I need to look into it to be able to
 * XXX: to tell if the source port is 123 and possibly detect NAT'ing.
 * XXX: It might make sense to change this to a simpel "iswkp" boolean.
 */
void
RcvPacket(struct instance *inp, void *packet, int len, struct sockaddr_in *sa, struct ts *ts, void *cookie, u_int cookielen, u_char *ref)
{
	uint8_t		*s;
	struct ntp_packet *p;
	struct partner *pr;
	uint32_t ip;
	u_char refid[4];
	int port;
	double d;

	ip = ntohl(sa->sin_addr.s_addr);
	port = ntohs(sa->sin_port);
	s = packet;

	if ((inp->filter & NTPV4_FILTER_FRAC0) && be32dec(s + 44) == 0)
		return;

	pr = LookupPartner(ts->ts.bt.sec, ip, port == WKP);
	Log(lsocket,
	    "RcvPacket(%p, %d, %08x, %d, %p, %p, %d, -> %p) %d.%d.%d.%d:%d",
	    packet, len, ip, port, ts, cookie, cookielen, pr,
		(ip >> 24) & 255,
		(ip >> 16) & 255,
		(ip >> 8) & 255,
		(ip >> 0) & 255,
		port);
	if (pr == NULL)
		return;

	if (pr->configured == NULL)
		memcpy(&pr->addr, sa, sizeof pr->addr);

	pr->inp = inp;	/* XXX also dependent on pr->configured ? */

	if (cookie != NULL) {
		if (pr->cookielen != cookielen)
			pr->cookie = realloc(pr->cookie, cookielen);
		if (pr->cookie != NULL) {
			memcpy(pr->cookie, cookie, cookielen);
			pr->cookielen = cookielen;
		} else {
			pr->cookielen = 0;
		}
	} else if (pr->cookie != NULL) {
		free(pr->cookie);
		pr->cookie = NULL;
		pr->cookielen = 0;
	}

	p = &pr->packet;

	/* Decode the packet */
	p->leap = (s[0] >> 6) & 0x3;
	p->version = (s[0] >> 3) & 0x7;
	p->mode = s[0] & 0x7;
	p->stratum = s[1];
	p->poll = s[2];
	p->prec = s[3];
	p->delay = (int32_t)be32dec(s + 4) / 65536.0;
	p->disp = (int32_t)be32dec(s + 8) / 65536.0;
	p->refid[0] = s[12];
	p->refid[1] = s[13];
	p->refid[2] = s[14];
	p->refid[3] = s[15];
	p->ours = !memcmp(p->refid, ref, 4);
	p->ts[0].ts.ntp = be64dec(s + 16);
	p->ts[0].fmt = TS_NTPTIME;
	p->ts[1].ts.ntp = be64dec(s + 24);
	p->ts[1].fmt = TS_NTPTIME;
	p->ts[2].ts.ntp = be64dec(s + 32);
	p->ts[2].fmt = TS_NTPTIME;
	p->ts[3].ts.ntp = be64dec(s + 40);
	p->ts[3].fmt = TS_NTPTIME;
	p->ts[4] = *ts;
	
	switch(p->mode) {
	case MODE_0:
	case MODE_CONTROL:
	case MODE_7:
		return;
	}


	if (pr->configured == NULL ||
	    p->mode == MODE_SYMACT || p->mode == MODE_CLIENT)
		ReplyNtp(pr);

	d = TsDiff(&p->ts[3], &p->ts[4]);
	TsNormalize(&p->ts[1]);
	TsNormalize(&p->ts[2]);
	if (p->ts[1].fmt != TS_INVALID && p->ts[2].fmt != TS_INVALID) {
		d += TsDiff(&p->ts[2], &p->ts[1]);
		d /= 2;
	}
	pr->delta = d;

	if (pr->sid != NULL && pr->originate == be64dec(s + 24)) {
		SourceParam(pr->sid, SRC_DELAY, p->delay);
		SourceParam(pr->sid, SRC_DISPERSION, p->disp);
		SourceParam(pr->sid, SRC_STRATUM, p->stratum);
		be32enc(refid, ip);
		SourceRefid(pr->sid, refid);
		SourceDelta(pr->sid, &p->ts[3], pr->delta, SRC_UTC);
		if (pr->configured != NULL)
			SourceParam(pr->sid, SRC_VOTES, pr->configured->votes);
		switch(p->leap) {
		case 0x0: SourceLeap(pr->sid, SRC_LEAP_NO); break;
		case 0x1: SourceLeap(pr->sid, SRC_LEAP_INSERT); break;
		case 0x2: SourceLeap(pr->sid, SRC_LEAP_DELETE); break;
		case 0x3: SourceLeap(pr->sid, SRC_LEAP_UNKNOWN); break;
		}
	}
}

void
poll_timer(evContext ctx __unused, void *uap, struct timespec due __unused, struct timespec inter __unused)
{
	struct partner *p;

	p = uap;
	Assert(p->magic == PARTNER_MAGIC, ("Wrong partner magic"));
	SendNtp(p);
}

int
cfg_ntpv4_filter(CK_ARG, const char **key)
{
	struct instance *ip;

	ip = cfa->priv;	
	Assert(ip->magic == NTPV4_MAGIC, ("Wrong NTPv4 magic"));

	ck_BoolKeyVar(CK_PASS, key,
	    "frac0", &ip->filter, NTPV4_FILTER_FRAC0, 0);
	return (0);
}
