/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore_msg.c,v 1.61 2005/12/30 10:07:58 phk Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/endian.h>
#include <pthread.h>

#include <ntpns.h>

#include "oncore.h"
#include "oncore_hash.h"

struct oncore_msg **oncore_msgs;

/*
 * Almanac subroutines
 */
static void
almanac_load(struct oncore *ip)
{
	int j;

	/*
	 * The entire almanac is too big to be loaded in one second
	 * so load only half of it each second
	 */
	for (j = 0; j < NALMANAC / 2 && ip->almanac_n < NALMANAC; j++) {
		oncore_sendmsg(ip, ip->almanac_ptr + 2, LALMANAC - 5);
		ip->almanac_ptr += LALMANAC;
		ip->almanac_n++;
	}
	/*
	 * Set state back to almanac collection, the receiver will
	 * return the almanac we just loaded to us.
	 */
	if (ip->almanac_n == NALMANAC) {
		free(ip->almanac);
		ip->almanac = NULL;
		ip->almanac_ptr = NULL;
		ip->almanac_work = ALMANAC_DONE;
	}
}

static void
almanac_done(struct oncore *ip)
{
	/*
	 * We are done with the almanac thread, join it to recover
	 * the status block.
	 * Based on a theory on last minute mistakes, ask for a
	 * leap-second status, just in case...
	 */
	pthread_join(ip->almanac_thread, NULL);
	ip->almanac_work = ALMANAC_COLLECT;
	oncore_sendmsg(ip, "Gj", 2);     /* Leap sec status */
}

static void *
Cb_thread(void *p)
{
	struct oncore *ip;
	FILE *f;
	const char *fn2;
	char *fn;
	int i;

	ip = p;
	if (ip->almanac_file != NULL)
		fn2 = ip->almanac_file;
	else
		fn2 = DEFAULT_ALMANAC_FILE;
	if (ip->almanac_work == ALMANAC_SAVE) {
		asprintf(&fn, "%s.%d.tmp", fn2, getpid());
		if (fn == NULL) {
			nomem();
			pthread_exit(NULL);
		}
		f = fopen(fn, "w");
		if (f == NULL) {
			warn("cannot open %s", fn);
			free(fn);
			pthread_exit(NULL);
		}
		i = fwrite(ip->almanac, 33 * 34, 1, f);
		if (fclose(f) != 0 || i != 1) {
			warn("cannot write/close %s", fn);
			unlink(fn);
			free(fn);
			pthread_exit(NULL);
		}
		i = rename(fn, fn2);
		if (i != 0) {
			warn("cannot rename %s to %s", fn, fn2);
			unlink(fn);
		}
		free(fn);
		free(ip->almanac);
		ip->almanac = NULL;
		ip->almanac_ptr = NULL;
		ip->almanac_work = ALMANAC_DONE;
		pthread_exit(NULL);
	}

	if (ip->almanac_work == ALMANAC_LOAD) {
		if (ip->almanac == NULL)
			ip->almanac = calloc(1, 33 * 34);
		if (ip->almanac == NULL) {
			nomem();
			pthread_exit(NULL);
		}
		f = fopen(fn2, "r");
		if (f == NULL) {
			warn("cannot open %s", fn2);
			pthread_exit(NULL);
		}
		i = fread(ip->almanac, 33 * 34, 1, f);
		if (fclose(f) != 0 || i != 1) {
			warn("cannot read/close %s", fn2);
			free(ip->almanac);
			ip->almanac = NULL;
			pthread_exit(NULL);
		}
		ip->almanac_ptr = ip->almanac;
		ip->almanac_work = ALMANAC_LOADED;
		ip->almanac_n = 0;
		pthread_exit(NULL);
	}
	warnx("ip->almanac_work = %d ??", ip->almanac_work);
	pthread_exit(NULL);
}

static void
handle_Cb(struct oncore *ip, struct oncore_msg *mp)
{
	int s, p, i;

	if (ip->almanac_work != ALMANAC_COLLECT)
		return;
	s = ip->serial_buf[4];
	p = ip->serial_buf[5];
	if (s == 0 && p == 0) {
		/* Receiver has no almanac, try to get it one */
		ip->almanac_work = ALMANAC_LOAD;
		i = pthread_create(&ip->almanac_thread, NULL, Cb_thread, ip);
		if (i != 0) {
			warn("pthread_create failed");
			free(ip->almanac);
			ip->almanac = ip->almanac_ptr = NULL;
		}
		return;
	}
	if (s == 5 && p == 1) {
		if (ip->almanac == NULL) {
			ip->almanac = calloc(1, 34 * mp->rs_len);
			if (ip->almanac == NULL) {
				nomem();
				return;
			}
		}
		ip->almanac_ptr = ip->almanac;
	}
	if (ip->almanac == NULL)
		return;
	memcpy(ip->almanac_ptr, ip->serial_buf, mp->rs_len);
	ip->almanac_ptr += mp->rs_len;
	/* are we done yet ? */
	if (ip->almanac_ptr != ip->almanac + 34 * mp->rs_len)
		return;

	Log(Linfo, "Got full almanac");

	ip->almanac_work = ALMANAC_SAVE;
	i = pthread_create(&ip->almanac_thread, NULL, Cb_thread, ip);
	if (i != 0) {
		warn("pthread_create failed");
		free(ip->almanac);
		ip->almanac = ip->almanac_ptr = NULL;
	}
}

/*
 * The rest of the stuff
 */

static void
oncore_params(struct oncore *ip)
{
	u_char msg[10];
	unsigned u;

	u = (uint)(ip->cf_raim_limit / 1e-7);
	if (ip->raim_limit != u) {
		msg[0] = 'G';		/* TRAIM */
		msg[1] = 'f';
		be16enc(msg + 2, u);
		oncore_sendmsg(ip, msg, 4);
	}
	if (ip->mask_angle != ip->cf_mask_angle) {
		msg[0] = 'A';		/* Mask Angle */
		msg[1] = 'g';
		msg[2] = ip->cf_mask_angle;
		oncore_sendmsg(ip, msg, 3);
	}
}

void
oncore_initialize(struct oncore *ip)
{
	u_char msg[50];

	oncore_sendmsg(ip, "Cj", 2); /* Receiver ID */
	oncore_sendmsg(ip, "Bb\001", 3); /* Sat visibility */
	oncore_sendmsg(ip, "Bd\001", 3); /* Almanac status */   
	oncore_sendmsg(ip, "Be\001", 3); /* Req Almanac */
	oncore_sendmsg(ip, "Bj\001", 3); /* Leap sec status */
	oncore_sendmsg(ip, "Bo\001", 3); /* Req UTC offset */
	if (ip->type == UT) {
		oncore_sendmsg(ip, "Ea\001", 3); /* Req pos/data */
		oncore_sendmsg(ip, "At\000", 3); /* No pos hold */
		ip->site_survey = 0;

		msg[0] = 'E';		/* TRAIM */
		msg[1] = 'n';
		msg[2] = 1;
		msg[3] = 1;
		be16enc(msg + 4, 10);
		msg[6] = 2;
		memset(msg + 7, 0, 10);
		oncore_sendmsg(ip, msg, 17);

		if (ip->tsat > 2) {
			msg[0] = 'A';
			msg[1] = 'd';
			be32enc(msg + 2, ip->lat);
			oncore_sendmsg(ip, msg, 6);

			msg[0] = 'A';
			msg[1] = 'e';
			be32enc(msg + 2, ip->lon);
			oncore_sendmsg(ip, msg, 6);

			msg[0] = 'A';
			msg[1] = 'f';
			be32enc(msg + 2, ip->ht);
			msg[6] = 0;
			oncore_sendmsg(ip, msg, 7);
		}
	}
	if (ip->type == M12) {
		oncore_sendmsg(ip, "Gc\003", 3); /* 1pps when RAIM happy */
		oncore_sendmsg(ip, "Gd\003", 3); /* pos mode (sitesurv) */
		ip->site_survey_timer = 10000;
		oncore_sendmsg(ip, "Ge\001", 3); /* TRAIM (enable)*/
		oncore_sendmsg(ip, "Ag\377", 3); /* Mask Angle (poll) */
		oncore_sendmsg(ip, "Gf\377\377", 4); /* TRAIM limit (poll) */
		oncore_sendmsg(ip, "Ha\001", 3); /* pos msg */
		oncore_sendmsg(ip, "Hn\001", 3); /* TRAIM msg */

		if (ip->tsat > 2 && (ip->rcv_status >> 13) >= 4) {
			msg[0] = 'G';
			msg[1] = 'a';
			be32enc(msg + 2, ip->lat);
			be32enc(msg + 6, ip->lon);
			be32enc(msg + 10, ip->ht);
			msg[14] = 0;
			oncore_sendmsg(ip, msg, 15);
		}
	}
	/*
	 * XXX: It seems to be a bad idea to set the clock.  The receiver
	 * will find time after the first sat tracked in mode 8, so little
	 * time is saved in any case.
	 * If we set the clock wrong, the oncore will find the next 1024week
	 * period and we risk jumping several years forward in time.
	 * It is not clear that the receiver ever finds out.  Appearantly
	 * any 1024week disambiguation info in the almanac is not read
	 */
}

/*
 * This handler does nothing, it is only here to suppress a hexdump in the
 * debugging output.
 */
static void
handle_NOOP(struct oncore *ip __unused, struct oncore_msg *mp __unused)
{
}

/*
 * Response from set to defaults
 */

static void
handle_Cf(struct oncore *ip, struct oncore_msg *mp __unused)
{

	if (ip->state == ST_RESET_WAIT)
		oncore_state(ip, ST_TEST);
	else
		oncore_state(ip, ST_RESET);
}

/*
 * Response to self test
 */

static void
handle_IaFa(struct oncore *ip, struct oncore_msg *mp __unused)
{
	uint32_t st;

	if (msg_mash("@@Fa") == msg_mash(ip->serial_buf)) {
		st = be16dec(ip->serial_buf + 4);
		ip->type = UT;
	} else {
		st = be32dec(ip->serial_buf + 4) >> 8;
		ip->type = M12;
	}
	if (st == 0) {
		oncore_state(ip, ST_INIT);
		return;
	}
	Log(Lwarn, "oncore %d: Receiver failed selftest: %03x", ip->unit, st);
	oncore_state(ip, ST_SICK);
}

/*
 * Health:
 *	0x01 removed
 *	0x02 unhealthy
 */
static void
handle_Bb(struct oncore *ip, struct oncore_msg *mp __unused)
{
	int i, j, k, ns;
	u_char *buf = ip->serial_buf;

	for (i = 0; i < NSAT; i++) {
		ip->persat[i].doppler = 0;
		ip->persat[i].elevation = 0;
		ip->persat[i].azimuth = 0;
		ip->persat[i].health = 4;
	}
	
	ns = buf[4];
	if (ns > 12)
		ns = 12;
	for (i = 0; i < ns; i++) {
		j = buf[5 + i * 7];
		if (j == 0)
			continue;
		if (j > 32) {
			Log(Lwarn, "SVID %d in @@Bb", j);
			continue;
		}
		k =  be16dec(buf + 6 + i * 7);
		if (k > 32767)
			k -= 65536;
		ip->persat[j].doppler = k;
		ip->persat[j].elevation = buf[8 + i * 7];
		ip->persat[j].azimuth = be16dec(buf + 9 + i * 7);
		ip->persat[j].health = buf[11 + i * 7];
	}
}

static void
handle_Ha(struct oncore *ip, struct oncore_msg *mp __unused)
{
	int i, j, k;
	u_char *buf = ip->serial_buf;
	char txt[50], txt2[50];
	const char *p;
	struct ts ts2[2];

	i = PpsPickup(&ip->pps, ts2, 1);
	ip->ts = ts2[0];
	record("ONCOREPPS", ip->unit, &ip->ts, sizeof ip->ts);

	/*
	 * XXX: we should compensate for the negative sawtooth here, but
	 * XXX: the magnitude is trivial compared to our PPS jitter so
	 * XXX: leave it for later
	 */
	if (i == 1)
		p = TsFormat(&ip->ts, txt);
	else 
		p = "no_pps_pulse";

	if (ip->almanac_work == ALMANAC_LOADED)
		almanac_load(ip);
	if (ip->almanac_work == ALMANAC_DONE)
		almanac_done(ip);
	
	for (i = 0; i < NSAT; i++) {
		ip->persat[i].mode = 0;
		ip->persat[i].signal_strength = 0;
		ip->persat[i].iode = 0;
		ip->persat[i].status = 0;
	}
	ip->tm.tm_mon = buf[4] - 1;
	ip->tm.tm_mday = buf[5];
	ip->tm.tm_year = be16dec(buf + 6) - 1900;
	ip->tm.tm_hour = buf[8];
	ip->tm.tm_min = buf[9];
	ip->tm.tm_sec = buf[10];
	ip->fraction = be32dec(buf + 11);
	ip->flat = be32dec(buf + 15);
	ip->flon = be32dec(buf + 19);
	ip->fht = be32dec(buf + 23);
	ip->lat = be32dec(buf + 31);
	ip->lon = be32dec(buf + 35);
	ip->ht = be32dec(buf + 39);
	ip->dop = be16dec(buf + 53);
	ip->vsat = buf[55];
	ip->tsat = buf[56];
	k = 0;
	for (i = 0; i < 12; i++) {
		j = buf[57 + i * 6];
		if (j == 0)
			continue;
		ip->persat[j].mode = buf[58 + i * 6];
		if (ip->persat[j].mode == 8)
			k++;
		ip->persat[j].signal_strength = buf[59 + i * 6];
		ip->persat[j].iode = buf[60 + i * 6];
		ip->persat[j].status = be16dec(buf + 61 + i * 6);
	}
	ip->lsat = k;
	ip->rcv_status = be16dec(buf + 129);
	if (((ip->rcv_status & 0xe000) == 0xe000) &&
	    ip->site_survey_timer > 0)
		ip->site_survey_timer--;
	ip->clock_bias = be16dec(buf + 133);
	ip->osc_offset = be32dec(buf + 135);
	ip->osc_temp = be16dec(buf + 139);
	ip->lutc_status = ip->utc_status;
	ip->utc_status = buf[141];
	ip->gmt_sign = buf[142];
	ip->gmt_hour = buf[143];
	ip->gmt_min = buf[144];
	strlcpy(ip->serno, buf + 145, sizeof ip->serno);

	ip->t = timegm(&ip->tm);

	/*
	 * Every so often, the Oncore will mis-estimate things and end
	 * up being too early.  Since the fraction field cannot be negative
	 * what happens is that the time fields are for the previous second
	 * and the fraction field is almost a full second.
	 * To fix this we add a second if the fraction is above 900msec,
	 * a number which is a pure guess.
	 */
	if (ip->fraction > 900000000)
		ip->t++;

	TsNormalize(&ip->ts);
	if (ip->utc_status & 0x40) {
		/*
		 * UTC mode
		 * The fraction is a proper fraction of the seconds we have
		 * so we treat it as unsigned, and adjust for the seconds.
		 */
		ip->delta = ip->ts.ts.bt.frac;
		ip->delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
		ip->delta += ip->ts.ts.bt.sec - ip->t;
	} else {
		/*
		 * Fraction mode
		 * we have no usable seconds part, so we treat the fractions
		 * as a signed number and report it directly.
		 */
		ip->delta = (int64_t)ip->ts.ts.bt.frac;
		ip->delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
	}

	gmtime_r(&ip->t, &ip->tm);

	/*
	 * Check for LeapSecond status once per minute
	 */
	if (ip->tm.tm_sec == 22)
		oncore_sendmsg(ip, "Gj", 2);     /* Leap sec status */

	strftime(txt2, sizeof txt2, "%Y-%m-%d %H:%M:%S", &ip->tm);
	Log(loncore_msg,
	    "%s %ld %12.9f %s  %d of %d stat (%x %x %x %x %x %x %x %x %x %x %x %x)  %d %d %d %d",
	    p, (long)ip->t, ip->delta, txt2,
	    ip->tsat, ip->vsat,
	    (ip->rcv_status >> 13) & 7,
	    (ip->rcv_status >> 11) & 3,
	    (ip->rcv_status >> 10) & 1,
	    (ip->rcv_status >> 9) & 1,
	    (ip->rcv_status >> 8) & 1,
	    (ip->rcv_status >> 7) & 1,
	    (ip->rcv_status >> 6) & 1,
	    (ip->rcv_status >> 5) & 1,
	    (ip->rcv_status >> 4) & 1,
	    (ip->rcv_status >> 3) & 1,
	    (ip->rcv_status >> 1) & 3,
	    (ip->rcv_status >> 0) & 1,
	    ip->utc_status, ip->gmt_sign, ip->gmt_hour, ip->gmt_min);
	oncore_params(ip);
}

static void
handle_Hn(struct oncore *ip, struct oncore_msg *mp __unused)
{
	int i, j;
	u_char *buf = ip->serial_buf;

	/* Copy stuff we want to retain */
	for (i = 1; i < NSAT; i++)
		ip->persat[i].offset = 0;
	for (i = 0; i < NCHAN; i++) {
		j = buf[15 + i * 5];
		if (j == 0)
			continue;
		ip->persat[j].offset = be32dec(buf + 16 + i * 5);
	}

	ip->raim_solution = buf[6];
	ip->raim_status = buf[7];
	ip->raim_removed = be32dec(buf + 8);
	ip->raim_sigma = be16dec(buf + 12) * 1e-9;
	ip->raim_sawtooth = (int8_t)buf[14];

	do {
		/* If we're not really ready anyway, bail */
		if (ip->state != ST_SYNC_WAIT && ip->state != ST_SYNC)
			break;

		/* If RAIM is not happy, bail */
		if (ip->raim_solution != 0)
			break;

		/*
		 * If the utc_status is not stable, bail.  (This is becuase
		 * the flag gets set one second earlier than the offset gets
		 * applied)
		 */
		if (ip->utc_status != ip->lutc_status)
			break;

		/* one tracked sat, one vote */
		SourceParam(ip->sid, SRC_VOTES, ip->tsat);

		/*
		 * Record dispersion.  We have 1sigma number, arbitrarily
		 * define dispersion as two-sigma.
		 */
		SourceParam(ip->sid, SRC_DISPERSION, 2.0 * ip->raim_sigma);

		if (ip->utc_status & 0x40) {
			SourceDelta(ip->sid, &ip->ts, ip->delta,  SRC_UTC);
		} else {
			SourceDelta(ip->sid, &ip->ts, ip->delta, SRC_FRACTION);
		}
	} while (0);

}

static void
handle_Gj(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;
	struct tm tm;
	time_t now;

	Log(loncore_msg,
	    "LEAP Gj: p %d f %d Y %04u M %u D %d i.f %u.%09u h %u m %u s %u",
		buf[4], buf[5],
		be16dec(buf + 6), buf[8], buf[9],
		buf[10], be32dec(buf + 11),
		buf[15], buf[16], buf[17]);
	if (be16dec(buf + 6) == 0) {
		ip->leap_info[0] = '\0';
		SourceLeap(ip->sid, SRC_LEAP_UNKNOWN);
		return;
	}
	tm.tm_sec = buf[17];
	tm.tm_min = buf[16];
	tm.tm_hour = buf[15];
	tm.tm_mday = buf[9];
	tm.tm_mon = buf[8] - 1;
	tm.tm_year = be16dec(buf + 6) - 1900;
	ip->leap_when = timegm(&tm);
	time(&now);
	if (ip->leap_when - now > 24 * 3600 ||
	    buf[4] == buf[5])
		SourceLeap(ip->sid, SRC_LEAP_NO);
	else if (buf[4] < buf[5])
		SourceLeap(ip->sid, SRC_LEAP_INSERT);
	else if (buf[4] > buf[5])
		SourceLeap(ip->sid, SRC_LEAP_DELETE);

	sprintf(ip->leap_info,
	    "%04u-%02u-%02u %02u:%02u:%02u ",
	    be16dec(buf + 6), buf[8], buf[9], buf[15], buf[16], buf[17]);
	if (buf[4] < buf[5])
		strcat(ip->leap_info, "INSERT");
	else if (buf[4] > buf[5])
		strcat(ip->leap_info, "DELETE");
	else
		strcat(ip->leap_info, "NONE");
}

static void
handle_Bj(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;

	Log(loncore_msg,
	    "LEAP Bj: s %d", buf[4]);
}

static void
handle_Ag(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;

	ip->mask_angle = buf[4];
}

static void
handle_Gf(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;
	u_int u;

	u = be16dec(buf + 4);
	ip->raim_limit = u;
}

static void
handle_Co(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;

	Log(loncore_msg,
	    "Ion: alpha %d %d %d %d beta %d %d %d %d",
	    (char)buf[4], (char)buf[5], (char)buf[6], (char)buf[7],
	    (char)buf[8], (char)buf[9], (char)buf[10], (char)buf[11]);
	Log(loncore_msg,
	    "UTC: A0 %d A1 %d d(tLS) %d t(ot) %d WNt %d WN(lsf) %d DN %d d(LSF) %d",
	    (int)be32dec(buf + 12), (int)be32dec(buf + 16),
	    (char)buf[20], buf[21], buf[22], buf[23], buf[24], buf[25]);
	    
		
}

static void
handle_At(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;

	ip->site_survey = buf[4];
}

static void
handle_Bo(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;

	ip->utc_offset = buf[4];
}

static void
handle_Cj(struct oncore *ip, struct oncore_msg *mp)
{
	u_char *buf = ip->serial_buf;

	memcpy(ip->Cj, buf, mp->rs_len);
	ip->Cj[mp->rs_len - 3] = '\0';
}

static void
handle_Sz(struct oncore *ip, struct oncore_msg *mp __unused)
{

	Log(Lerror, "oncore %d GPS receiver ROM FAILURE (Sz %d)",
	    ip->unit, ip->serial_buf[4]);
	oncore_state(ip, ST_SICK);
}

static void
handle_Ea(struct oncore *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;
	int i, j, k;
	const char *p;
	struct ts ts2[2];
	char txt[50];

	i = PpsPickup(&ip->pps, ts2, 1);
	ip->ts = ts2[0];
	if (i == 1)
		p = TsFormat(&ip->ts, txt);
	else 
		p = "no_pps_pulse";

	if (ip->almanac_work == ALMANAC_LOADED)
		almanac_load(ip);
	if (ip->almanac_work == ALMANAC_DONE)
		almanac_done(ip);
	
	for (i = 0; i < NSAT; i++) {
		ip->persat[i].mode = 0;
		ip->persat[i].signal_strength = 0;
		ip->persat[i].iode = 0;
		ip->persat[i].status = 0;
	}
	ip->tm.tm_mon = buf[4] - 1;
	ip->tm.tm_mday = buf[5];
	ip->tm.tm_year = be16dec(buf + 6) - 1900;
	ip->tm.tm_hour = buf[8];
	ip->tm.tm_min = buf[9];
	ip->tm.tm_sec = buf[10];
	ip->fraction = be32dec(buf + 11);
	ip->lat = be32dec(buf + 15);
	ip->lon = be32dec(buf + 19);
	ip->ht = be32dec(buf + 23);
	/* 27 mmmm unused */
	/* 31 vv velocity */
	/* 33 hh heading */
	ip->dop = be16dec(buf + 35);
	/* 37 t DOP type */
	ip->vsat = buf[38];
	ip->tsat = buf[39];

	k = 0;
	for (i = 0; i < 8; i++) {
		j = buf[40 + i * 4];
		if (j == 0)
			continue;
		ip->persat[j].mode = buf[41 + i * 4];
		if (ip->persat[j].mode == 8)
			k++;
		ip->persat[j].signal_strength = buf[42 + i * 4];
		ip->persat[j].iode = 0;
		ip->persat[j].status = buf[43 + i * 4];
	}

	ip->lutc_status = ip->utc_status;
	ip->utc_status = buf[72] & 0x01;
	ip->rcv_status = buf[72];

	if ((ip->tm.tm_sec % 10) == 5)
		oncore_sendmsg(ip, "Bo\000", 3);     /* Utc offset */

	if (ip->tm.tm_sec == 22) {
		oncore_sendmsg(ip, "Gj", 2);     /* Leap sec status */
	}

	if (ip->tm.tm_sec == 33) {
		oncore_sendmsg(ip, "Bj0", 3);     /* Leap sec status */
	}

	if (ip->site_survey == 2) {
		if (ip->site_survey_timer > 0)
			ip->site_survey_timer--;
	} else {
		ip->site_survey_timer = 0;
	}

	if (ip->state == ST_SYNC_WAIT || ip->state == ST_SYNC) {
		if ((ip->rcv_status & 0x20) && !(ip->site_survey)) {
			oncore_sendmsg(ip, "At\002", 3); /* Start site survey */
			Log(loncore_msg, "Start Site Survey");
			ip->site_survey = 2;
			ip->site_survey_timer = 10000;
		}
	}

	ip->t = timegm(&ip->tm);

	/*
	 * Every so often, the Oncore will mis-estimate things and end
	 * up being too early.  Since the fraction field cannot be negative
	 * what happens is that the time fields are for the previous second
	 * and the fraction field is almost a full second.
	 * To fix this we add a second if the fraction is above 900msec,
	 * a number which is a pure guess.
	 */
	if (ip->fraction > 900000000) {
printf("Fraction %d\n", ip->fraction);
		ip->t++;
	}

	TsNormalize(&ip->ts);
	if (ip->utc_status == 0) {
		/*
		 * UTC mode
		 * The fraction is a proper fraction of the seconds we have
		 * so we treat it as unsigned, and adjust for the seconds.
		 */
		ip->delta = ip->ts.ts.bt.frac;
		ip->delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
		ip->delta += ip->ts.ts.bt.sec - ip->t;
	} else {
		/*
		 * Fraction mode
		 * we have no usable seconds part, so we treat the fractions
		 * as a signed number and report it directly.
		 */
		ip->delta = (int64_t)ip->ts.ts.bt.frac;
		ip->delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
	}

	gmtime_r(&ip->t, &ip->tm);

	Log(loncore_msg,
	    "EA %04d-%02d-%02d %02d:%02d:%02d.%09.0f v%dt%dl%d ST:%02x %s %.9f",
		ip->tm.tm_year + 1900, ip->tm.tm_mon + 1, ip->tm.tm_mday,
		ip->tm.tm_hour, ip->tm.tm_min, ip->tm.tm_sec,
		ip->fraction / 4.294967296,
		ip->vsat, ip->tsat, k, ip->rcv_status, p, ip->delta);
}

static void
handle_En(struct oncore *ip, struct oncore_msg *mp __unused)
{
	int i, j;
	u_char *buf = ip->serial_buf;

	/* Copy stuff we want to retain */
	for (i = 1; i < NSAT; i++)
		ip->persat[i].offset = 0;

	for (i = 0; i < 8; i++) {
		j = buf[26 + i * 5];
		if (j == 0)
			continue;
		ip->persat[j].offset = be32dec(buf + 26 + i * 5);
	}

	do {
		/* If we're not really ready anyway, bail */
		if (ip->state != ST_SYNC_WAIT && ip->state != ST_SYNC)
			break;

		/* If time is not UTC locked */
		if (buf[20] != 0)
			break;

		/* If RAIM is not happy, bail */
		if (buf[21] != 0)
			break;

		if (ip->utc_status != ip->lutc_status)
			break;

		/* one tracked sat, one vote */
		SourceParam(ip->sid, SRC_VOTES, ip->tsat);

		/*
		 * Record dispersion.  We have 1sigma number, arbitrarily
		 * define dispersion as two-sigma.
		 */
		j = be16dec(buf + 19);
		SourceParam(ip->sid, SRC_DISPERSION, 2 * j * 1e-9);

		if (ip->utc_status == 0x00 && ip->utc_offset != 0) {
			SourceDelta(ip->sid, &ip->ts, ip->delta,  SRC_UTC);
		} else {
			SourceDelta(ip->sid, &ip->ts, ip->delta, SRC_FRACTION);
		}
	} while (0);
}

/*
 * Initialize the hashed tables of known messages. 
 * The asserts is to make sure that the hash we use does what we think.
 */
int
init_oncore_tables(void)
{
	char const **cp;
	struct oncore_msg *mp;
	int i;

	oncore_msgs = calloc(sizeof *oncore_msgs, N_MASH);
	if (oncore_msgs == NULL) {
		nomem();
		return (-1);
	}
	for (cp = oncore_request; *cp != NULL; cp++) {
		i = msg_mash(*cp);
		Assert(i != N_MASH - 1, "Hash not perfect RQ(%4.4s) -> %d",
		    *cp, i);
		mp = oncore_msgs[i];
		if (mp == NULL) {
			mp = calloc(sizeof *mp, 1);
			if (mp == NULL)
				err(1, "calloc() failed");
			oncore_msgs[i] = mp;
		} 
		Assert(mp->rq_len == 0,
		    "Hash was not perfect RQ(%4.4s <--> %4.4s) -> %d",
		    *cp, mp->rq_format, i);
		mp->rq_len = strlen(*cp);
		mp->rq_format = *cp;
	}
	for (cp = oncore_response; *cp != NULL; cp++) {
		i = msg_mash(*cp);
		Assert(i != N_MASH - 1, "Hash not perfect RS(%4.4s) -> %d",
		    *cp, i);
		mp = oncore_msgs[i];
		if (mp == NULL) {
			mp = calloc(sizeof *mp, 1);
			if (mp == NULL)
				err(1, "calloc() failed");
			oncore_msgs[i] = mp;
		}
		Assert(mp->rs_len == 0,
		    "Hash was not perfect RS(%4.4s <--> %4.4s) -> %d",
		    *cp, mp->rs_format, i);
		mp->rs_len = strlen(*cp);
		mp->rs_format = *cp;
	}
	oncore_msgs[msg_mash("@@Ag")]->handler = handle_Ag;
	oncore_msgs[msg_mash("@@Bb")]->handler = handle_Bb;
	oncore_msgs[msg_mash("@@Cb")]->handler = handle_Cb;
	oncore_msgs[msg_mash("@@Ch")]->handler = handle_NOOP;
	oncore_msgs[msg_mash("@@Ea")]->handler = handle_Ea;
	oncore_msgs[msg_mash("@@En")]->handler = handle_En;
	oncore_msgs[msg_mash("@@Ha")]->handler = handle_Ha;
	oncore_msgs[msg_mash("@@Hn")]->handler = handle_Hn;
	oncore_msgs[msg_mash("@@Cf")]->handler = handle_Cf;
	oncore_msgs[msg_mash("@@Ia")]->handler = handle_IaFa;
	oncore_msgs[msg_mash("@@Fa")]->handler = handle_IaFa;
	oncore_msgs[msg_mash("@@Gf")]->handler = handle_Gf;
	oncore_msgs[msg_mash("@@Gj")]->handler = handle_Gj;
	oncore_msgs[msg_mash("@@Co")]->handler = handle_Co;
	oncore_msgs[msg_mash("@@Cj")]->handler = handle_Cj;
	oncore_msgs[msg_mash("@@Bo")]->handler = handle_Bo;
	oncore_msgs[msg_mash("@@At")]->handler = handle_At;
	oncore_msgs[msg_mash("@@Bj")]->handler = handle_Bj;
	oncore_msgs[msg_mash("@@Sz")]->handler = handle_Sz;
	return (0);
}
