/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore_serial.c,v 1.32 2005/12/30 10:07:58 phk Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <termios.h>
#include <pthread.h>
#include <configkit.h>
#include <isc/eventlib.h>

#include <ntpns.h>
#include "oncore_cfg.h"
#include "oncore_hash.h"
#include "oncore.h"


static void serial_input(evContext ctx, void *uap, int fd, int eventmask);

int
cfg_oncore_serial(CK_ARG, const char **serial)
{
	struct oncore *ip;
        int fd, i;
        struct termios t;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong Oncore magic");
	if (!ck_WordVar(CK_PASS, serial, &ip->serial))
		return (0);

	if (ip->serial == NULL) {
		if (ip->serial_fd >= 0) {
			evDeselectFD(C, ip->serial_fid);
			close(ip->serial_fd);
			ip->serial_fd = -1;
			oncore_state(ip, ST_UNCONF);
			return (0);
		}
	}
        fd = open(ip->serial, O_RDWR | O_NDELAY);
        if (fd < 0) {
                ck_printf(CK_PASS, "open(%s): %s\n",
		    ip->serial, strerror(errno));
		return (-1);
	}
        i = tcgetattr(fd, &t);
        if (i < 0) {
		ck_printf(CK_PASS, "tcgetattr(%s): %s\n",
		    ip->serial, strerror(errno));
		close(fd);
		return (-1);
	}
        cfmakeraw(&t);
        cfsetspeed(&t, B9600);
        t.c_cflag &= ~CSIZE;
        t.c_cflag |= CS8;
        t.c_cflag |= CLOCAL;
        t.c_cc[VMIN] = 0;
        t.c_cc[VTIME] = 20;
        i = tcsetattr(fd, TCSANOW, &t);
        if (i < 0) {
		ck_printf(CK_PASS, "tcsetattr(%s): %s\n", *serial, strerror(errno));
		close(fd);
		return (-1);
	}
	i = evSelectFD(C, fd, EV_READ | EV_EXCEPT, serial_input,
	    ip, &ip->serial_fid);
	if (i) 
		err(1, "evSelectFD(%s)", *serial);
	ip->serial_fd = fd;
	ip->serial = strdup(*serial);
	ip->serial_idx = 0;
	oncore_state(ip, ST_CONFIG);
	return (0);
}


static void
serial_input(evContext ctx __unused, void *uap, int fd, int eventmask __unused)
{
	struct oncore *ip;
	int i, l, mm;
	struct oncore_msg *mp;
	u_char s;

	ip = uap;

	/*
	 * Read as much as we can fit in our buffer
	 */
	i = read(fd, ip->serial_buf + ip->serial_idx,
	     sizeof(ip->serial_buf) - ip->serial_idx);
	if (i <= 0)
		return;
	ip->serial_idx += i;

	for(;;) {
		/*
		 * Stop if don't have enough data for a complete message
		 * the shortest possible message is 7 bytes: "@@xxC\r\n"
		 */
		if (ip->serial_idx < 7)
			return;

		/*
		 * See if we know this one and if the header looks OK
		 */
		mm = msg_mash(ip->serial_buf);
		mp = oncore_msgs[mm];
		if (ip->serial_buf[0] != '@' ||
		    ip->serial_buf[1] != '@' ||
		    mp == NULL ||
		    mp->rs_len == 0 ||
		    ip->serial_buf[2] != mp->rs_format[2] ||
		    ip->serial_buf[3] != mp->rs_format[3]) {
			/*
			 * Look for a '\n' and up to and including that,
			 * otherwise, skip entire buffer.
			 */
			for (i = 0; i < ip->serial_idx; i++)
				if (ip->serial_buf[i] == '\n')
					break;
			if (i < ip->serial_idx)
				i++;
			record("M12_DISC", ip->unit, ip->serial_buf, i);
			Log(loncore_serial,
			    "%02x %02x %02x %02x skipping %d chars",
			    ip->serial_buf[0], ip->serial_buf[1],
			    ip->serial_buf[2], ip->serial_buf[3], i);
			memcpy(ip->serial_buf,
			    ip->serial_buf + i,
			    ip->serial_idx - i);
			ip->serial_idx -= i;
			continue;
		}

		l = mp->rs_len;

		/*
		 * Do we have all the bytes in this message yet ?
		 */
		if (ip->serial_idx < l)
			return;

		/* Calculate XOR checksum */
		s = 0;
		for (i = 2; i < l - 2; i++)
			s ^= ip->serial_buf[i];

		if (s != 0) {
			Log(loncore_serial,
			    "wrong XOR at end of %4.4s (%02x/%02x)",
			    mp->rs_format, ip->serial_buf[l - 3], s);
			record("M12_XOR", ip->unit, ip->serial_buf, l);
		} else if (ip->serial_buf[l - 2] != '\r') {
			Log(loncore_serial, "no CR at end of %4.4s (%02x)",
			    mp->rs_format, ip->serial_buf[l - 2]);
			record("M12_CR", ip->unit, ip->serial_buf, l);
		} else if (ip->serial_buf[l - 1] != '\n') {
			Log(loncore_serial, "no LF at end of %4.4s (%02x)",
			    mp->rs_format, ip->serial_buf[l - 1]);
			record("M12_LF", ip->unit, ip->serial_buf, l);
		} else {
			record("M12_OK", ip->unit, ip->serial_buf, l);
			Log(loncore_serial, "Rcv: %4.4s[%d]", ip->serial_buf, l);
			if (mp->handler != NULL)
				mp->handler(ip, mp);
			if (ip->state == ST_LISTEN)
				oncore_state(ip, ST_CONTACT);
			if (ip->state == ST_SYNC_WAIT ||
			    ip->state == ST_SYNC)
				oncore_reset_timer(ip);
		}

		/*
		 * Pull up anything at the end and start over
		 */
		memcpy(ip->serial_buf, ip->serial_buf + l, ip->serial_idx - l);
		ip->serial_idx -= l;
	}
}

void
oncore_sendmsg(struct oncore *ip, const void *p, u_int len)
{
	u_char cs = 0;
	const u_char *msg;

	msg = p;
	record("M12_XMIT", ip->unit, msg, len);
	Log(loncore_serial, "Xmt: @@%2.2s[%d]", msg, len + 5);
	write(ip->serial_fd, "@@", 2);
	write(ip->serial_fd, msg, len);
	while (len--)
		cs ^= *msg++;
	write(ip->serial_fd, &cs, 1);
	write(ip->serial_fd, "\r\n", 2);
}
